/*
 * KROC translator: Instruction prototypes.
 *
 * $Source: /extra/play/cvs/kroc/src/host/f_proto.h,v $
 *
 * $Id: f_proto.h,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 * 
 * (c) Copyright 1999 R. Ivimey-Cook <ruthc@sharra.demon.co.uk>
 *
 * (C) Copyright 1995,1996,1997 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *   /proj/kroc/develop/src/RCS/f_proto.h,v
 *   f_proto.h,v 1.14 1997/01/08 17:24:59 djb1 Exp
 *
 * These are all the Transputer instructions listed in
 * "The transputer instruction set - a compiler writer's guide",
 * (INMOS, March 1987) including those that are not implemented.
 * Also directives generated by the "-zasm" option,
 * and pseudo-directives used for debugging etc.
 *
 * Should be independent of target, except perhaps for floating-point.
 */

/* f_alt.c */

void ALT (void);
void ALTWT (void);
void ALTEND (void);
void ENBC (void);
void ENBS (void);
void DISC (void);
void DISS (void);
void TIN (void);
void TALT (void);
void TALTWT (void);
void ENBT (void);
void DIST (void);

/* f_arith.c */

void DIFF (void);
void ADD (void);
void PROD (void);
void SUB (void);
void REM (void);
void DIV (void);
void NOT (void);
void XOR (void);
void SHR (void);
void SHL (void);
void AND (void);
void OR (void);
void SUM (void);
void MUL (void);

/* f_byte.c */

int hex1 (int i);
int hex2 (int i);
int word (char *buff);
void BYTE (void);

/* f_comm.c */

void IN (void);
void OUT (void);
void OUTBYTE (void);
void OUTWORD (void);

/* f_direct.c */

void CALL_KRN (const int);
void GEN_TRAP (const int);

#if defined(TARGET_CPU_ALPHA)
int big (int n, int nbits);
#elif defined(TARGET_CPUxsPARC)
int big (int n);
#endif

void J (void);
void LDLP (void);
void PFIX (void);
void LDNL (void);
void LDC (void);
void LDNLP (void);
void NFIX (void);
void LDL (void);
void ADC (void);
void CALL (void);
void CJ (void);
void AJW (void);
void EQC (void);
void STL (void);
void STNL (void);
void OPR (void);

/* f_directives.c - Directives */

void ALIGN (void);
void GLOBAL (void);
void DESCRIPTOR (void);
void EXTERN (void);
void PATCH (void);

/* f_error.c */

void SETERR (void);
void TESTERR (void);
void STOPERR (void);
void CLRHALTERR (void);
void SETHALTERR (void);
void TESTHALTERR (void);
void FPCHKERR (void);
void FPTESTERR (void);
void FPUSETERR (void);
void FPUCLRERR (void);

/* f_etc.c */

void REV (void);
void LB (void);
void BSUB (void);
void GCALL (void);
void GT (void);
void WSUB (void);
void RESETCH (void);
void CSUB0 (void);
void LDPI (void);
void XDBLE (void);
void LDPRI (void);
void RET (void);
void LEND (void);
void LDTIMER (void);
void STTIMER (void);
void TESTPRANAL (void);
void BCNT (void);
void XWORD (void);
void SB (void);
void GAJW (void);
void WCNT (void);
void MINT (void);
void MOVE (void);
void CSNGL (void);
void CCNT1 (void);
void CWORD (void);

/* f_float.c */

#if defined(FLOAT)
void FPDUP (void);
void FPREV (void);
void FPLDNLSN (void);
void FPLDNLDB (void);
void FPLDNLSNI (void);
void FPLDNLDBI (void);
void FPSTNLSN (void);
void FPSTNLDB (void);
void FPURN (void);
void FPURZ (void);
void FPURP (void);
void FPURM (void);
void FPADD (void);
void FPSUB (void);
void FPMUL (void);
void FPDIV (void);
void FPUSQRTFIRST (void);
void FPUSQRTSTEP (void);
void FPUSQRTLAST (void);
void FPREMFIRST (void);
void FPREMSTEP (void);
void FPLDZEROSN (void);
void FPLDZERODB (void);
void FPUMULBY2 (void);
void FPUDIVBY2 (void);
void FPUEXPINC32 (void);
void FPUEXPDEC32 (void);
void FPUABS (void);
void FPLDNLADDSN (void);
void FPLDNLADDDB (void);
void FPLDNLMULSN (void);
void FPLDNLMULDB (void);
void FPGT (void);
void FPEQ (void);
void FPORDERED (void);
void FPNAN (void);
void FPNOTFINITE (void);
void FPUR32TOR64 (void);
void FPUR64TOR32 (void);
void FPINT (void);
void FPSTNLI32 (void);
void FPUCHKI32 (void);
void FPUCHKI64 (void);
void FPRTOI32 (void);
void FPI32TOR32 (void);
void FPI32TOR64 (void);
void FPB32TOR64 (void);
void FPUNOROUND (void);
void FPENTRY (void);
#endif /* FLOAT */

/* f_long.c */

void LADD (void);
void LDIV (void);
void LMUL (void);
void LSHR (void);
void LSHL (void);
void LSUM (void);
void LSUB (void);
void LDIFF (void);
void NORM (void);

/* f_move2d.c */

void MOVE2DINIT (void);
void MOVE2DALL (void);
void MOVE2DNONZERO (void);
void MOVE2DZERO (void);

/* f_proc.c */

void ENDP (void);
void STARTP (void);
void STOPP (void);
void STHF (void);
void STLB (void);
void STLF (void);
void RUNP (void);
void SAVEH (void);
void SAVEL (void);
void STHB (void);

/* f_pseudo.c - Pseudo-directives for debugging etc.*/

void dcheck (int max);
void message (const char *s, int m, int n);
void _CHECK (void);
void _COND (void);
void _ECHO (void);
void _STACK (void);
void _STATE (void);
void _STIME (void);
#if defined(FLOAT)
void _DOUBLE (void);
void _FSTACK (void);
void _FTYPE (void);
void _SINGLE (void);
void _FDUMP (void);
#endif /* FLOAT */
void _DUMP (void);
void _ERROR (void);
void _ABORT (void);
void unimp (void);

/* f_t414.c */

void UNPACKSN (void);
void ROUNDSN (void);
void POSTNORMSN (void);
void LDINF (void);
void CFLERR (void);

/* f_t800.c */

void DUP (void);
void FMUL (void);
void CRCWORD (void);
void CRCBYTE (void);
void BITCNT (void);
void BITREVWORD (void);
void BITREVNBITS (void);
void WSUBDB (void);

/* f_t9000.c */

void CAUSEERROR (void);
void CB (void);
void CBU (void);
void CS (void);
void CSU (void);
void FPADDDBSN (void);
void FPGE (void);
void FPLG (void);
void FPREM (void);
void GTU (void);
void LBX (void);
void LDCNT (void);
void LDTH (void);
void LS (void);
void LSX (void);
void SS (void);
void SSUB (void);
void VIN (void);
void VOUT (void);
void XSWORD (void);
