/*
 * KROC translator: Deal with cases when different paths
 * to a label have different stack levels.
 *
 * $Source: /extra/play/cvs/kroc/src/host/label.c,v $
 *
 * $Id: label.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 * Derived from label.c:
 * (C) Copyright 1995,1996,1997 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *
 *   /proj/kroc/develop/src/RCS/label.c,v    
 *   label.c,v 1.16 1997/01/08 17:23:42 djb1 Exp
 *
 * label.c:
 * (C) Copyright 1999 R. Ivimey-Cook <ruthc@sharra.demon.co.uk>
 *
 * The shuffling code depends on the target.
 *
 * Should make more use of shuffle information.
 */

static char label_c_rcsid[] = "$Id: label.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $";

#include "octran.h"
#include <stdlib.h>

static state_type labtab[LT_SIZE];	/* label table */

static void plant (const char *l);

/* 'l' is always "b_label") */
static void plant (const char *l)		/* plant label in output code */
{
    posn = LINEUP;		/* lie about alignement for comments */
    
#if defined(TARGET_CPU_ARM)
    if (l[0] == 'L')
    {
        fprintf (out, "%s", l); NL;
    }
    else
    {
        fprintf (out, "|%s|", l); NL;
    }
#else
    fprintf (out, "%s:", l); NL;	/* prefix ".L" for Solaris? */
#endif
}

void init_labtab (void)
{
    int i;
    for (i = 0; i < LT_SIZE; ++i)
        labtab[i].i = S_EMPTY;
}

int shuffle (int s, int t)		/* shuffle stack from level s to t */
{
    int n = 0;
    istackp = s;
    if (s < t) {
        fputs ("\n*** Impossible shuffle", stderr);
        error ();
    }
  
    if (s - t > ishuff) {
        fputs ("\n*** Shuffling too far", stderr);
        /*	error ();	*/
        where ();
    }
    /* else */ if (s > t && t > 0) {
        if (o_verbose) {
            fprintf (stderr, "\n... Shuffling stack from %d to %d\n", s, t);
            where ();
        }

        if (s == 3) {
            if (t == 2) {
                mnem("mov"); B; C; COMMENT("shuffle 3 to 2"); nl;
                mnem("mov"); A; B; nl;
            } else { /* t == 1 */
                mnem("mov"); A; C; COMMENT("shuffle 3 to 1"); nl;
            }
        } else { /* s == 2 */
            mnem("mov"); A; B; COMMENT("shuffle 2 to 1"); nl;
        }

        n = 1;
    }
    istackp = t;
    return n;
}


void deflab (void)			/* called when label defined */
{
    int l;
    state_type state;
    if (b_label[0] != 'L') {
#if defined(TARGET_CPU_SPARC)
        if (o_global & (b_label[0] == '$'))
            fputc (CCHAR, out);
#endif
        plant (b_label);
        return;
    }
  
    l = atoi (&b_label[1]);
    state = labtab[l];			/* previous state */

    if (state.i == S_EMPTY)		/* no (forward) reference to it */
        plant (b_label);
    else {				/* shuffle if necessary */
        if (istackp >= state.B_ISTACK) {
            lshufflecount += shuffle (istackp, state.B_ISTACK); /* (may not shuffle) */
            plant (b_label);
        } else { /* istackp < state.B_ISTACK */
            if (istackp == 0)
                plant (b_label);
            else { /* istackp > 0 */
#if defined(TARGET_CPU_SPARC)
                mnem("ba,a");
#elif defined(TARGET_CPU_ALPHA)
                mnem("br");
#elif defined(TARGET_CPU_ARM)
                mnem("jump"); /* do we need this ??? */
#endif
                f(1); COMMENT("skip shuffle"); nl;
                plant (b_label);
                lshufflecount += shuffle (state.B_ISTACK, istackp);
                lab(1); nl;
            }
        }
    }

    labtab[l].B_ISTACK = istackp;
    labtab[l].B_ISHUFF = ishuff;
/* etc.	*/
    labtab[l].B_DEF = 1;
}


state_type reflab (int s)		/* called when label referenced */
{
    int l;
    state_type state;

    state.i=0;
        
    if (b_operand[0] != 'L')
        return state;

    l = atoi (&b_operand[1]);
    state = labtab[l];			/* state when referenced */

    if (state.i == S_EMPTY) {		/* first time */
        labtab[l].B_ISTACK = s;		/* record state */
        labtab[l].B_ISHUFF = ishuff;
        /* etc.	*/
        labtab[l].B_FLAG = 1;
    } else {				/* already met */
        if (s > state.B_ISTACK) {
            if (state.B_ISTACK == 0)		/* ??? */
                state.B_ISTACK = 1;		/* !!! */
            lshufflecount += shuffle (s, state.B_ISTACK);
        } else if (s == 0) {		/* must be J */
            labtab[l].B_ISTACK = 0;
            labtab[l].B_ISHUFF = 0;
            /* etc.	*/
            labtab[l].B_FLAG = 1;
        } else if (s < state.B_ISTACK) {
            fprintf (stderr, "\n*** Conflicting stack levels %d and %d",
                     s, state.B_ISTACK);
            /*	error ();	*/
            where ();
        }

    } /* if (state.i == S_EMPTY) */
    return state;
}
