/*
 * $Id: lookup.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 * (C) Copyright 1995,1996,1997 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *
 * Should be independent of target.
 */

static char lookup_c_rcsid[] = "$Id: lookup.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $";

#include "octran.h"
#include <string.h>

/*      Symbol table for Transputer mnemonics, in ASCII order
 *
 * Notes:
 *      (1)     Should not occur at this level
 *      (2)     Don't understand
 *      (3)     Require high priority
 *      (4)     Not implemented (not required?)
 *      (5)     T414 instructions (not implemented)
 *      (6)     These have operands in the wrong place
 *      (7)     T9000 instructions (not all used by occam)
 *              Most T9000 floating-point instructions are synonyms
 */

const it_type itable [] =
/*      {"mnemonic",    function,       flags},         */
{       {"",            _ERROR,         0},     /* dummy */

	{".BYTE",       BYTE,           0},     /* real directive */

	{".abort",      _ABORT,         0},     /* pseudo-directives */
	{".check",      _CHECK,         F_OPERAND},
	{".cond",       _COND,          F_OPERAND},
#ifdef FLOAT
	{".double",     _DOUBLE,        0},
#endif /* FLOAT */
	{".dump",       _DUMP,          0},
	{".echo",       _ECHO,          F_OPERAND},
#ifdef FLOAT
	{".fdump",      _FDUMP,         0},
	{".fstack",     _FSTACK,        F_OPERAND},
	{".ftype",      _FTYPE,         F_OPERAND},
	{".single",     _SINGLE,        0},
#endif /* FLOAT */
	{".stack",      _STACK,         F_OPERAND},
	{".state",      _STATE,         F_OPERAND},
	{".time",       _STIME,         F_OPERAND},

	{"adc",         ADC,            F_OPERAND},     /* instructions */
	{"add",         ADD,            0},
	{"ajw",         AJW,            F_OPERAND},
	{"align",       ALIGN,          0},
	{"alt",         ALT,            0},
	{"altend",      ALTEND,         0},
	{"altwt",       ALTWT,          0},
	{"and",         AND,            0},
	{"bcnt",        BCNT,           0},
	{"bitcnt",      BITCNT,         0},
	{"bitrevnbits", BITREVNBITS,    0},
	{"bitrevword",  BITREVWORD,     0},
	{"bsub",        BSUB,           0},
	{"byte",        BYTE,           0},     /* (6) */
	{"call",        CALL,           F_OPERAND},
	{"causeerror",  CAUSEERROR,     0},     /* (7) */
	{"cb",          CB,             0},     /* (7) */
	{"cbu",         CBU,            0},     /* (7) */
	{"ccnt1",       CCNT1,          0},
	{"cflerr",      CFLERR,         0},     /* (5) */
	{"cj",          CJ,             F_OPERAND | F_CJ},
	{"clrhalterr",  CLRHALTERR,     0},
	{"crcbyte",     CRCBYTE,        0},
	{"crcword",     CRCWORD,        0},
	{"cs",          CS,             0},     /* (7) */
	{"csngl",       CSNGL,          0},
	{"csu",         CSU,            0},     /* (7) */
	{"csub0",       CSUB0,          0},
	{"cword",       CWORD,          0},
	{"descriptor",  DESCRIPTOR,     0},     /* (6) */
	{"diff",        DIFF,           0},
	{"disc",        DISC,           0},
	{"diss",        DISS,           0},
	{"dist",        DIST,           0},
	{"div",         DIV,            0},
	{"dup",         DUP,            0},
	{"enbc",        ENBC,           0},
	{"enbs",        ENBS,           0},
	{"enbt",        ENBT,           0},
	{"endp",        ENDP,           0},
	{"eqc",         EQC,            F_OPERAND | F_EQC},
	{"extern",      EXTERN,         F_OPERAND},
	{"fmul",        FMUL,           0},

#ifdef FLOAT

	{"fpabs",       FPUABS,         0},     /* (7) */
	{"fpadd",       FPADD,          0},
	{"fpadddbsn",   FPADDDBSN,      0},     /* (7) */
	{"fpb32tor64",  FPB32TOR64,     0},
	{"fpchkerr",    FPCHKERR,       0},
	{"fpchki32",    FPUCHKI32,      0},     /* (7) */
	{"fpchki64",    FPUCHKI64,      0},     /* (7) */
	{"fpdiv",       FPDIV,          0},
	{"fpdivby2",    FPUDIVBY2,      0},     /* (7) */
	{"fpdup",       FPDUP,          0},
	{"fpentry",     FPENTRY,        0},     /* (1) */
	{"fpeq",        FPEQ,           0},
	{"fpexpdec32",  FPUEXPDEC32,    0},     /* (7) */
	{"fpexpinc32",  FPUEXPINC32,    0},     /* (7) */
	{"fpge",        FPGE,           0},     /* (7) */
	{"fpgt",        FPGT,           0},
	{"fpi32tor32",  FPI32TOR32,     0},
	{"fpi32tor64",  FPI32TOR64,     0},
	{"fpint",       FPINT,          0},
	{"fpldnladddb", FPLDNLADDDB,    0},
	{"fpldnladdsn", FPLDNLADDSN,    0},
	{"fpldnldb",    FPLDNLDB,       0},
	{"fpldnldbi",   FPLDNLDBI,      0},
	{"fpldnlmuldb", FPLDNLMULDB,    0},
	{"fpldnlmulsn", FPLDNLMULSN,    0},
	{"fpldnlsn",    FPLDNLSN,       0},
	{"fpldnlsni",   FPLDNLSNI,      0},
	{"fpldzerodb",  FPLDZERODB,     0},
	{"fpldzerosn",  FPLDZEROSN,     0},
	{"fplg",        FPLG,           0},     /* (7) */
	{"fpmul",       FPMUL,          0},
	{"fpmulby2",    FPUMULBY2,      0},     /* (7) */
	{"fpnan",       FPNAN,          0},
	{"fpnotfinite", FPNOTFINITE,    0},
	{"fpordered",   FPORDERED,      0},
	{"fpr32tor64",  FPUR32TOR64,    0},     /* (7) */
	{"fpr64tor32",  FPUR64TOR32,    0},     /* (7) */
	{"fprem",       FPREM,          0},     /* (7) */
	{"fpremfirst",  FPREMFIRST,     0},
	{"fpremstep",   FPREMSTEP,      0},
	{"fprev",       FPREV,          0},
	{"fprm",        FPURM,          0},     /* (7) */
	{"fprn",        FPURN,          0},     /* (7) */
	{"fprp",        FPURP,          0},     /* (7) */
	{"fprtoi32",    FPRTOI32,       0},
	{"fprz",        FPURZ,          0},     /* (7) */
	{"fpsqrt",      FPUSQRTLAST,    0},     /* (7) */
	{"fpstnldb",    FPSTNLDB,       0},
	{"fpstnli32",   FPSTNLI32,      0},
	{"fpstnlsn",    FPSTNLSN,       0},
	{"fpsub",       FPSUB,          0},
	{"fptesterr",   FPTESTERR,      0},
	{"fpuabs",      FPUABS,         0},
	{"fpuchki32",   FPUCHKI32,      0},     /* (4) */
	{"fpuchki64",   FPUCHKI64,      0},     /* (4) */
	{"fpuclrerr",   FPUCLRERR,      0},
	{"fpudivby2",   FPUDIVBY2,      0},
	{"fpuexpdec32", FPUEXPDEC32,    0},
	{"fpuexpinc32", FPUEXPINC32,    0},
	{"fpumulby2",   FPUMULBY2,      0},
	{"fpunoround",  FPUNOROUND,     0},
	{"fpur32tor64", FPUR32TOR64,    0},
	{"fpur64tor32", FPUR64TOR32,    0},
	{"fpurm",       FPURM,          0},
	{"fpurn",       FPURN,          0},
	{"fpurp",       FPURP,          0},
	{"fpurz",       FPURZ,          0},
	{"fpuseterr",   FPUSETERR,      0},
	{"fpusqrtfirst",FPUSQRTFIRST,   0},
	{"fpusqrtlast", FPUSQRTLAST,    0},
	{"fpusqrtstep", FPUSQRTSTEP,    0},

#endif /* FLOAT */

	{"gajw",        GAJW,           0},
	{"gcall",       GCALL,          0},
	{"global",      GLOBAL,         F_OPERAND},     /* (6) */
	{"gt",          GT,             0},
	{"gtu",         GTU,            0},     /* (7) */
	{"in",          IN,             0},
	{"j",           J,              F_OPERAND},
	{"ladd",        LADD,           0},
	{"lb",          LB,             0},
	{"lbx",         LBX,            0},     /* (7) */
	{"ldc",         LDC,            F_OPERAND},
	{"ldcnt",       LDCNT,          0},     /* (7) */
	{"ldiff",       LDIFF,          0},
	{"ldinf",       LDINF,          0},     /* (5) */
	{"ldiv",        LDIV,           0},
	{"ldl",         LDL,            F_OPERAND},
	{"ldlp",        LDLP,           F_OPERAND},
	{"ldnl",        LDNL,           F_OPERAND},
	{"ldnlp",       LDNLP,          F_OPERAND},
	{"ldpi",        LDPI,           F_O_LABEL},
	{"ldpri",       LDPRI,          0},
	{"ldth",        LDTH,           0},     /* (7) */
	{"ldtimer",     LDTIMER,        0},
	{"lend",        LEND,           F_O_LABEL},
	{"lmul",        LMUL,           0},
	{"ls",          LS,             0},     /* (7) */
	{"lshl",        LSHL,           0},
	{"lshr",        LSHR,           0},
	{"lsub",        LSUB,           0},
	{"lsum",        LSUM,           0},
	{"lsx",         LSX,            0},     /* (7) */
	{"mint",        MINT,           0},
	{"move",        MOVE,           0},
	{"move2dall",   MOVE2DALL,      0},
	{"move2dinit",  MOVE2DINIT,     0},
	{"move2dnonzero",
			MOVE2DNONZERO,  0},
	{"move2dzero",  MOVE2DZERO,     0},
	{"mul",         MUL,            0},
	{"nfix",        NFIX,           F_OPERAND},     /* (1) */
	{"norm",        NORM,           0},
	{"not",         NOT,            0},
	{"opr",         OPR,            F_OPERAND},     /* (1) */
	{"or",          OR,             0},
	{"out",         OUT,            0},
	{"outbyte",     OUTBYTE,        0},
	{"outword",     OUTWORD,        0},
	{"patch",       PATCH,          0},
	{"pfix",        PFIX,           F_OPERAND},     /* (1) */
	{"postnormsn",  POSTNORMSN,     0},     /* (5) */
	{"prod",        PROD,           0},
	{"rem",         REM,            0},
	{"resetch",     RESETCH,        0},
	{"ret",         RET,            0},
	{"rev",         REV,            0},
	{"roundsn",     ROUNDSN,        0},     /* (5) */
	{"runp",        RUNP,           0},
	{"saveh",       SAVEH,          0},     /* (3) */
	{"savel",       SAVEL,          0},
	{"sb",          SB,             0},
	{"seterr",      SETERR,         0},
	{"sethalterr",  SETHALTERR,     0},
	{"shl",         SHL,            0},
	{"shr",         SHR,            0},
	{"ss",          SS,             0},     /* (7) */
	{"ssub",        SSUB,           0},     /* (7) */
	{"startp",      STARTP,         0},
	{"sthb",        STHB,           0},     /* (3) */
	{"sthf",        STHF,           0},     /* (3) */
	{"stl",         STL,            F_OPERAND},
	{"stlb",        STLB,           0},
	{"stlf",        STLF,           0},
	{"stnl",        STNL,           F_OPERAND},
	{"stoperr",     STOPERR,        0},
	{"stopp",       STOPP,          0},
	{"sttimer",     STTIMER,        0},     /* (2) */
	{"sub",         SUB,            0},
	{"sum",         SUM,            0},
	{"talt",        TALT,           0},
	{"taltwt",      TALTWT,         0},
	{"testerr",     TESTERR,        0},
	{"testhalterr", TESTHALTERR,    0},
	{"testpranal",  TESTPRANAL,     0},
	{"tin",         TIN,            0},
	{"unpacksn",    UNPACKSN,       0},     /* (5) */
	{"vin",         VIN,            0},     /* (7) */
	{"vout",        VOUT,           0},     /* (7) */
	{"wcnt",        WCNT,           0},
	{"wsub",        WSUB,           0},
	{"wsubdb",      WSUBDB,         0},
	{"xdble",       XDBLE,          0},
	{"xor",         XOR,            0},
	{"xsword",      XSWORD,         0},     /* (7) */
	{"xword",       XWORD,          0},

	{"",            _ERROR,         0}
};


static int last = sizeof (itable)/sizeof (itable[1]) - 1;

/* look up Transputer mnemonic in symbol table */
int lookup (const char *string)
{
  int low = 0;
  int high = last;
  int mid, cmp;
  while ((high - low) > 1) {
    mid = (high + low)/2;
    cmp = strcmp (string, itable[mid].mnem);
    if (cmp < 0)
      high = mid;
    else if (cmp > 0)
      low  = mid;
    else /* cmp == 0 */
      return mid;     /* got it */
  }
  return last;                    /* not there */
}


/* translate Transputer token to target instructions */
void translate (int token)
{
  char flags = itable[token].flags;
  if (o_opt && (flags & F_CJ) == 0 && cond &&
      ((flags & F_EQC) == 0 || n_operand != 0))
    defcond (cond); /* see EQC */

  if (o_label && l_operand[0] != 0 && (flags & F_O_LABEL) == 0)
    defldc ();

  COMMENT(itable[token].mnem); nl;
  
  (itable[token].func)();
  
  started = True;
}
