/*
 * KROC translator: Utility functions.
 *
 * $Source: /extra/play/cvs/kroc/src/host/util.c,v $
 *
 * $Id: util.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 * Derived from util.c:
 * (C) Copyright 1995,1996,1997 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *
 *   /proj/kroc/develop/src/RCS/util.c,v
 *   util.c,v 1.44 1997/01/08 17:24:37 djb1 Exp
 *
 * util.c:
 * (C) Copyright 1998 R. Ivimey-Cook <ruthc@sharra.demon.co.uk>
 *
 * Most of the functions here are concered with generating parts
 * of assembly-language instructions for the target machine.
 * Hence the details depend on the target (and its assembly
 * language), though similar functions will usually be required.
 */

static char util_c_rcsid[] = "$Id: util.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $";

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/time.h>
#include "octran.h"

static void date (void);

pt_type proctab[PT_SIZE];	/* procedure table */

/*	Should be independent of target */

void printst (void)			/* print symbol table */
{
    int i;
    if (o_verbose) {
        fputs ("\n... Symbol table\n", stderr);
        fprintf (stderr, "    %-5s  %-8s  %-20s  %2s  %2s  %2s\n",
                 "Label", "Type", "Name", "is", "fs", "ft");

        for (i = 0; i <= stcount; ++i)
            fprintf (stderr, "    %-5s  %-8s  %-20s  %2d  %2d  %2d\n",
                     proctab[i].label, ptype[proctab[i].S_PTYPE],
                     proctab[i].proc, proctab[i].S_ISTACK,
                     proctab[i].S_FSTACK, proctab[i].S_FTYPES);
    }
}


void where (void)			/* report where something happened */
{
    fprintf (stderr, "\n    at line %d", iline);
    if (b_proc[0] != 0)
        fprintf (stderr, " in %s %s", ptype[proctab[stcount].S_PTYPE], b_proc);
    if (filename[0] != 0)
        fprintf (stderr, "\n    near Occam line %d in file %s", fline, filename);
    fputc ('\n', stderr);
}


void error (void)			/* report fatal error */
{
    where ();
    fprintf (stderr, "\n%s\n", buffer);
    printst ();
    fputs ("\n*** Exit\n", stderr);
    fputc ('\n', out);
    if (o_abort)
        abort ();
    exit (1);
}


void lineup (void)			/* line up comments at column LINEUP */
{
    do {
        fputc ('\t', out);
        posn += 8;
    } while (posn < LINEUP);
}


void newline (void)			/* comment/newline between target statements */
{
    if (cflag && b_comment[0] != 0) {
        lineup ();
        fprintf (out, "%c-%s", CCHAR2, b_comment);
    }
    cflag = False;
}


void COMMENT (const char *s)		/* generate comment */
{
    lineup ();
    fprintf (out, "%c+ %s", CCHAR2, s);
}


/*	Details depend on target	*/
const char *rgname (int r)		/* generate register name */
{
    static const char names[4][4] = /* Evaluation stack */
#if defined(TARGET_CPU_SPARC)
    {"...", "%l0", "%l1", "%l2"};
#elif defined(TARGET_CPU_ALPHA)
    {"...", "$1", "$2", "$3"};
#elif defined(TARGET_CPU_ARM)
    {"...", "r4", "r5", "r6"};
#endif

    if (r < 1)
    {
        fputs ("\n*** Integer stack underflow", stderr);
        error ();
    }
/* else	*/
    return names[r];
}


void mnem (const char *op)			/* generate target mnemonic */
{
    posn = 7 + fprintf (out, "\t%s", op);
    comma = 0;
}


void dummy (const char *op)			/* dummy instruction */
{
    posn = 6 + fprintf (out, "%c\t%s", DCCHAR, op);
    comma = 0;
}


void sep (void)			/* tab or comma before operand */
{
    if (comma == 0) {
        putc ('\t', out);
        posn |= 7;
        ++comma;
    } else
        putc (',', out);
    ++posn;
}


void dec (int op)			/* generate decimal constant operand */
{
    sep ();
    posn += fprintf (out, "%d", op);
}


void immdec (int op)			/* generate immediate decimal constant operand */
{
    sep ();
    posn += 1 + fprintf (out, "#%d", op);
}

void ldrcon (int n)			/* generate immediate decimal constant operand */
{
    sep ();
    posn += 1 + fprintf (out, "=%d", n);
}

void ldrequ (char *s)			/* generate immediate string constant operand */
{
    sep ();
    posn += 1 + fprintf (out, "=%s", s);
}


void hex (int op)			/* generate hexadecimal constant operand */
{
    sep ();
    posn += fprintf (out, "0x%x", op);
}


void str (const char *op)			/* generate string operand */
{
    sep ();
    posn += fprintf (out, "%s", op);
}


void qstr (const char *op)			/* generate quoted string operand */
{
    sep ();
    posn += fprintf (out, "\"%s\"", op);
}

#if defined(TARGET_CPU_ARM)
void barstr (const char *op)			/* generate quoted string operand */
{
    sep ();
    posn += fprintf (out, "|%s|", op);
}
#endif

void quotedstr (const char *op)			/* generate string between quotes */
{
    posn += fprintf (out, " \"%s\"", op);
}


/*	Do we need both?	*/
#if defined(TARGET_CPU_SPARC)
void sched (const char *op)			/* generate scheduler entry point */
{
    sep ();
    posn += fprintf (out, "%s%s", SCHEDNAME, op);
}
#endif


#if defined(TARGET_CPU_ALPHA) || defined(TARGET_CPU_ARM)
void tline (const char *op)			/* generate text line */
{
    posn += fprintf (out, "%s\n", op);
}
#endif


#if defined(TARGET_CPU_ALPHA)
void offset0 (int c)			/* byte offset (no scaling) */
{
    if (c >= 0)
        posn += fprintf (out, "%1d", c);
    else if (c < 0)
        posn += fprintf (out, "-%1d", -c);
    /* changed from: nothing if c == 0 */
}
#endif


void offset (int c)			/* word offset */
{
#if defined(TARGET_CPU_ALPHA) || defined(TARGET_CPU_ARM)
    if (c >= 0)
        posn += fprintf (out, "%1d", c*BytesPerWord);
#elif defined(TARGET_CPU_SPARC)
    if (c > 0)
        posn += fprintf (out, "+%1d", c*BytesPerWord);
#endif
    else if (c < 0)
        posn += fprintf (out, "-%1d", -c*BytesPerWord);
}

#if defined(TARGET_CPU_ARM)
void shftop1 (const char *op, int c)		/* generate shifter operand, const */
{
    sep ();
    posn += fprintf (out, "%s #%d", op, c);
}
void shftop2 (const char *op, const char *r)  /* generate shifter operand, reg */
{
    sep ();
    posn += fprintf (out, "%s %s", op, r);
}
#endif


#if defined(TARGET_CPU_ALPHA)
void index0 (const char *r, int c)	/* generate unscaled index operand */
{
    sep ();
    offset0 (c);
    posn += fprintf (out, "(%s)", r);
}
#endif


#if defined(TARGET_CPU_ALPHA)
void index1 (const char *r, int c)		/* generate scaled index operand */
{
    sep ();
    offset (c);
    posn += fprintf (out, "(%s)", r);
}
#endif

#if defined(TARGET_CPU_ARM)
void index1 (const char *r, int c)		/* generate scaled index operand */
{
    sep ();
    if (c == 0)
    {
        posn += fprintf (out, "[%s]", r);
    }
    else
    {
        posn += fprintf (out, "[%s, #", r);
        offset (c);
        fputc (']', out);
    }
    ++posn;
}
#endif


#if defined(TARGET_CPU_SPARC)
void sindex (const char *r, int c)		/* generate scaled index operand */
{
    sep ();
    posn += fprintf (out, "[%s", r);
    offset (c);
    fputc (']', out);
    ++posn;
}
#endif


void index2 (const char *r1, const char *r2)		/* generate double index operand */
{
#if defined(TARGET_CPU_SPARC)
    sep ();
    posn += fprintf (out, "[%s+%s]", r1, r2);
#elif defined(TARGET_CPU_ARM)
    sep ();
    posn += fprintf (out, "[%s, %s]", r1, r2);
#elif defined(TARGET_CPU_ALPHA)
    unimp ();
#endif
}


#if defined(TARGET_CPU_ALPHA)
void indirect (const char *r)		/* generate indirect operand */
{
    sep ();
    posn += fprintf (out, "(%s)", r);
}
#endif


#if defined(TARGET_CPU_ALPHA)
void indirect0 (const char *r)		/* generate indirect operand with explicit zero offset */
{
    sep ();
    posn += fprintf (out, "0(%s)", r);
}
#endif


/*	Not used for SPARC	*/
void label (void)			/* plant label "LX<lcount>" */
{
    fprintf (out, "LX%d:", lcount++);
}


/*	Not used for SPARC	*/
void LAB (int n)			/* generate label "LX<lcount + n>" */
{
    sep ();
    posn += fprintf (out, "LX%d", lcount + n);
}


/*	Not used.  (Counting bytes is a bad idea)	*/
void here (int c)			/* relative to current location */
{
    str(HERE); offset(c);
}


/*	Local numerical labels	*/
void lab (int n)			/* plant label "<n>" */
{
    fprintf (out, "%d", n); nl;
}


void f (int n)			/* generate label "<n>f" */
{
    sep ();
    posn += fprintf (out, "%%F%d", n);
}


void b (int n)			/* generate label "<n>b" */
{
    sep ();
    posn += fprintf (out, "%%B%d", n);
}


void equate (const char *s)	/* equate symbol to value */
{
    posn = fprintf (out, s);
    if (posn >= 8) {
        fputc (' ', out);
        posn += 1;
    } else {
        fputc ('\t', out);
        posn = 8;
    }

#if defined(TARGET_CPU_ARM)
    fputs ("EQU", out);
#elif defined(TARGET_CPU_SPARC) || defined(TARGET_CPU_ALPHA)
    fputc ('=', out);
#endif
    posn += 1;
    comma = False;
}


#if defined(TARGET_CPU_SPARC)
/*	External functions called from translated occam		*/
/*	Must be in the same order as jump table in scheduler	*/

const char *extname (int op)
{
    typedef struct {
        const char *name;
        const char *prefix;
    } en_type;

    const en_type extnames[X_SIZE] = {
        {"fpconst",	  SCHEDNAME},
        {"outbyte",	  SCHEDNAME},
        {"outword",	  SCHEDNAME},
        {"out",	  SCHEDNAME},
        {"in",	  SCHEDNAME},
        {"move",	  SCHEDNAME},
        {"stopp",	  SCHEDNAME},
        {"scheduler", SCHEDNAME},
        {"resched",	  SCHEDNAME},
        {"tin",       SCHEDNAME},
        {"taltwt",	  SCHEDNAME},
        {"dist",	  SCHEDNAME},
        {"vin",       SCHEDNAME},
        {"vout",	  SCHEDNAME},
    
        {"dump",	  SCHEDNAME},
        {"fdump",	  SCHEDNAME},
    
/*  {"error",	  SCHEDNAME}, */
    
        {"mul",       "."},
        {"div",       "."},
        {"rem",       "."},
        {"umul",	  "."},
    
        {"rtime",	  CPREFIX},
        {"utime",	  CPREFIX},
        {"remainder", CPREFIX}
    };
  
    static char ename_buffer[20];	/* bigger than "$$scheduler" */


    if (op < 0 || op >= X_SIZE) {
        fprintf (stderr, "\n*** Unknown external entry %d", op);
        error ();
    }
    sprintf (ename_buffer, "%s%s", extnames[op].prefix, extnames[op].name);
    return (const char*)ename_buffer;
}


void extcall (int op)
{
    if (o_kjumptable) {
        mnem("ld"); sindex(Vptr, op); T0; COMMENT(extname(op)); nl;
        mnem("call"); T0;
    } else {
        mnem("call"); str(extname(op));
    }
}
#endif


void push (void)			/* push stack */
{
    if (istackp == 3) {
        if (ishuff == 0) {	/* unexpected */
            fputs ("\n*** Integer stack overflow", stderr);
            /*	error ();	*/
            where ();
        }
        else
        {
            /* expected */
            if (o_verbose)
            {
                fputs ("\n... Adjusting stack", stderr);
                where ();
            }
#if defined(TARGET_CPU_ARM)
            mnem("mov"); C; B; COMMENT("shuffle"); nl;
            mnem("mov"); B; A; nl;
#elif defined(TARGET_CPU_SPARC) || defined(TARGET_CPU_ALPHA)
            mnem("mov"); B; C; COMMENT("shuffle"); nl;
            mnem("mov"); A; B; nl;
#endif
            shufflecount += 1;
            if (ishuff > 0)
                ishuff -= 1;
        }
    } else
        ++istackp;

/*  shufflecount += shuffle (istackp, istackp + 1); */
}


void pop (int n)			/* pop n words off stack */
{
    istackp -= n;

    if (ishuff > istackp) {
        ishuff = istackp;
        if (o_verbose && ishuff == 0) {
            fputs ("\n... Shuffling off", stderr);
            where ();
        }
    }
}


void printglob (void)		/* print PROC/FUNCTION names as globals */
{
    int i, j;
#if defined(TARGET_CPU_SPARC)
    char proc[PROC_SIZE + 2];
#endif

    for (i = stcount; i > 0; --i) {		/* remove repeats */
        if (proctab[i].proc[0] != 0) {
            for (j = 0; j < i; ++j) {
                if (strcmp (proctab[j].proc, proctab[i].proc) == 0) {
                    if (o_verbose)
                        fprintf (stderr, "\n... Duplicate name %s %s",
                                 proctab[j].label, proctab[j].proc);
                    proctab[j].proc[0] = 0;
                }
            }
        }
    }

    for (i = 0; i <= stcount; ++i) {
        if (proctab[i].proc[0] != 0 && proctab[i].S_PTYPE < P_EXTERNAL) {
#if defined(TARGET_CPU_SPARC)
            sprintf (proc, "$%s", proctab[i].proc);
            mnem(".global"); str(proc); nl;
            equate(proc); str(proctab[i].label); nl;
#elif defined(TARGET_CPU_ALPHA)
            mnem(".globl"); str(proctab[i].proc); nl;
            equate(proctab[i].proc); str(proctab[i].label); nl;
#elif defined(TARGET_CPU_ARM)
            mnem("EXTERN"); barstr(proctab[i].proc); nl;
            equate(proctab[i].proc); str(proctab[i].label); nl;
#endif
        }
    }
}


void entry (void)			/* mark (global) entry point */
{
#if defined(TARGET_CPU_SPARC)
    mnem(".global"); sched("entry"); COMMENT("entry point"); nl;
    fprintf (out, "%sentry:", SCHEDNAME); nl;
#elif defined(TARGET_CPU_ALPHA)
    mnem(".globl"); str("$entry"); COMMENT("entry point"); nl;
    fputs ("$entry:", out); nl;
#elif defined(TARGET_CPU_ARM)
    mnem("EXPORT"); barstr("$entry"); COMMENT("entry point"); nl;
    fputs ("|$entry|", out); nl;
#endif
}


void debug (void)			/* plant debugging directives */
{
#if !defined(TARGET_CPU_ARM)
    if (strcmp (filename, lastname) != 0) {
        mnem(".stabs"); qstr(filename); dec(132); dec(0); dec(0); here(0); nl;
    }
    mnem(".stabn"); dec(68); dec(0); dec(fline); here(0); nl;
#endif
}


void debugp (const char *name)		/* debugging PROC/FUNCTIONs */
{
    mnem(".stabs"); qstr(name); dec(36); dec(0); dec(fline); here(0); nl;
}


#define DT_SIZE		50

static void date (void)			/* date/time stamp */
{
    time_t now;
    char dt_buff[DT_SIZE];
    now = time (NULL);
    (void) strftime (dt_buff, DT_SIZE, "%Y/%B/%d at %H:%M:%S", localtime (&now));
#if defined(TARGET_CPU_SPARC) || defined(TARGET_CPU_ARM)
    fprintf (out, "%c%s", CCHAR, dt_buff); nl;
#elif defined(TARGET_CPU_ALPHA)
    dummy(dt_buff);
#endif
}


extern char kroc_version_message[];     /* In version.c */

void initial (void)			/* initial stuff */
{
#if defined(TARGET_CPU_ALPHA)
    fprintf (out, "\n");		/* Required, bizarrely */
#endif

    dummy(kroc_version_message); nl;
    fprintf (out, "%c\tSource file \"%s\" translated ", CCHAR, source);
    dummy(dest); nl;
    date(); nl; nl;

#if defined(TARGET_CPU_SPARC)
    equate("case_scale"); dec(BytesPerWord); COMMENT("just in case"); nl; nl;
    mnem(".text"); nl;
    mnem(".align"); dec(BytesPerWord); nl; nl;
    
#elif defined(TARGET_CPU_ARM)
    mnem("CODE32"); nl; nl; nl;

    mnem("; Macro to  load a constant known not to fit in an immediate"); nl;


    mnem("MACRO"); nl;
    mnem("ldimm"); str("$reg"); str("$value"); nl; nl;
    mnem("IF ($value :AND: 0xFFFF0000) = 0"); COMMENT("Load 16 bit v+ve value"); nl;
    mnem("  LCLA"); str("temp0"); nl;
    mnem("  LCLA"); str("temp1"); nl;
    fprintf (out, "temp0\t  SETA ($value :AND: 0xFF)"); nl;
    fprintf (out, "temp1\t  SETA ($value :AND: 0xFF00)"); nl;
    mnem("  mov"); str("$reg"); str("temp0"); nl;
    mnem("  orr"); str("$reg"); str("$reg"); str("temp1"); nl;
    mnem("ELIF ($value :AND: 0xFFFF0000) = 0xFFFF0000"); COMMENT("Load 16 bit -ve value"); nl; nl;
    fprintf (out, "temp0\t  SETA ($value :OR: 0xFF00)"); nl;
    fprintf (out, "temp1\t  SETA (:NOT: $value) :AND: 0xFF00"); nl;
    mnem("  mov"); str("$reg"); str("temp0"); COMMENT("really mvn"); nl;
    mnem("  eor"); str("$reg"); str("$reg"); str("temp1"); nl; nl;
    mnem("ELSE"); COMMENT("Load 'hard' $value: > 16 bits needed"); nl; nl;
    mnem("  ldr"); str("$reg"); str("=value"); nl;
    mnem("ENDIF"); nl; nl;
    mnem("MEND"); nl;

    nl; nl;

    mnem("AREA"); str("|Occam$$Code|,CODE,READONLY"); nl;
    mnem("IMPORT"); str("||Lib$$Request$$armlib||, WEAK"); nl;
    if (!o_lib) {
        mnem("EXPORT"); str("entry_user"); nl;
        tline("entry_user");
    }
    
#elif defined(TARGET_CPU_ALPHA)
    equate(" \tcase_scale"); dec(4); COMMENT("just in case"); nl; nl;

    tline(".text");
    tline(".align 3");
    if (!o_lib) {
        tline(".globl $entry_user");
        tline("$entry_user:");
    }
#endif
}


void final (void)			/* final stuff */
{
#if defined(TARGET_CPU_SPARC)
    int c_ws;

    if (!o_lib) {
        c_ws = ((ws	/* workspace - seems to include 3(?!) words for entry stack frame */
                 + vs	/* vector space */
                 + 6	/* entry stack frame - 2 would probably do */
                 + 8	/* locals */
                 + 24)	/* system (rounded up) */
                *4	/* bytes */
                + 7) & ~7; /* rounded up to double-word alignment ... */
        mnem(".global"); sched("space"); nl;
        equate("$$space"); dec(-c_ws); nl;	/* ... and negated */
        mnem(".global"); sched("fsr"); nl;
        equate("$$fsr"); hex((o_checks == CH_TRAP) ? TEM : 0); nl;
    }

    fprintf (out, "\n%c\t%d input lines\n", CCHAR, iline);
    fprintf (out, "%c\t%d output lines\n", CCHAR, oline);
    fprintf (out, "%c\tWorkspace = %d bytes\n", CCHAR, ws*4);
    if (vs != 0)
        fprintf (out, "%c\tVector space = %d bytes\n", CCHAR, vs*4);

    if (shufflecount != 0) {
        fprintf (out, "%c\t%d stack shuffle", CCHAR, shufflecount);
        if (shufflecount > 1)
            fputc ('s', out);
        fputc ('\n', out);
    }
  
    if (lshufflecount != 0) {
        fprintf (out, "%c\t%d label stack shuffle", CCHAR, lshufflecount);
        if (lshufflecount > 1)
            fputc ('s', out);
        fputc ('\n', out);
    }
  
    if (!o_lib) {
        if (b_entry[0] != 0)
            fprintf (out, "%c\tEntry point = %s\n", CCHAR, b_entry);
        else {
            fputs ("\n... Incomplete program - no entry point\n", stderr);
            exit (1);	/* so shell file won't run it */
        }
    }

#elif defined(TARGET_CPU_ALPHA)
    if (!o_lib) {
        equate("WSSIZE"); dec(ws+6); nl;     /* workspace + entry stack frame */
        equate("VSSIZE"); dec(vs); nl;
        tline(".globl $vsbytes");
        tline(".globl $wsbytes");
        tline(".globl $vectorspace");
        tline("$vsbytes: .long 4*VSSIZE");
        if (o_stackws) {
            tline("$wsbytes: .long 4*WSSIZE");
            fprintf (out, "$vectorspace: .long 0 %c- %s\n", CCHAR2, "dummy to satisfy linker");
        } else {
            tline("$wsbytes: .long 0");
            tline(".data");
            mnem(".space"); dec(4*(ws+6)); nl;
            tline(".globl $vectorspace");
            tline("$vectorspace:");
            mnem(".space"); dec(4*vs); nl;
        }
    }
#elif defined(TARGET_CPU_ARM)
    int c_ws;

    if (!o_lib) {
        c_ws = ((ws	/* workspace - includes 3 words for entry stack frame */
                 + vs	/* vector space */
                 + 6	/* entry stack frame - 2 would probably do */
                 + 8	/* locals */
                 + 24)	/* system (rounded up) */
                *4	/* bytes */
                + 7) & ~7; /* rounded up to double-word alignment ... */
/*         mnem("EXTERN"); str("space"); nl; */
/*         equate("space"); dec(c_ws); nl; */
    }

    fprintf (out, "\n%c\t%d input lines\n", CCHAR, iline);
    fprintf (out, "%c\t%d output lines\n", CCHAR, oline);
    fprintf (out, "%c\tWorkspace = %d bytes\n", CCHAR, ws*4);
    if (vs != 0)
        fprintf (out, "%c\tVector space = %d bytes\n", CCHAR, vs*4);

    if (shufflecount != 0) {
        fprintf (out, "%c\t%d stack shuffle", CCHAR, shufflecount);
        if (shufflecount > 1)
            fputc ('s', out);
        fputc ('\n', out);
    }
  
    if (lshufflecount != 0) {
        fprintf (out, "%c\t%d label stack shuffle", CCHAR, lshufflecount);
        if (lshufflecount > 1)
            fputc ('s', out);
        fputc ('\n', out);
    }
  
    fprintf (out, "\n\tEND\n");
    if (!o_lib) {
        if (b_entry[0] != 0)
        {
            fprintf (out, "%c\tEntry point = %s\n", CCHAR, b_entry);
        }
        else {
            fputs ("\n... Incomplete program - no entry point\n", stderr);
            exit (1);	/* so shell file won't run it */
        }
    }

#endif
}


void align (void)			/* align to address boundary */
{
#if defined(TARGET_CPU_SPARC)
    /* Align to (double) word address */
    mnem(".align"); dec(o_real64 ? BytesPerWord*2 : BytesPerWord); nl;
#elif defined(TARGET_CPU_ALPHA)
    /* Align to lword address */
    mnem(".align"); dec(3); nl;
#elif defined(TARGET_CPU_ARM)
    /* Align to (l)word address */
    mnem("ALIGN"); dec(o_real64 ? BytesPerWord*2 : BytesPerWord); nl;
#endif
}


/*	SPARC only		*/
void mnema (const char *op)			/* annulled branch */
{
    mnem(op);
    fputs (",a", out);
    posn += 2;
}


void delay (const char *s)			/* delay slot comment */
{
    lineup ();
    fprintf (out, "%c* %s", CCHAR, s);
}


void annul (const char *s)			/* annul slot comment */
{
    lineup ();
    fprintf (out, "%c@ %s", CCHAR, s);
}


/*	Depends on target assembler	*/
const char map (char ch)		/* map characters in names */
{
    if (isalnum (ch))
        return (ch);
    else if (ch == '$')
        return (ch);
    else if (ch == '.')
        return ('_');
    else if (ch == '-')		/* kludge for expressions in operand field */
        return (ch);
    else if (ch == '%' && !o_trunc)
        return ('$');
    else
        return (0);
}


void LCOMMENT (void)
{
    if (l_comment[0] != 0) {
        lineup ();
        fprintf (out, "%c* %s", CCHAR, l_comment);
    }
}


void defldc (void)			/* do deferred "ldc" */
{
    mnem("set"); str(l_operand); A; LCOMMENT(); nl;
    l_operand[0] = 0;
}


#if defined(TARGET_CPU_SPARC)
void vindex (const char *op)	/* symbolic index relative to Vptr */
{
    sep ();
    posn += fprintf (out, "[%s+%s%s]", Vptr, SCHEDNAME, op);
}
#endif
