#ifdef CSPLIB
/* CSPlib's entry point.  Can optionally initialise the KRoC I/O model
 * with keyboard, screen and error channels corresponding to stdin/out/err.
 * This allow's I/O to occur without blocking
 */

#include "sched_consts.h"
#include "sched_types.h"
#include "sched.h"
#include <stdio.h>
#include "timer.h"
#include "queue_defs.h"

#include "csplib.h"

#ifdef GSL_SMP
  #include <pthread.h>
#endif

#ifdef CSPLIB_CHANNEL_IO
  #include "occam_io.h"
  extern void _occam_start(); /* occam code entry point */
  extern word _wsbytes; /* occam code size */
  extern void *kernel(); /* set up jump table */

  extern FILE *kroc_out;
  extern FILE *kroc_err;
#endif

#ifdef GSL_SMP
void RunPrimary(void *p) {
  fprintf(stderr, "RunPrimary: Primary thread started.\n");
  RunMain((Process *)p);
}

void RunSecondary(void *vp) {
  Process *p = (Process *)vp;
  fprintf(stderr, "RunSecondary: Secondary thread started.\n");
  ProcStop();
}
#endif

#warning Using C entry point code
int c_enter () {
  Process *p = ProcAlloc(NULL, 0, 0);
  #ifdef PRIORITIES
  int i;
  #endif
  #ifdef GSL_SMP
  Process *s = ProcAlloc(NULL, 0, 0);
  pthread_t th_a;
  pthread_t th_b;
  int retcode;
  void *retval;
  #endif
  #ifdef CSPLIB_CHANNEL_IO 
    void *hard_print = kernel(); 
    /* returns pointer to hard print code to pass to */
    /* init_occam_io - need to change this... */
    init_occam_io(hard_print);
    sf.combined_sync = 0;
    #ifdef PRIORITIES
      for (i=0; i<NUM_PRIORITIES; i++) {
        runq.Fptr[i] = runq.Bptr[i] = NotProcess_p;
      }
      Tptr = NotProcess_p;
      runq.priority_table = (word)0;
    #else
    Fptr = Bptr = Tptr = NotProcess_p;
    #endif
    STATUS = 0;
    scr_workspace[Device] = (word)kroc_out;
    err_workspace[Device] = (word)kroc_err;
    #ifdef CPU_TIMERS
      Time_Init();
    #endif
    #ifdef GSL_SMP
      #warning using SMP code
      retcode = pthread_create(&th_a, NULL, RunPrimary, (void *)(p));
      if (retcode != 0) {
        fprintf(stderr, "c_entry: create primary thread failure code %d.\n",
                         retcode);
      } else {
        fprintf(stderr, "c_entry: primary thread creation successful.\n");
      }
      retcode = pthread_create(&th_b, NULL, RunSecondary, (void *)(s));
      if (retcode != 0) {
        fprintf(stderr, "c_entry: create secondary thread failure code %d.\n",
                        retcode);
      } else {
        fprintf(stderr, "c_entry: secondary thread creation successful.\n");
      }
      retcode = pthread_join(th_a, &retval);
      if (retcode != 0) {
        fprintf(stderr, "c_entry: primary thread join failure code %d.\n");
      }
      retcode = pthread_join(th_b, &retval);
      if (retcode != 0) {
        fprintf(stderr, "c_entry: secondary thread join failure code %d.\n");
      }
    #else
      RunMain(p);
    #endif
    STATUS &= 1;
    return STATUS;
  #else
    #ifdef PRIORITIES
      for (i=0; i<NUM_PRIORITIES; i++) {
        Fptr[i] = Bptr[i] = NotProcess_p;
      }
      Tptr = NotProcess_p;
      priority_table = (word)0;
    #else
      Fptr = Bptr = Tptr = NotProcess_p;
    #endif
    sf.combined_sync = 0;
    STATUS = 0;
    #ifdef CPU_TIMERS
      Time_Init();
    #endif
    #ifdef GSL_SMP
      #warning using SMP code
      retcode = pthread_create(&th_a, NULL, RunPrimary, (void *)p);
      if (retcode != 0) {
        fprintf(stderr, "c_entry: create primary thread failure code %d.\n",
                         retcode);
      } else {
        fprintf(stderr, "c_entry: primary thread creation successful.\n");
      }
      retcode = pthread_create(&th_b, NULL, RunSecondary, (void *)s);
      if (retcode != 0) {
        fprintf(stderr, "c_entry: create secondary thread failure code %d.\n",
                        retcode);
      } else {
        fprintf(stderr, "c_entry: secondary thread creation successful.\n");
      }
      retcode = pthread_join(th_a, &retval);
      if (retcode != 0) {
        fprintf(stderr, "c_entry: primary thread join failure code %d.\n");
      }
      retcode = pthread_join(th_b, &retval);
      if (retcode != 0) {
        fprintf(stderr, "c_entry: secondary thread join failure code %d.\n");
      }
    #else
      RunMain(p);
    #endif
    return STATUS;
  #endif
}
#endif
