/* {{{  */

/* Transputer style kernel written in mixed C and assembler */
/* Copyright (C) 1998 Jim Moores */

/* }}} */

/* {{{  includes*/

#include <stdio.h>

#include "sched_types.h"

#include "sched_consts.h"
#include "sched_asm_inserts.h"

#include "occam_io.h"

#include "timer.h"

#include "csplib_consts.h"
#include "csplib_types.h"

#ifndef RTS_H
#define RTS_H
#include "rts.h"
#endif

#ifdef EXTERNAL_COMMS
#include "external.h"
#include "k_extern_macros.h"
#include "k_usr_extern.h"
#endif

/* }}} */

typedef void (*FUNCPTR)();

#include "debug.h"

/* {{{  variables for global spin lock SMP support*/
#ifdef GSL_SMP
word global_spin_lock = 0;
word wait_lock = 0;
#endif
/* }}} */
/* PUBLIC variables that are used elsewhere in the run time system. */
/* {{{  variables for priority*/
#ifdef PRIORITIES
int priority_level;
int priority; /* should just have one or the other ... */
int current_max_priority; /* Shared variable - should not be saved before
                           * calling the external comms. subsystem.
                           */
#endif
/* }}} */
/* this has to go here, because it refers to current_max_priority */
#include "queues.h"

extern T_norm;

/* {{{  global static variables.*/

#if 0
/* actually declared in queue_macros because of inline function there */
pritority_t priority_table;
#endif
sync_type sf;
word STATUS=0;
word* Wptr;
word* Tptr;
/* paramter passing variables */
word** channel_address;
word return_address;
byte* source_address;
byte* destination_address;
word count;
Time wait_time;
word start_address;
word start_offset;
word guard;
word process_address;
word code_offset;
word precondition;
word* usage;
#if 0
word timeout;
#else
Time timeout;
#endif
word result;

/* pseudo private variables (there names are externally visible, but they */
/* should not be used. */
Time now; /* must be int to work (signed) */
word temp1;
word *temp_ptr;
word* other_workspace;
word *current;
word *previous;
word fired;
bool isexternal;

#ifdef CSPLIB
Semaphore* semaphore;
#endif

/* }}} */

void print_tq(void);
void kernel_panic(void);
void *kernel(void);
void X_savel(word *PQptrs);
void X_stlf(word fp);
void X_stlb(word bp);
void X_norm(void);
void X_Seterr(void);
void X_RangeCheckError(void);
void hard_print(void);
void  kbd_process(void);
void X_outbyte(word val, word channel_address);
void X_outword(void);
void X_out(word count, word chaddr, word srcaddr);
void X_release_alt(void);
void X_nb_out(void);
void X_zlnb_in(void);
void X_in(void);
void external_in(void);
void X_pause(void); /* called on loopend */
void X_resched(void);
void X_stopp(void);
void X_scheduler(void);
void X_C_arbitrary_priority_change(void);
void X_C_own_priority_change(void);
void X_C_pri_schedule(void);
void X_C_pri_reschedule(void);
void X_C_pri_startp(void);
void X_C_startp(void);
void X_C_endp(void);
void X_endp(void);
void X_startp(void);
void X_runp(void);
void X_ldtimer(void);
void X_tin(void);
void X_disc(void);
void X_diss(void);
void X_dist(void);
void X_enbs(void);
void X_enbc(void);
void X_enbt(void);
void X_C_claim_semaphore(void);
void X_C_release_semaphore(void);
void X_C_release_all_semaphore(void);

/* {{{  print out timer queue for debugging*/

void print_tq()
{
    word *tqptr = Tptr;
    fprintf(stderr, "Timer Queue dump.\n");
    while ((word)tqptr != NotProcess_p) {
        fprintf(stderr, "(Wptr = 0x%x, Wptr[Time_f] = %d, Wptr[State] = 0x%x)\n", tqptr, tqptr[Time_f], tqptr[State]);
        tqptr = (word*)tqptr[TLink];
    }
    fprintf(stderr, "--- end Timer queue dump.\n");
}

/* }}} */

/* {{{  kernel panic function*/
void kernel_panic()
{
    fprintf(stderr,"Kernel Panic: called kernel function not implemented.\n");
    exit(1);
}
/* }}} */

KernelProcVec procvec =
{
    sizeof(KernelProcVec),
    K_VER, K_REV,
    1, 1, /* occam & c support */
#ifdef EXTERNAL_COMMS
    1,  /* external chan support */
#else
    0,
#endif
#ifdef PRIORITIES
    1,  /* multi-priority support */
#else
    0,
#endif
    0, /* usused */
#ifdef PRIORITIES
    NUM_PRIORITIES,
#else
    2,
#endif

    X_savel,  /* if moved/deleted, change kernel() */
    X_stlf,
    X_stlb,
    X_norm,
    X_Seterr,
    X_RangeCheckError,
    X_outbyte,
    X_outword,
    X_out,
    X_release_alt,
    X_zlnb_in,
    X_nb_out,
    X_in,
    X_pause,
    X_resched,
    X_stopp,
    X_scheduler,
    X_C_arbitrary_priority_change,
    X_C_own_priority_change,
    X_C_pri_schedule,
    X_C_pri_reschedule,
    X_C_pri_startp,
    X_C_startp,
    X_C_endp,
    X_endp,
    X_startp,
    X_runp,
    X_ldtimer,
    X_disc,
    X_diss,
    X_dist,
    X_enbt,
    X_enbc,
    X_enbs,
    X_taltwt,
    X_altwt,
    X_tin,
#ifdef CSPLIB
    X_C_claim_semaphore,
    X_C_release_semaphore,
    X_C_release_all_semaphore,
#else
    X_null,
    X_null,
    X_null,
#endif
};

void *kernel()
{
    word err_stop = (word)kernel_panic;
    occam_table = &procvec.krnSavel;
  
    return 0;
}


/*** MAIN KERNEL CODE ***/
/* Notes
 * The operation var & ~1 is used to mask the bottom bit from the
 * workspace pointer which is used as the priority bit on the transputer.
 * It is not actually used by this kernel but is included so it does not
 * have to be removed from any of the algorithms (which may break them).
 */

/* This function is used by the new synchronisation primitives in occam
 * (BUCKETs, EVENTs etc).
 */
/* {{{  savel */

void X_savel(word *PQptrs)
{
#ifdef PRIORITIES
#error X_savel will not work with priorities enabled.
#else
    PQptrs[0] = (word)Bptr;
    PQptrs[1] = (word)Fptr;
#endif
}

/* }}} */

/* {{{  stlf */

void X_stlf(word fp)
{
#ifndef PRIORITIES
    Fptr = fp;
#endif
    return ;
}

/* }}} */

/* {{{  stlb */

void X_stlb(word bp)
{
#ifndef PRIORITIES
    Bptr = bp;
#endif
    return ;
}

/* }}} */

/* {{{  norm */
void X_norm()
{
/*  The implementation of normalise
 */
    fprintf(stderr, "Norm: not implemented, exiting.\n");
    user_bad_exit();
}
/* }}} */

/* {{{  Seterr */
void X_Seterr()
{        
    fprintf(stderr, "SetErr: Application level error at 0x%x, Wptr = 0x%x.\n",
            return_address, Wptr);
    user_bad_exit();
    return ;
}
/* }}} */

/* {{{  RangeCheckError */
void X_RangeCheckError()
{
    fprintf(stderr, "RangeCheckError: not implemented, exiting.\n");
    user_bad_exit();
    return ;
} 
/* }}} */

/* {{{  outbyte */
/* X_outbyte(val, channel): output a byte (val) to a channel */
void X_outbyte(word val, word channel_address)
{
    source_address = (byte *)&(Wptr[0]);
    count = 1;

    /* {{{  debug*/
    DBGMSG("out: channel address = 0x%x\n" \
           "     count = %d\n" \
           "     source = 0x%x\n" \
           "     return_address = 0x%x\n", \
           channel_address, count, source_address, \
           return_address);
    /* }}} */
#ifdef EXTERNAL_COMMS
    if (IsExternal(channel_address))
    {
        KJUMP(external_out);
    }
#endif
    if (*channel_address == NotProcess_p)
    {
        *channel_address = Wptr;
        Wptr[Pointer] = (word)source_address;
        Wptr[Iptr] = return_address; /* save return address in workspace. */
        KJUMP(X_scheduler);
    }
    else
    {
        temp_ptr = (word *)(((word)(*channel_address)) & (~1));
        DBGMSG("out: other_process[State] = 0x%x", temp_ptr[State]);
        /* mask off priority bit */
        if (temp_ptr[State] <= Ready_p) {
            /* {{{  debug*/
            DBGMSG( "out: channel state is something ALTy.\n");
            /* }}} */
            if (temp_ptr[State] == Waiting_p) {
                QUEUE(temp_ptr);
            }
            temp_ptr[State] = Ready_p;
            *channel_address = Wptr;
            Wptr[Pointer] = (word)source_address;
            Wptr[Iptr] = return_address; /* save return address in workspace. */
            KJUMP(X_scheduler);
        } else {
            *channel_address = (word *)(((word)(*channel_address)) & (~1));
            destination_address = (byte *)((*channel_address)[Pointer]);
            /* copy... */
            /* put first process on back of queue */
            DBGMSG("in: queuing this process and other process.\n");
            /* New code for dealing with priorities (optionally) */
#ifdef PRIORITIES
            /* {{{  check if other process is a higher priority than the current*/
            if (GetPriority(*channel_address) > GetPriority(Wptr)) {
                /* We know because we (Wptr) are running currently, that we
                 * _were_ the highest priority runnable process, but that now
                 * the other process (*channel_address) is, so we deschedule
                 * Wptr and continue as (*channel_address).  There is no need
                 * to queue because we know that (*channel_address)'s run queue
                 * is empty (because Wptr was running before at a lower priority.
                 */
                Wptr[Iptr] = return_address; /* save Iptr for current process */
                QUEUE(Wptr); /* put it on its appropriate run queue */
                Wptr = *channel_address; /* make the other process the current */
                return_address = Wptr[Iptr]; /* get its continuation address */
                /* {{{  do copy and exit.  (copied above)*/
                *channel_address = NotProcess_p; /* same as NULL */
                /* BLOCK COPY */
#ifdef DONT_USE_MEMCPY
                /* {{{  do the copy with a while loop*/
                while (count--) {
                    *(destination_address++) = *(source_address++);
                }
                /* }}} */
#else
                /* {{{  use memcopy (custom version)*/
                MEMCPY(source_address, destination_address, count);
                /* }}} */
#endif
                /* set up ebp = Wptr and esi = occam_table for occam */
                RESTORE;
                return ;
                /* }}} */
            } else {
                QUEUE(*channel_address);
                Wptr[Iptr] = return_address;
                QUEUE(Wptr);
                /* gets a process from Q at current (GetPriority(Wptr)) priority
                 * level.
                 */
                Wptr = GetNextProcess_NonEmptyQ();
                return_address = Wptr[Iptr];
                /* {{{  do copy and exit.  (copied above)*/
                *channel_address = NotProcess_p; /* same as NULL */
                /* BLOCK COPY */
#ifdef DONT_USE_MEMCPY
                /* {{{  do the copy with a while loop*/
                while (count--) {
                    *(destination_address++) = *(source_address++);
                }
                /* }}} */
#else
                /* {{{  use memcopy (custom version)*/
                MEMCPY(source_address, destination_address, count);
                /* }}} */
#endif
                /* set up ebp = Wptr and esi = occam_table for occam */
                RESTORE;
                return ;
                /* }}} */
            }
            /* }}} */
#else
            /* {{{  queue both processes and get a new one from run queue*/
            QUEUE(*channel_address);
            Wptr[Iptr] = return_address; /* save this processes Iptr */
            QUEUE(Wptr);
            Wptr = GetNextProcess_NonEmptyQ();
            return_address = Wptr[Iptr]; /* get new processes old Iptr */
            /* }}} */
#endif
            /* {{{  do copy and exit.  (copied above for optimisation)*/
            *channel_address = NotProcess_p; /* same as NULL */
            /* BLOCK COPY */
#ifdef DONT_USE_MEMCPY
            /* {{{  do the copy with a while loop*/
            while (count--) {
                *(destination_address++) = *(source_address++);
            }
            /* }}} */
#else
            /* {{{  use memcopy (custom version)*/
            MEMCPY(source_address, destination_address, count);
            /* }}} */
#endif
            /* set up ebp = Wptr and esi = occam_table for occam */
            RESTORE;
            return ;
            /* }}} */
        }
    }
}
/* }}} */

/* {{{  outword*/
void X_outword()
{
    LOADWPTR;
    LOADRETURN;
    TWOPARAMS(Wptr[0], channel_address);
    source_address = (byte *)&(Wptr[0]);
    count = 4;
    /* {{{  debug*/
    DBGMSG("out: channel address = 0x%x\n" \
           "     count = %d\n" \
           "     source = 0x%x\n" \
           "     return_address = 0x%x\n", \
           channel_address, count, source_address, \
           return_address);
    /* }}} */
#ifdef EXTERNAL_COMMS
    if (IsExternal(channel_address)) {
        KJUMP(external_out);
    }
#endif
    if (*channel_address == NotProcess_p) {
        *channel_address = Wptr;
        Wptr[Pointer] = (word)source_address;
        Wptr[Iptr] = return_address; /* save return address in workspace. */
        KJUMP(X_scheduler);

    } else {
        temp_ptr = (word *)(((word)(*channel_address)) & (~1));
        DBGMSG("out: other_process[State] = 0x%x", temp_ptr[State]);
        /* mask off priority bit */
        if (temp_ptr[State] <= Ready_p) {
            /* {{{  debug*/
            DBGMSG( "out: channel state is something ALTy.\n");
            /* }}} */
            if (temp_ptr[State] == Waiting_p) {
                QUEUE(temp_ptr);
            }
            temp_ptr[State] = Ready_p;
            *channel_address = Wptr;
            Wptr[Pointer] = (word)source_address;
            Wptr[Iptr] = return_address; /* save return address in workspace. */
            KJUMP(X_scheduler);
        } else {
            *channel_address = (word *)(((word)(*channel_address)) & (~1));
            destination_address = (byte *)((*channel_address)[Pointer]);
            /* copy... */
            /* put first process on back of queue */
            DBGMSG("in: queuing this process and other process.\n");
            /* New code for dealing with priorities (optionally) */
#ifdef PRIORITIES
            /* {{{  check if other process is a higher priority than the current*/
            if (GetPriority(*channel_address) > GetPriority(Wptr)) {
                /* We know because we (Wptr) are running currently, that we
                 * _were_ the highest priority runnable process, but that now
                 * the other process (*channel_address) is, so we deschedule
                 * Wptr and continue as (*channel_address).  There is no need
                 * to queue because we know that (*channel_address)'s run queue
                 * is empty (because Wptr was running before at a lower priority.
                 */
                Wptr[Iptr] = return_address; /* save Iptr for current process */
                QUEUE(Wptr); /* put it on its appropriate run queue */
                Wptr = *channel_address; /* make the other process the current */
                return_address = Wptr[Iptr]; /* get its continuation address */
                /* {{{  do copy and exit.  (copied above)*/
                *channel_address = NotProcess_p; /* same as NULL */
                /* BLOCK COPY */
#ifdef DONT_USE_MEMCPY
                /* {{{  do the copy with a while loop*/
                while (count--) {
                    *(destination_address++) = *(source_address++);
                }
                /* }}} */
#else
                /* {{{  use memcopy (custom version)*/
                MEMCPY(source_address, destination_address, count);
                /* }}} */
#endif
                /* set up ebp = Wptr and esi = occam_table for occam */
                RESTORE;
                return ;
                /* }}} */
            } else {
                QUEUE(*channel_address);
                Wptr[Iptr] = return_address;
                QUEUE(Wptr);
                /* gets a process from Q at current (GetPriority(Wptr)) priority
                 * level.
                 */
                Wptr = GetNextProcess_NonEmptyQ();
                return_address = Wptr[Iptr];
                /* {{{  do copy and exit.  (copied above)*/
                *channel_address = NotProcess_p; /* same as NULL */
                /* BLOCK COPY */
#ifdef DONT_USE_MEMCPY
                /* {{{  do the copy with a while loop*/
                while (count--) {
                    *(destination_address++) = *(source_address++);
                }
                /* }}} */
#else
                /* {{{  use memcopy (custom version)*/
                MEMCPY(source_address, destination_address, count);
                /* }}} */
#endif
                /* set up ebp = Wptr and esi = occam_table for occam */
                RESTORE;
                return ;
                /* }}} */
            }
            /* }}} */
#else
            /* {{{  queue both processes and get a new one from run queue*/
            QUEUE(*channel_address);
            Wptr[Iptr] = return_address; /* save this processes Iptr */
            QUEUE(Wptr);
            Wptr = GetNextProcess_NonEmptyQ();
            return_address = Wptr[Iptr]; /* get new processes old Iptr */
            /* }}} */
#endif

            /* {{{  do copy and exit.  (copied above for optimisation)*/
            *channel_address = NotProcess_p; /* same as NULL */
            /* BLOCK COPY */
#ifdef DONT_USE_MEMCPY
            /* {{{  do the copy with a while loop*/
            while (count--) {
                *(destination_address++) = *(source_address++);
            }
            /* }}} */
#else
            /* {{{  use memcopy (custom version)*/
            MEMCPY(source_address, destination_address, count);
            /* }}} */
#endif
            /* set up ebp = Wptr and esi = occam_table for occam */
            RESTORE;
            return ;
            /* }}} */
        }
    }
}
/* }}} */

/* {{{  out*/
void X_out(word count, word chaddr, word srcaddr)
{
    /* {{{  debug*/
    DBGMSG("out: channel address = 0x%x\n" 
           "     count = %d\n" 
           "     source = 0x%x\n" 
           "     return_address = 0x%x\n", 
           channel_address, count, source_address, 
           return_address);
    /* }}} */
#ifdef EXTERNAL_COMMS
    /* {{{ check for and deal with external comms */

if (IsExternal(channel_address))
{
    /* {{{  debug*/
        DBGMSG("external_out: channel address = 0x%x\n" 
               "              count = %d\n" 
               "              source = 0x%x\n" 
               "              return_address = 0x%x\n", 
               channel_address, count, source_address, 
               return_address);
        DBGMSG("out: external channel detected.\n");
        /* }}} */
    /* {{{  point channel_address at soft channel word*/
#ifdef SLOW
        (word *)channel_address = GetAddrInternal(channel_address);
#else
        ((word *)channel_address)++;
#endif
        /* point channel_address at the soft, internal channel */
        /* Note that if this is an external output then the channel word should
         * be empty.
         */
        /* }}} */
     if (*channel_address == NotProcess_p)
     {
         *channel_address = Wptr;
         Wptr[Pointer] = (word)source_address;
         Wptr[Iptr] = return_address; /* save return address in workspace. */
         /* {{{  Notify the external comms. system of output (External_DoOut)*/
            DBGMSG("out: performing early data copy for exxternal channel.\n");
            /* {{{  point channel_address back at external channel word*/
#ifdef SLOW
#error X_out: possiblilty not dealt with
#else
            /* I will be sent to hell for using pointer arithmetic */
            ((word *)channel_address)--;
#endif
            /* }}} */
            /* Move channel pointer back to top of external channel structure */
            /* {{{  use the source address and count given to out */
            /*
             * copy data before synchronisation, but instead of doing a data
             * copy we just copy the source address and length into the (p,len)
             * pair in the external channel structure.  The external communications
             * system must then actually do any transfer.  This is done so that the
             * data is available to the communications system before we release
             * the outputting process (which happens when the communication is
             * complete.  The data transfer usually happens at the end in a normal
             * transaction on internal channels.
             */
            ((External *)channel_address)->p = source_address;
            ((External *)channel_address)->len = count;
            /* }}} */
            /* {{{  save static variables*/
            /* save necessary static variables so that there is no problem if
             * External_DoOut() calls the kernel back.
             */
            SAVESTATIC(count);
            SAVESTATIC(channel_address);
            SAVESTATIC(source_address);
            SAVESTATIC(Wptr);
            SAVESTATIC(isexternal);
            SAVESTATIC(return_address);
            /* }}} */
            /* Notify the communications system of the output request */
            External_DoOut((External *)channel_address);
            /* {{{  restore static variables*/
            RESTORESTATIC(return_address);
            RESTORESTATIC(isexternal);
            RESTORESTATIC(Wptr);
            RESTORESTATIC(source_address);
            RESTORESTATIC(channel_address);
            RESTORESTATIC(count);
            /* }}} */
            /* }}} */
         KJUMP(X_scheduler);
         
     } else {
         /* {{{  error*/
            fprintf(stderr, "X_out: attempt to output on non-empty external channel."
                    ".\n");
            user_bad_exit();
            /* }}} */
     }
}
else

/* }}} */
#endif
    if (*channel_address == NotProcess_p)
    {
        *channel_address = Wptr;
        Wptr[Pointer] = (word)source_address;
        Wptr[Iptr] = return_address; /* save return address in workspace. */
        KJUMP(X_scheduler);
    }
    else
    {
        temp_ptr = (word *)(((word)(*channel_address)) & (~1));
        DBGMSG("out: other_process[State] = 0x%x", temp_ptr[State]);
        /* mask off priority bit */
        if (temp_ptr[State] <= Ready_p) {
            /* {{{  debug*/
            DBGMSG( "out: channel state is something ALTy.\n");
            /* }}} */
            if (temp_ptr[State] == Waiting_p) {
                QUEUE(temp_ptr);
            }
            temp_ptr[State] = Ready_p;
            *channel_address = Wptr;
            Wptr[Pointer] = (word)source_address;
            Wptr[Iptr] = return_address; /* save return address in workspace. */
            KJUMP(X_scheduler);
        } else {
            *channel_address = (word *)(((word)(*channel_address)) & (~1));
            destination_address = (byte *)((*channel_address)[Pointer]);
            /* copy... */
            /* put first process on back of queue */
            DBGMSG("in: queuing this process and other process.\n");
            /* New code for dealing with priorities (optionally) */
#ifdef PRIORITIES
            /* {{{  check if other process is a higher priority than the current*/
            if (GetPriority(*channel_address) > GetPriority(Wptr)) {
                /* We know because we (Wptr) are running currently, that we
                 * _were_ the highest priority runnable process, but that now
                 * the other process (*channel_address) is, so we deschedule
                 * Wptr and continue as (*channel_address).  There is no need
                 * to queue because we know that (*channel_address)'s run queue
                 * is empty (because Wptr was running before at a lower priority.
                 */
                Wptr[Iptr] = return_address; /* save Iptr for current process */
                QUEUE(Wptr); /* put it on its appropriate run queue */
                Wptr = *channel_address; /* make the other process the current */
                return_address = Wptr[Iptr]; /* get its continuation address */
                /* {{{  do copy and exit.  (copied above)*/
                *channel_address = NotProcess_p; /* same as NULL */
                /* BLOCK COPY */
#ifdef DONT_USE_MEMCPY
                /* {{{  do the copy with a while loop*/
                while (count--) {
                    *(destination_address++) = *(source_address++);
                }
                /* }}} */
#else
                /* {{{  use memcopy (custom version)*/
                MEMCPY(source_address, destination_address, count);
                /* }}} */
#endif
                /* set up ebp = Wptr and esi = occam_table for occam */
                RESTORE;
                return ;
                /* }}} */
            } else {
                QUEUE(*channel_address);
                Wptr[Iptr] = return_address;
                QUEUE(Wptr);
                /* gets a process from Q at current (GetPriority(Wptr)) priority
                 * level.
                 */
                Wptr = GetNextProcess_NonEmptyQ();
                return_address = Wptr[Iptr];
                /* {{{  do copy and exit.  (copied above)*/
                *channel_address = NotProcess_p; /* same as NULL */
                /* BLOCK COPY */
#ifdef DONT_USE_MEMCPY
                /* {{{  do the copy with a while loop*/
                while (count--) {
                    *(destination_address++) = *(source_address++);
                }
                /* }}} */
#else
                /* {{{  use memcopy (custom version)*/
                MEMCPY(source_address, destination_address, count);
                /* }}} */
#endif
                /* set up ebp = Wptr and esi = occam_table for occam */
                RESTORE;
                return ;
                /* }}} */
            }
            /* }}} */
#else
            /* {{{  queue both processes and get a new one from run queue*/
            QUEUE(*channel_address);
            Wptr[Iptr] = return_address; /* save this processes Iptr */
            QUEUE(Wptr);
            Wptr = GetNextProcess_NonEmptyQ();
            return_address = Wptr[Iptr]; /* get new processes old Iptr */
            /* }}} */
#endif

            /* {{{  do copy and exit.  (copied above for optimisation)*/
            *channel_address = NotProcess_p; /* same as NULL */
            /* BLOCK COPY */
#ifdef DONT_USE_MEMCPY
            /* {{{  do the copy with a while loop*/
            while (count--) {
                *(destination_address++) = *(source_address++);
            }
            /* }}} */
#else
            /* {{{  use memcopy (custom version)*/
            MEMCPY(source_address, destination_address, count);
            /* }}} */
#endif
            /* set up ebp = Wptr and esi = occam_table for occam */
            RESTORE;
            return ;
            /* }}} */
        }
    }
}
/* }}} */
#ifdef EXTERNAL_COMMS

/* {{{  release_alt*/

#ifdef EXTERNAL_COMMS
/* This kernel function should release any waiting alt guard - if the guard
 * is selected (it may not be), then the following X_in call will block.
 * Normally once an alt guard has been released the input succeeds (because
 * it is an output that normally unblocks and alt.  We require this behaviour
 * to allow for more efficient external channel buffer management; we don't
 * have to receive the whole message (ie request more than just the first bit)
 * before we can release an alt and get a committed input.  This should
 * reduce network congestion.
 */
void X_release_alt()
{
    LOADWPTR;
    LOADRETURN;
    PARAM(channel_address);

    /* {{{  run time testing if (is an external channel)*/
#ifdef EXTERNAL_RUNTIME_CHECKS
    if (IsExternal(channel_address)) {
#endif
        /* }}} */
        channel_address++; /* point at soft channel */
        if (*channel_address == NULL) {
            *channel_address = Wptr;
            /* so that the Alt can think it should release us */
            Wptr[Pointer] = BlockIn_p;
            /* {{{  Comment*/
            /* when the input reads this it will block and wait to be released by
             * the comms system (it will do what it would normally do if the
             * channel word was empty).
             */
            /* }}} */
            RESTORE;
            return ;
        } else {
            temp_ptr = (word *)(((word)(*channel_address)) & (~1));
            if (temp_ptr[State] <= Ready_p) {
                /* {{{  debug*/
#ifdef DEBUG_KERNEL
                DBGMSG( "out: channel state is something ALTy.\n");
#endif
                /* }}} */
                if (temp_ptr[State] == Waiting_p) {
                    QUEUE(temp_ptr);
                }
                temp_ptr[State] = Ready_p;
                *channel_address = Wptr;
                Wptr[Pointer] = BlockIn_p;
                /* {{{  Comment*/
                /* we don't save an Iptr as we rely on the inputter not trying to
                 * schedule the pseudo process.
                 * It used to be that we'd deschedule here, but as this is called by
                 * the communications system, it isn't a process, so we can't, this
                 * call must not block, or transfer control elsewhere.
                 */
                /* }}} */
                RESTORE;
                return ;
            } /* else do nothing */
        } /* else do nothing */
        /* {{{  run time testing endif (is it an external channel)*/
#ifdef EXTERNAL_RUNTIME_CHECKS
    } else {
        fprintf(stderr, "release_alt: attempt to release_alt an internal channel.\n");
    }
#endif
    /* }}} */
    RESTORE;
    return ;
}
#endif

/* }}} */

/* {{{  nb_out*/
#ifdef EXTERNAL_COMMS
void X_nb_out()
{
    LOADWPTR;
    LOADRETURN;
    THREEPARAMS(count, channel_address, source_address);
    DBGMSG("X_nb_out: channel address = 0x%x\n" \
           "     count = %d\n" \
           "     source = 0x%x\n" \
           "     return_address = 0x%x\n", \
           channel_address, count, source_address, \
           return_address);
    /* {{{  if external channel (only if checks enabled)*/
#if EXTERNAL_RUNTIME_CHECKS
    if (isexternal = IsExternal(channel_address)) {
#endif
        /* }}} */
        /* {{{  */
        /* {{{  Comment*/
        /* If internal channels points at pseudo address workspace then one
         * of two things could have happened.
         *  1) The channel was release_alt'ed in which case the the pseudo
         *     workspace's Pointer field will contain BlockIn_p
         *  2) The channel was nb_out'ed twice in a row without the data
         *     being accepted (in which case the Pointer field should point
         *     somewhere else.
         * because of the distinction we detect the second condition as an error
         * while the first condition means we need to treat the channel word
         * as empty.
         * This error detection is included really to aid debugging of new
         * external communications packages.
         */
        /* }}} */
        /* {{{  if attempt to renew data before channel accepted then error (opt)*/
#ifdef EXTERNAL_RUNTIME_CHECKS
        if (
            (GetInternal(channel_address) ==
             (word)GetAddrWorkspace(channel_address))
            &&
            (GetAddrWorkspace(channel_address)[Pointer] != BlockIn_p)
            ) {
            /* check that channel not 'released' by comms sys before, and data
             * has not yet been accepted.
             */
            fprintf(stderr, "nb_out: attempt to renew channel data before accepted"
                    ".\n");
            user_bad_exit();
        }
#endif
        /* }}} */
#ifdef SLOW
        (word *)channel_address = GetAddrInternal(channel_address);
#else
        ((word *)channel_address)++;
#endif
        /* point channel_address at the soft, internal channel */
      
        /* }}} */
        /* {{{  else error*/
#ifdef EXTERNAL_RUNTIME_CHECKS
    } else {
        fprintf(stderr, "nb_out: attempt to nb_out a soft channel.\n");
        user_bad_exit();
    }
#endif
    /* }}} */
    /* Note that if this is an external output then the channel word should
     * be empty.
     */
    /* Always an internal, soft channel */
    /* if it points to pseudo workspace then it was release_alt'ed or nb_out'ed
     * twice.  We detected multiple nb_out's above, so it must be that the
     * channel was release_alt'ed.
     */
    if ( (*channel_address == NotProcess_p)
         ||
         (
             (*channel_address != NotProcess_p)
             &&
             (*channel_address ==
              GetAddrWorkspace((External *)((word *)channel_address-1)))
             )
         ) {
        /* {{{  Comment*/
        /* This happens if some data has arrived, but not yet been requested by
         * a process
         */
        /* }}} */
        /* {{{  debug*/
        DBGMSG("nb_out: Attempt by communications system to release an" \
               " outputting process on\n""an empty channel.\n");
        /* }}} */
        *channel_address = Wptr;
        Wptr[Pointer] = (word)source_address;
        /* {{{  Comment*/
        /* we don't save an Iptr as we rely on the inputter not trying to
         * schedule the pseudo process.
         * It used to be that we'd deschedule here, but as this is called by
         * the communications system, it isn't a process, so we can't, this
         * call must not block, or transfer control elsewhere.
         */
        /* }}} */
        RESTORE;
        return ;

    } else {
        /* {{{  debug*/
        DBGMSG("nb_out: channel word not empty.\n");
        /* }}} */
        temp_ptr = (word *)(((word)(*channel_address)) & (~1));
        /* {{{  debug*/
        DBGMSG("nb_out: other_process[State] = 0x%x\n", temp_ptr[State]);
        /* }}} */
        /* mask off priority bit */
        if (temp_ptr[State] <= Ready_p) {
            /* {{{  debug*/
            DBGMSG( "out: channel state is something ALTy.\n");
            /* }}} */
            if (temp_ptr[State] == Waiting_p) {
                QUEUE(temp_ptr);
            }
            temp_ptr[State] = Ready_p;
            *channel_address = Wptr;
            Wptr[Pointer] = (word)source_address;
            /* {{{  Comment*/
            /* we don't save an Iptr as we rely on the inputter not trying to
             * schedule the pseudo process.
             * It used to be that we'd deschedule here, but as this is called by
             * the communications system, it isn't a process, so we can't, this
             * call must not block, or transfer control elsewhere.
             */
            /* }}} */
            RESTORE;
            return ;
        } else {
            *channel_address = (word *)(((word)(*channel_address)) & (~1));
            destination_address = (byte *)((*channel_address)[Pointer]);
            /* {{{  Comment*/
            /* copy... */
            /* put waiting process on back of queue */
            /* Continue as us. */
            /* }}} */
            QUEUE(*channel_address);

            *channel_address = NotProcess_p; /* same as NULL */

            /* {{{  block copy*/
#ifdef DONT_USE_MEMCPY
            /* {{{  do it with a while loop*/
            /* BLOCK COPY */
            while (count--) {
                *(destination_address++) = *(source_address++);
            }
            /* }}} */
#else
            /* {{{  MEMCPY asm macro*/
            MEMCPY(source_address, destination_address, count);
            /* }}} */
#endif
            /* }}} */
            /* save necessary static variables so their values can be restored
             * if DoneIn calls the kernel back.  Here we only need two.
             */
            /* {{{  save statics variables*/
            SAVESTATIC(Wptr);
            SAVESTATIC(return_address);
            /* }}} */
            External_DoneIn((External *)((word *)channel_address-1));
            /* {{{  restore static variables*/
            RESTORESTATIC(return_address);
            RESTORESTATIC(Wptr);
            /* }}} */
            /* set up ebp = Wptr and esi = occam_table for occam */
            RESTORE;
            return ;
        }
    }
}
#endif
/* }}} */

/* {{{  zlnb_in*/
/* This is a zero length, non-blocking versions of X_in.  It is just used
 * to synchronise external communications on the internal, soft channel in
 * the external channel structure.  It is used to allow the communications
 * system to release outputting processes.  The reason we cannot just call
 * X_in with zero length message is that X_in reschedules both processes.
 * On the transputer the control flow continued with the inputting process
 * rather than the rescheduled outputting process.  Also, the data transfer
 * has already happened, it must be done when X_out is called on an external
 * channel so that data is available to the communications system.
 */
void X_zlnb_in()
{
    LOADWPTR;
    LOADRETURN;
    PARAM(channel_address);
    /* {{{  debug*/
#ifdef DEBUG_KERNEL
    DBGMSG("external_in_release: return address = 0x%x\n" \
           "    channel address = 0x%x\n", \
           return_address, channel_address \
           );
#endif
    /* }}} */
#ifdef EXTERNAL_COMMS
     /* {{{  if its an external channel (optional) {*/
#ifdef EXTERNAL_RUNTIME_CHECKS
    if (isexternal = IsExternal(channel_address))
    {
#endif
        /* }}} */
#ifdef SLOW
        (word *)channel_address = GetAddrInternal(channel_address);
#else
        ((word *)channel_address)++;
#endif
        /* point channel_address at the soft, internal channel */
    }
    /* {{{  } else error (optional)*/
#ifdef EXTERNAL_RUNTIME_CHECKS
    else
    {
        fprintf(stderr, "zlnb_in: attempt to zlnb_in a soft channel.\n");
        user_bad_exit();
    }
#endif
    /* }}} */
#endif
    /* {{{  if (check channel isn't empty) (optional)*/
#ifdef EXTERNAL_RUNTIME_CHECKS
    if (*channel_address == NotProcess_p)
    {
        /* This should never happen, we should always be releasing a process */
        fprintf(stderr, "X_zlnb_in: invalid use of this call, channel empty.\n");
        user_bad_exit();
        *channel_address = Wptr;
        RESTORE;
        return ;
    }
    else
#endif
        /* }}} */
    {
        QUEUE(*channel_address);
        *channel_address = NotProcess_p; /* used to be done earlier */
        /* and pointer cached in register */
        /* set up ebp = Wptr and esi = occam_table for occam */
        RESTORE;
        return ;
    }
}
/* }}} */
#endif /*  EXTERNAL_COMMS */

/* {{{  in*/
/* Input from an internal or external channel */
void X_in()
{
    LOADWPTR;
    LOADRETURN;
    THREEPARAMS(count, channel_address, destination_address);
    /* {{{  debug*/
    DBGMSG("in: return address = 0x%x\n" \
           "    count = %d\n" \
           "    channel address = 0x%x\n" \
           "    destination address = 0x%x\n", \
           return_address, count, channel_address, \
           destination_address);
    /* }}} */
    /* {{{  if external channel*/
#ifdef EXTERNAL_COMMS
    if (IsExternal(channel_address)) {
        KJUMP(external_in);
    }
#endif
    /* }}} */
    /* {{{  Comment*/
    /* Re: EXTERNAL_COMMS optional code in if statement below.
     * This section of code deals with the case where an ALT guard has
     * been release_alt'ed and we don't want the input to succeed until
     * we recieve a PerformedIn.  We know this because the Pointer field is
     * set to a special value BlockIn_p.
     */
    /* }}} */
    if (*channel_address == NotProcess_p) {
        *channel_address = Wptr;
        Wptr[Pointer] = (word)destination_address;
        Wptr[Iptr] = return_address; /* save return address in workspace. */
        KJUMP(X_scheduler);
    } else {
        source_address = (byte*)(*channel_address)[Pointer];
        /* copy ... */
        /* put first process on back of queue */
        DBGMSG("in: queuing this process and other process.\n");
        /* New code for dealing with priorities (optionally) */
#ifdef PRIORITIES
        /* {{{  check if a higher priority process is available*/
        if (GetPriority(*channel_address) > GetPriority(Wptr)) {
            /* We know because we (Wptr) are running currently, that we
             * _were_ the highest priority runnable process, but that now
             * the other process (*channel_address) is, so we deschedule
             * Wptr and continue as (*channel_address).  There is no need
             * to queue because we know that (*channel_address)'s run queue
             * is empty (because Wptr was running before at a lower priority.
             */
            Wptr[Iptr] = return_address; /* save Iptr for current process */
            QUEUE(Wptr); /* put it on its appropriate run queue */
            Wptr = *channel_address; /* make the other process the current */
            return_address = Wptr[Iptr]; /* get its continuation address */
            /* {{{  copy data and return*/
            /* {{{  Comment*/
            /* BUG FIX - this didn't used to be done here, but in the queue juggling
             * stuff above - I therefore accidentally left it out if the channel was
             * external, which meant that the channel was never reset.
             * NB: need to be careful of this if doing multiprocessor SMP version
             */
            /* }}} */
            *channel_address = NotProcess_p;
            /* BLOCK COPY */
            /* {{{  debug*/
            DBGMSG("in: Source address = 0x%x, Destination address = 0x%x\n", \
                   source_address, destination_address);
            /* }}} */
#ifdef DONT_USE_MEMCPY
            /* {{{  copy using a while loop*/
            while (count) {
                *(destination_address++) = *(source_address++);
                count--;
            }
            /* }}} */
#else
            MEMCPY(source_address, destination_address, count);
#endif
            /* set up ebp = Wptr and esi = occam_table for occam */
            RESTORE;
            return ;
            /* }}} */
        } else {
            QUEUE(*channel_address);
            Wptr[Iptr] = return_address;
            QUEUE(Wptr);
            /* gets a process from Q at current (GetPriority(Wptr)) priority
             * level.
             */
            Wptr = GetNextProcess_NonEmptyQ();
            return_address = Wptr[Iptr];
            /* {{{  copy data and return*/
            /* {{{  Comment*/
            /* BUG FIX - this didn't used to be done here, but in the queue juggling
             * stuff above - I therefore accidentally left it out if the channel was
             * external, which meant that the channel was never reset.
             * NB: need to be careful of this if doing multiprocessor SMP version
             */
            /* }}} */
            *channel_address = NotProcess_p;
            /* BLOCK COPY */
            /* {{{  debug*/
            DBGMSG("in: Source address = 0x%x, Destination address = 0x%x\n", \
                   source_address, destination_address);
            /* }}} */
#ifdef DONT_USE_MEMCPY
            /* {{{  copy using a while loop*/
            while (count) {
                *(destination_address++) = *(source_address++);
                count--;
            }
            /* }}} */
#else
            MEMCPY(source_address, destination_address, count);
#endif
            /* set up ebp = Wptr and esi = occam_table for occam */
            RESTORE;
            return ;
            /* }}} */
        }
        /* }}} */
#else
        /* {{{  queue both processes and get a new one from the front of the run q*/
        QUEUE(*channel_address);
        Wptr[Iptr] = return_address;
        QUEUE(Wptr);
        Wptr = GetNextProcess_NonEmptyQ();
        return_address = Wptr[Iptr];
        /* }}} */
#endif
        /* {{{  copy data and return (rememeber there are copies above)*/
        /* {{{  Comment*/
        /* BUG FIX - this didn't used to be done here, but in the queue juggling
         * stuff above - I therefore accidentally left it out if the channel was
         * external, which meant that the channel was never reset.
         * NB: need to be careful of this if doing multiprocessor SMP version
         */
        /* }}} */
        *channel_address = NotProcess_p;
        /* BLOCK COPY */
        /* {{{  debug*/
        DBGMSG("in: Source address = 0x%x, Destination address = 0x%x\n", \
               source_address, destination_address);
        /* }}} */
#ifdef DONT_USE_MEMCPY
        /* {{{  copy using a while loop*/
        while (count) {
            *(destination_address++) = *(source_address++);
            count--;
        }
        /* }}} */
#else
        MEMCPY(source_address, destination_address, count);
#endif
        /* set up ebp = Wptr and esi = occam_table for occam */
        RESTORE;
        return ;
        /* }}} */
    }
} 
/* }}} */

/* {{{  external_in*/
#ifdef EXTERNAL_COMMS
void external_in()
{             
    /* Input from an external channel */
    /* {{{  debug*/
    DBGMSG("external_in: return address = 0x%x\n" \
           "             count = %d\n" \
           "             channel address = 0x%x\n" \
           "             destination address = 0x%x\n", \
           return_address, count, channel_address, \
           destination_address);
    /* }}} */
    /* {{{  point channel_address at soft channel*/
    /* Inform the communications system of input, note that it _must_ not
     * generate a PerformedIn callback from here, it must wait until it is
     * next polled by the scheduler.   I can't really see why the comms.
     * system needs this call.
     */
    ((word *)channel_address)++;
    /* point it at the soft, internal channel */
    /* }}} */
    /* {{{  Comment*/
    /* Re: EXTERNAL_COMMS optional code in if statement below.
     * This section of code deals with the case where an ALT guard has
     * been release_alt'ed and we don't want the input to succeed until
     * we recieve a PerformedIn.  We know this because the Pointer field is
     * set to a special value BlockIn_p.
     */
    /* }}} */

    if ((*channel_address == NotProcess_p) || ((*channel_address != NotProcess_p)
                                               && ((*channel_address)[Pointer] == BlockIn_p))) {
        *channel_address = Wptr;
        Wptr[Pointer] = (word)destination_address;
        Wptr[Iptr] = return_address; /* save return address in workspace. */
        (word *)channel_address--; /* move pointer back to base of external
                                    * channel structure, nasty but efficient
                                    */
        /* {{{  Comment*/
        /* Save variables that might be corrupted during external call by the
         * called routine calling back the kernel, so the values in the static
         * variables change.
         */
        /* }}} */
        /* {{{  save static variables*/
        SAVESTATIC(count);
        SAVESTATIC(channel_address);
        SAVESTATIC(destination_address);
        SAVESTATIC(Wptr);
        SAVESTATIC(isexternal);
        SAVESTATIC(return_address);
        /* }}} */
        External_DoIn((External *)channel_address);
        /* {{{  restore static variables*/
        RESTORESTATIC(return_address);
        RESTORESTATIC(isexternal);
        RESTORESTATIC(Wptr);
        RESTORESTATIC(destination_address);
        RESTORESTATIC(channel_address);
        RESTORESTATIC(count);
        /* }}} */
        /* {{{  Comment*/
        /* We don't want to say DoneIn here because the input hasn't acutally
         * happened yet (the data transfer that is).  This is done by the
         * External_PerformedIn() function.
         */
        /* }}} */
        /*External_DoneIn((External *)channel_address);*/
        KJUMP(X_scheduler);
    } else {
        source_address = (byte*)(*channel_address)[Pointer];
        /* copy ... */
        *channel_address = NotProcess_p;
        /* BLOCK COPY */
        DBGMSG("in: Source address = 0x%x, Destination address = 0x%x\n", \
               source_address, destination_address);
        while (count) {
            *(destination_address++) = *(source_address++);
            count--;
        }
        /* {{{  inform comms. system. that we've completed the transfer.*/
#ifndef DONT_SYNCRONISE_EXTERNAL_INPUT
        if (isexternal) {
            (word *)channel_address--;
            /* Save necessary variables for external call.  Here we only save
             * what we absolutely have to.
             */
            SAVESTATIC(Wptr);
            SAVESTATIC(return_address);
            External_DoneIn((External *)channel_address);
            RESTORESTATIC(return_address);
            RESTORESTATIC(Wptr);
        }
#endif
        /* }}} */
        /* set up ebp = Wptr and esi = occam_table for occam */
        RESTORE;
        return ;
    }
}
#endif
/* }}} */

/* {{{  pause*/
void X_pause() /* called on loopend */
{
    KJUMP(X_resched);
}
/* }}} */

/* {{{  resched*/
void X_resched()
{
    LOADWPTR;
    LOADRETURN;
    QUEUE(Wptr);
    /* Inlined stopp more efficient */
    Wptr[Iptr] = return_address;
    DBGMSG("X_resched: Wptr = 0x%x.  Entering scheduler.",Wptr);
    /* Jump on */
    KJUMP(X_scheduler);
}
/* }}} */

/* {{{  stopp*/
void X_stopp()
{
    LOADWPTR;
    LOADRETURN;
    Wptr[Iptr] = return_address; /* save return address in workspace. */
    KJUMP(X_scheduler);
}
/* }}} */

/* {{{  scheduler*/
/* This routine cannot be "call'ed" because it doesn't deal with the stack, so
   it should never be called from C directly, only from assembler using a jump.
   It should not be called externally anyway... */
/* FOOTNOTE: But it is jumped to (now!) from CSPlib in ProcPar(). */


void X_scheduler()
{
#ifdef EXTERNAL_COMMS
    /* NB no need to save anything here */
    /*sf.s.extern_flag = (byte)External_Poll(); */
#endif
#ifdef CPU_TIMERS
    /* we don't need this (although it does no harm) if we are using the UNIX
     * timer mechanism
     */
    sf.s.tim_sync = (byte)Time_PollAlarm();
#endif
    DBGMSG("scheduler: called with Wptr = 0x%x.\n", Wptr);
    /* current_max_priority may be changed by a subsequent call to
     * External_HandleExternals() which may reschedule processes of higher
     * priority than the current processes priority.
     */
#ifdef PRIORITIES
    current_max_priority = GetPriority(Wptr);
#endif
    if (sf.combined_sync != 0) {
        DBGMSG("scheduler: one of the sync flags has been set.\n");
        /* one of the sync flags has been set */
        if (sf.s.kbd_sync) {
            DBGMSG( "scheduler: it was the keyboard sync.\n");
            sf.s.kbd_sync = 0;
            goto kbd_process; /* jump could be made internal... */
        } else {
            if (sf.s.tim_sync) {
                sf.s.tim_sync = 0;
                now = Time_GetTime();
                /* possible problem in following while loop in 2nd expr */
                while ((Tptr != NotProcess_p) && (!(Time_AFTER(GetTimeField(Tptr),now)))) {
                    /* move first process on timer queue to the run queue */
                    temp_ptr = Tptr;
                    Tptr = (word *)Tptr[TLink];
                    temp_ptr[TLink] = TimeSet_p;
                    GetTimeField(temp_ptr) = now;
                    if (temp_ptr[State] != Ready_p) {
                        temp_ptr[State] = Ready_p;
                        QUEUE(temp_ptr);
                    }
                }
                if (Tptr != NotProcess_p) {
                    Time_SetAlarm( Time_MINUS(GetTimeField(Tptr), now) );
                } else {
                    Time_ResetAlarm();
                }
            } else {
#ifdef EXTERNAL_COMMS
                if (sf.s.extern_flag) {
                    /* I don't /think/ that there is any need to save any static
                     * variables here.
                     * There will be if the scheduler is rearranged.
                     */
                    External_HandleExternals();
                } else {
#endif
                    kernel_panic();
#ifdef EXTERNAL_COMMS
                }
#endif
            }
        }
    }
    DBGMSG("scheduler: there are processes waiting.\n");
#ifdef PRIORITIES
    /* {{{  debug*/
#if 0 /* def DEBUG_KERNEL*/
    DBGMSG("scheduler: before loop Wptr = 0x%x.\n", Wptr);
    for (temp1=0; temp1<32; temp1++) {
        DBGMSG("scheduler: in loop Wptr = 0x%x.\n", Wptr);
        if (!IsRunQEmpty(temp1)) {
            DBGMSG("Run queue %d is tagged full, Fptr[%d] = 0x%x.  " \
                   "Bptr[%d] = 0x%x.\n",temp1,temp1,Fptr[temp1], \
                   temp1,Bptr[temp1]);
        }
    }
#endif
    DBGMSG("scheduler: after loop Wptr = 0x%x.\n", Wptr);
      
    DBGMSG("scheduler: about to call IsRunQEmpty with priority = %d" \
           " Wptr = %x.\n", current_max_priority, Wptr);
    DBGMSG("scheduler: after after loop Wptr = 0x%x.\n");
    /* }}} */
    if (!IsRunQEmpty(current_max_priority)) {
        /* still some processes left at current priority */
        Wptr = GetNextProcessFrom(current_max_priority);
    } else {
        /* {{{  debug*/
        DBGMSG("scheduler: about to call GetNextProcess with priority = %d" \
               " Wptr = %x.\n", current_max_priority, Wptr);
        /* }}} */
        /* {{{  check that we've not deadlocked*/
        if (NoProcesses()) {
            DBGMSG( "scheduler: Run queue empty.\n");
#ifndef EXTERNAL_COMMS
            /* unfortunately we cannot have deadlock detection with external comms
             * because we have no way to detect processes waiting on external
             * channels.
             */
            if (kbd_chan==NotProcess_p) {
                if (Tptr == NotProcess_p) {
                    STATUS &= 1;
                    dead(STATUS);
                }
            }
            safe_pause();
#endif
#ifdef GSL_SMP
            RELEASE(global_spin_lock); /* allow other process(or) to get it */
            CLAIM(wait_lock); /* wait for the other process(or) to add a new process,
                               * it will release this spin lock when it does
                               */
            CLAIM(global_spin_lock); /* reaquire the lock to get back into the
                                      * scheduler
                                      */
#endif
            KJUMP(X_scheduler);
        }
        /* }}} */
        /* search for next process */
        /* from _current_ priority level downwards! */
        Wptr = GetNextProcess(current_max_priority);
    }
#else
    /* {{{  if (more processes) {*/
    if (NoProcesses()) {
        DBGMSG( "scheduler: Run queue empty.\n");
#ifndef EXTERNAL_COMMS
        /* unfortunately we cannot have deadlock detection with external comms
         * because we have no way to detect processes waiting on external
         * channels.
         */
        if (kbd_chan==NotProcess_p) {
            if (Tptr == NotProcess_p) {
                STATUS &= 1;
                dead(STATUS);
            }
        }
        safe_pause();
#endif
        KJUMP(X_scheduler);
    } else {
        /* }}} */
        Wptr = GetNextProcess();
    }
#endif
    return_address = Wptr[Iptr];
    DBGMSG("scheduler: next process continues from 0x%x.\n", return_address);
    RESTORE;
    return ;
    DBGMSG("scheduler: fallen off end of function!\n");
    exit(1);
/* }}} */

#ifdef PRIORITIES
/* {{{  C_arbitrary_priority_change*/
/* Arbitrary process priority change */
void X_C_arbitrary_priority_change()
{
    LOADWPTR;
    LOADRETURN;
    TWOPARAMS(other_workspace, priority);
    /* {{{  */
    DBGMSG("X_C_arbitrary_priority_change: other workspace = 0x%x, priority = %d.\n", other_workspace, priority);
    /* }}} */
    if (other_workspace[SuspendQ]==NotProcess_p) {
        /* Must be blocked, or the current process */
        /* {{{  */
        DBGMSG("X_C_arbitrary_priority_change: must be blocked or current process.\n");
        /* }}} */
        if (other_workspace != Wptr) {
            /* if not the current process */
            /* {{{  */
            DBGMSG("X_C_arbitrary_priority_change: not current process.\n");
            /* }}} */
            other_workspace[Priority_f] = priority;
            /* {{{  */
            DBGMSG("X_C_arbitrary_priority_change: other_workspace[Priority_f] = %d.\n", other_workspace[Priority_f]);
            /* }}} */
        } else {
            /* it is the current process */
            /* {{{  */
            DBGMSG("X_C_arbitrary_priority_change: is current process.\n");
            /* }}} */
            if (priority >= Wptr[Priority_f]) {
                DBGMSG("X_C_arbitrary_priority_change: we are raising our own priority.\n");
                /* {{{  */
                /* We are highering out own priority, so just change it */
                /* }}} */
                other_workspace[Priority_f] = priority;
            } else {
                /* we are lowering our own priority, so we must search from our old
                 * maximum level downwards.
                 */
                /* {{{  */
                DBGMSG("X_C_arbitrary_priority_change: we are lowering our own priority.\n");
                /* }}} */
                temp1 = Wptr[Priority_f]; /* save current priority level */
                Wptr[Priority_f] = priority;
                /* save context */
                Wptr[Iptr] = return_address;
                /* reschedule current process at new priority level */
                QUEUE(Wptr);
                /* Get new process */
                Wptr = GetNextProcess(temp1); /* minus 1? */
                /* Get its continuation address */
                return_address = Wptr[Iptr];
                RESTORE;
                return ;
            }
        }
    } else {
        /* its on a run queue */
        if (priority <= Wptr[Priority_f]) {
            /* if its of lower priority than us then remove it from its current
             * run queue.
             */
            /* {{{  */
            DBGMSG("X_C_arbitrary_priority_change: its of lower priority than us, so removing it from its current run queue and requeue.\n");
            /* }}} */
            RemoveFromRunQ(other_workspace[Priority_f], other_workspace);
            /* change its priority level */
            other_workspace[Priority_f] = priority;
            /* Requeue it at its new level */
            QUEUE(other_workspace);
        } else {
            /* if its going to be of higher priority than us then remove it from
             * its current run queue, reschedule us, and continue as it.
             */
            /* {{{  */
            DBGMSG("X_C_arbitrary_priority_change: its of higher priority than us, so removing it from its current run queue and requeue.\n");
            /* }}} */
            RemoveFromRunQ(other_workspace[Priority_f], other_workspace);
            /* change its priority level */
            other_workspace[Priority_f] = priority;
            /* save our continuation address */
            Wptr[Iptr] = return_address;
            /* queue us */
            QUEUE(Wptr);
            /* set up Wptr and return address to new process to continue as */
            Wptr = other_workspace;
            return_address = other_workspace[Iptr];
        }
    }
    RESTORE;
    return ;
}
/* }}} */
/* {{{  C_own_priority_change*/
void X_C_own_priority_change()
{
    LOADWPTR;
    LOADRETURN;
    PARAM(priority);
    if (priority >= Wptr[Priority_f]) {
        /* if we are raising our priority, or keeping it the same, just change the
         * priority field and continue as us
         */
        Wptr[Priority_f] = priority;
    } else {
        temp1 = Wptr[Priority_f];
        Wptr[Priority_f] = priority; /* set new priority level */
        Wptr[Iptr] = return_address; /* save continuation address */
        QUEUE(Wptr); /* reschedule current process at new priority level */
        Wptr = GetNextProcess(temp1); /* search for new process down from old
                                       * priority level
                                       */
        return_address = Wptr[Iptr]; /* get new continuation address */
    }
    RESTORE;
    return ;
}
/* }}} */
/* {{{  C_pri_schedule*/
/* pri_schedule schedules in the next process from a given priority
 * level.  This call is used by ProcPar and ProcPriPar to switch to the highest
 * priority process that they have just launched.
 */
void X_C_pri_schedule()
{
    LOADRETURN; /* just to keep stack balanced */
    PARAM(priority_level);
    Wptr = GetNextProcessFrom(priority_level);
    return_address = Wptr[Iptr];
    RESTORE;
    return ;
}
/* }}} */
/* {{{  C_pri_reschedule (not used)*/
#if 0
/* NOT NEEDED FOR NOW */
/* pri_reschedule: as pri_schedule, except save current processes context */
void X_C_pri_reschedule()
{
    LOADWPTR;
    LOADRETURN;
    PARAM(priority_level);
    /* save current context */
    Wptr[Iptr] = return_address;
    /* queue it for rescheduling */
    QUEUE(Wptr);
    /* get the next process */
    Wptr = GetNextProcessFrom(priority_level);
    /* get its continuation address */
    return_address = Wptr[Iptr];
    return ;
}
#endif
/* }}} */
#endif

/* {{{  C_pri_startp*/
#ifdef PRIORITIES
/* startp for CSPlib when ProcRun'ing a process - if the process we are
 * starting is of higher priority that the current process, then we switch
 * to the other process.  This is the C version, so uses an absolute address
 * for the start of the other process rather than an offset.
 */

void X_C_pri_startp()
{
    LOADWPTR;
    LOADRETURN;
    TWOPARAMS(other_workspace, start_address);
    other_workspace[Iptr] = start_address;
    if (GetPriority(other_workspace) > GetPriority(Wptr)) {
        /* We are starting a process of a higher priority that ourself. */
        /* deschedule the caller */
        DBGMSG("C_startp: starting a process of higher priority than me, ");
        DBGMSG("switching to it.");
        /* save continuation address for descheduled process */
        Wptr[Iptr] = return_address;
        QUEUE(Wptr);
        Wptr = other_workspace; /* Wptr = new processes Wptr */
        return_address = start_address;  /* return to new process */
        RESTORE; /* restore Wptr and esi */
        return ;
    } else {
        DBGMSG("C_startp: starting a process of equal or lower priority than me");
        QUEUE(other_workspace);
        RESTORE; /* restore Wptr and esi */
        return ;
    }
}
#endif
/* }}} */

/* {{{  C_startp*/
/* startp for CSPlib - address of process to start is absolute. */
void X_C_startp()
{
    LOADWPTR;
    LOADRETURN;
    TWOPARAMS(other_workspace, start_address);
    other_workspace[Iptr] = start_address;
    QUEUE(other_workspace);
    RESTORE; /* restore Wptr and esi */
    return ;
}
/* }}} */

/* {{{  C_endp*/
/* endp for CSPlib - sets up stack pointer */
void X_C_endp()
{
    LOADRETURN; /* junk the return address */
    LOADWPTR;
    PARAM(usage); /* Wptr becomes successor_workspace */
    usage[Count]--;
    if (usage[Count]) {
        DBGMSG("C_endp: terminating process and calling scheduler.\n");
        KJUMP(X_scheduler); /* Wptr is workspace so we don't need to set up ebp */
    } else {
        Wptr = usage; /* added to correctly update Wptr 091198 */
        /* for some reason, putting a print statement in here messes things up.
         * probably because the printf etc messes up the stack after the restoresp
         * because the compiler's doing something wierd.
         */
        /*DBGMSG("C_endp: terminating process and jumping to successor process.\n");
         */
        RESTORESP(usage[2]); /* for csplib to restore stack pointer for C process */
        /* any attempt to restore the frame pointer should be inserted here. */
        EJUMP(usage[IptrSucc]);
    }
}
/* }}} */

/* {{{  endp*/
/* endp for occam - doesn't restore system stack pointer */
void X_endp()
{
    LOADRETURN; /* junk the return address */
    LOADWPTR;
    PARAM(usage); /* Wptr becomes successor_workspace */
    usage[Count]--;
    if (usage[Count]) {
        Wptr = usage;
        KJUMP(X_scheduler); /* Wptr is workspace so we don't need to set up ebp */
    } else {
        Wptr = usage; /* added to correctly update WPtr 091198 */
        RESTORE;
        EJUMP(usage[IptrSucc]);
    }
}
/* }}} */

/* {{{  startp*/
/* startp for occam - address of start code is an offset */
void X_startp()
{
    LOADWPTR;
    LOADRETURN;
    TWOPARAMS(other_workspace, start_offset);
    DBGMSG("startp: called with other_workspace = 0x%x, offset = 0x%x", \
           other_workspace, start_offset);
    other_workspace[Iptr] = start_offset+return_address;
#ifdef PRIORITIES
    if (GetPriority(other_workspace) > GetPriority(Wptr)) {
        /* We are starting a process of a higher priority that ourself. */
        /* save continuation address for descheduled process */
        Wptr[Iptr] = return_address;
        /* deschedule the caller */
        QUEUE(Wptr);
        Wptr = other_workspace; /* Wptr = new processes Wptr */
        return_address += start_offset;  /* return to new process */
        RESTORE; /* restore Wptr and esi */
        return ;
    } else {
        QUEUE(other_workspace);
        RESTORE; /* restore Wptr and esi */
        return ;
    }
#else
    QUEUE(other_workspace);
    RESTORE; /* restore Wptr and esi */
    return ;
#endif
}
/* }}} */

/* {{{  runp*/

void X_runp()
{
    LOADWPTR;
    LOADRETURN;
    PARAM(other_workspace);
    (word)other_workspace &= 0xfffffffe; /* clear bit zero (priority bit) */
    QUEUE(other_workspace);
    RESTORE; /* restore Wptr and esi */
    return ;
}

/* }}} */

/* {{{  ldtimer*/

void X_ldtimer()
{
    LOADWPTR;
    LOADRETURN;
    now = Time_GetTime();
    DBGMSG( "ldtimer: now = %d.\n", now);
#if 0
    fprintf(stderr, "ldtimer: returning %Lx as time, hi = %x, lo = %x.\n",
            now,
            hi(now), lo(now));
#endif
#ifdef CPU_TIMERS
    RETURNVAL(hi(now));
    RETURNVAL(lo(now));
#else
    RETURNVAL(now);
#endif
    RESTORE;
    return ;
}

/* }}} */

/* {{{  tin*/

void X_tin()
{
    LOADWPTR;
    LOADRETURN;
#ifdef CPU_TIMERS
    TWOPARAMS(lo(wait_time), hi(wait_time));
#else
    PARAM(wait_time);
#endif
    now = Time_GetTime();
#if 0
    fprintf(stderr, "X_tin: now = %Lx, wait_time = %Lx.\n", now, wait_time);
#endif
#if 1
    if (Time_AFTER(now,wait_time)) {
#else
        if ((now-wait_time)>0) {
#endif
            DBGMSG( "tin: timer expired, returning to caller.\n");
            DBGMSG( "tin: wait time is %d, now is %d, now-wt = %d," \
                    " now-wt > 0 = %d\n", \
                    wait_time, now, (now-wait_time), \
                    ((now-wait_time)>0));
#if 0
            fprintf(stderr, "tin: now-wait_time >0.\n");
#endif
            RESTORE;
            return ;
        } else { asm("X_tin_no:");
        DBGMSG( "tin: wait time not expired.");
#ifdef CPU_TIMERS
        DBGMSG( "tin: wait time is %Ld, now is %Ld, now-wt = %Ld," \
                " now-wt > 0 = %d\n", \
                wait_time, now, (now-wait_time), \
                ((now-wait_time)>0));
#else
        DBGMSG( "tin: wait time is %d, now is %d, now-wt = %d," \
                " now-wt > 0 = %d\n", \
                wait_time, now, (now-wait_time), \
                ((now-wait_time)>0));
#endif
        Wptr[Iptr] = return_address;
        Wptr[State] = Waiting_p; /* T9000 book says enabling */
        wait_time++; /* from T9000 book... */
        GetTimeField(Wptr) = wait_time;
        /* start of pseudo code in SPARC code */
        current = Tptr;
        previous = NotProcess_p;
        while ((current != NotProcess_p) && (Time_AFTER(wait_time,GetTimeField(current)))) {
            previous = current;
            current = (word *)current[TLink];
        }
        Wptr[TLink] = (word)current;
        if (previous != NotProcess_p) {
            previous[TLink] = (word)Wptr;
        } else {
            Tptr = Wptr;
#ifdef CPU_TIMERS
            DBGMSG( "tin: setting alarm with delay %Ld.", (GetTimeField(Tptr)) - now);
#else
            DBGMSG( "tin: setting alarm with delay %d.", (GetTimeField(Tptr)) - now);
#endif
            Time_SetAlarm(Time_MINUS(GetTimeField(Tptr), now));
        }
        KJUMP(X_scheduler);
        }
    }

/* }}} */

/* {{{  altstart (not used)*/

/* NOT USED AT PRESENT */
void X_altstart()
{
    LOADRETURN;
        LOADWPTR;
        Wptr[State] = Enabling_p;
        RESTORE;
        return ;
}

/* }}} */

/* {{{  altend (not used)*/

/* NOT USED AT PRESENT */
void X_altend()
{
    LOADWPTR;
        LOADRETURN;
        DBGMSG("altend: called.\n");
        return_address += Wptr[Temp];
        RESTORE;
        return ;
}

/* }}} */

/* {{{  altwt*/

void X_altwt()
{
    LOADWPTR;
        LOADRETURN;
        DBGMSG("altwt: called.\n");
        Wptr[Temp] = NoneSelected_o;
        if (Wptr[State] == Ready_p) {
            DBGMSG("altwt: guard is Ready_p.  Returning to caller.\n");
            RESTORE;
            return ;
        } else {
            DBGMSG( "altwt: guard is not Ready_p.   Decheduling current process.\n");
            Wptr[State] = Waiting_p;
            SETRETURN(return_address);
            RESTORE;

            KJUMP(X_stopp);
        }
}

/* }}} */

/* {{{  taltwt*/

void X_taltwt()
{
    LOADWPTR;
        LOADRETURN;
        DBGMSG( "taltwt: called.\n");
        Wptr[Temp] = NoneSelected_o;
        if (Wptr[State] == Ready_p) {
            DBGMSG("taltwt: State is Ready_p so Wptr[Time_f] = now.\n");
            now = Time_GetTime();
            GetTimeField(Wptr) = now;
        } else {
            if (Wptr[TLink] == TimeNotSet_p) {
                DBGMSG("taltwt: TLink is TimeNotSet_p.  Descheduling.\n");
                Wptr[State] = Waiting_p;
                /* ... wait until Wptr^[State] = Ready_p (execute stopp) */
                SETRETURN(return_address);
                RESTORE;
                KJUMP(X_stopp);
            } else {
                DBGMSG("taltwt: TLink is TimeSet_p.\n");
                now = Time_GetTime();
                if (Time_AFTER(now, GetTimeField(Wptr))) { /* now AFTER Wptr^[Time_f] */
                    DBGMSG("taltwt: now AFTER Wptr[Time_f].\n");
                    Wptr[State] = Ready_p;                     /* from T9000 Instruction */
                    GetTimeField(Wptr) = now;                     /* Set p.436 */
                } else {
                    DBGMSG("taltwt: now is not AFTER Wptr[Time_f].");
                    DBGMSG("Inserting this process into timer queue.\n");
                    GetTimeField(Wptr)++;
                    /* ... insert this process in timer queue */
                    current = Tptr;
                    previous = (word *)NotProcess_p;
                    while ( (current != NotProcess_p) &&
                            (Time_AFTER(GetTimeField(Wptr), GetTimeField(current)))
                            ) {
                        previous = current; /* deduce: previous <> NotProcess_p */
                        current = (word *)current[TLink];
                    }
                    Wptr[TLink] = (word)current;
                    if (previous != NotProcess_p) { /* iff: current <> Tptr */
                        DBGMSG("taltwt: Inserting in middle of queue.\n");
                        previous[TLink] = (word)Wptr;       /* insert in middle of queue */
                    } else {
                        DBGMSG("taltwt: Inserting at head of queue (and setting alarm).\n");
                        DBGMSG("taltwt: Wptr[Time_f] = %d, now = %d", Wptr[Time_f], now);
                        Tptr = Wptr;                  /* insert at head of queue */
                        Time_SetAlarm(Time_MINUS(GetTimeField(Wptr), now));
                    }
                    /* ... end insert */
                    if (Wptr[State] != Ready_p) { /* redundant ? */
#ifdef DEBUG_KERNEL
                        DBGMSG("taltwt: Wptr[State] is not Ready_p, set to Waiting_p");
                        DBGMSG("and deschedule.\n");
#endif
                        Wptr[State] = Waiting_p; /* see altwt */
                        /* ... wait until Wptr[State] = Ready_p (execute stopp) */
                        SETRETURN(return_address);
                        RESTORE;
                        KJUMP(X_stopp);
                    } else {
                        impossible(1, return_address);
                    }
                }
            }
        }
        DBGMSG( "taltwt:exiting to caller.\n");
        RESTORE;
        return ;
}

/* }}} */

/* {{{  dist*/

void X_dist()
{
    LOADWPTR;
        LOADRETURN;
#ifdef CPU_TIMERS
        FOURPARAMS(code_offset, precondition, lo(timeout), hi(timeout));
#else
        THREEPARAMS(code_offset, precondition, timeout);
#endif
        DBGMSG( "dist: entered.\n");
        if (!precondition) {
            DBGMSG( "dist: precondition is false.\n");
            result = false;
        } else {
            DBGMSG( "dist: precondition is true.\n");
            if (Wptr[TLink] == TimeNotSet_p) {
                DBGMSG( "dist: no timer is ready.\n");
                result = false;
            } else {
                if (Wptr[TLink] == TimeSet_p) {
                    DBGMSG( "dist: some timer is ready...\n");
                    /* some timer has gone off */
                    if (!(Time_AFTER(GetTimeField(Wptr), (Time)timeout))) {
                        DBGMSG( "dist: ... but not this one.");
                        result = false;
                    } else {
                        if (Wptr[Temp] != NoneSelected_o) {
                            DBGMSG("dist: a previous guard was selected.\n");
                            result = false;
                        } else {
                            DBGMSG("dist: This is the first ready guard.\n");
                            Wptr[Temp] = code_offset;
                            result = true;
                        }
                    }
                } else {
                    /* {{{  cancel timer*/
                    DBGMSG("dist: strage else condition - remove process from tq.\n");
                    result = false;
                    /* remove process from timer queue */
                    if (Tptr == Wptr) {
                        DBGMSG( "dist: Process is at head of timer Q.\n");
                        Tptr = (word *)Tptr[TLink];
                        if (Tptr != NotProcess_p) {
                            DBGMSG("dist: there is something else on the timer queue.\n");
                            /* Tptr[Link] = NotProcess_p -- commented out in comment! */
                            now = Time_GetTime();
                            if (!(Time_AFTER(GetTimeField(Wptr), now))) { /*should be same as */
                                /*if (((int)Tptr[Time_f] - now) <= 0) { */
                                DBGMSG( \
                                    "dist: time out has happened.  Setting timer sync.\n");
                                sf.s.tim_sync = 1;
                            } else {
        
                                DBGMSG("dist: timeout not happened yet, set alarm.\n");
                                Time_SetAlarm(Time_MINUS(GetTimeField(Tptr), now));
                            }
                        }
                    } else {
                        if (Tptr == NotProcess_p) {
                            impossible(3, return_address);
                        } else {
                            DBGMSG( "dist: searching timer queue from" \
                                    " 2nd element.\n");
                            previous = Tptr; /* search from 2nd item on timer queue */
                            current = (word *)Tptr[TLink];
                            while ((current != NotProcess_p) && (current != Wptr)) {
                                DBGMSG( "dist: onto next item...\n");
                                previous = current;
                                current = (word *)current[TLink];
                            }
                            if (current == NotProcess_p) {
                                impossible(4, return_address);
                            }
                            /* deduce: current == Wptr */
                            previous[TLink] = Wptr[TLink];
                        }
                    }
                    DBGMSG("dist: Timer cancelled.\n");
                    Wptr[TLink] = TimeNotSet_p; /* i.e. timer cancelled */
                    /* }}} */
                }
            }
        }
        RETURNVAL(result);
        RESTORE;
        return ;
}

/* }}} */

/* {{{  enbs*/

/* This function is not used by occam, only by the CSPlib library.  It is
 * inlined in tranpc in the occam implementation.
 */
void X_enbs()
{
    LOADWPTR;
        LOADRETURN;
        ONEPARAM(guard);
        if (guard) {
            Wptr[State] = Ready_p;
        }
        RETURNVAL(guard);
        RESTORE;
        return ;
}

/* }}} */

/* {{{  enbc*/

void X_enbc()
{
    LOADWPTR;
        LOADRETURN;
        TWOPARAMS(guard, channel_address);
        DBGMSG("enbc: guard = %d, channel address = 0x%x\n",
               guard, channel_address);
#ifdef EXTERNAL_COMMS
        if (IsExternal(channel_address)) {
            /* This is an external channel */
            /* enable the internal channel field */
            DBGMSG("enbc: External channel detected.  Enabling soft channel.\n");
#ifdef SLOW
            channel_address = GetAddrInternal(channel_address);
#else
            channel_address++;
#endif
            /* because external channel defined as word, this will point at
             * next word, which is the internal channel.  Slow will be needed if
             * field is moved in structure, or structure's layout is unknown.
             */
        }
#endif
        if (guard) {
            if (*channel_address == NotProcess_p) {
                *channel_address = (word *)((word)Wptr | 1);
            } else {
                if (*channel_address != (word *)((word)Wptr | 1)) {
                    Wptr[State] = Ready_p;
                }
            }
        }
        RETURNVAL(guard);
        RESTORE;
        return ;
}

/* }}} */

/* {{{  enbt*/

void X_enbt()
{
    LOADWPTR;
        LOADRETURN;
#ifdef CPU_TIMERS
        THREEPARAMS(guard, lo(timeout), hi(timeout));
#else
        TWOPARAMS(guard, timeout);
#endif
        DBGMSG("enbt: entered. timeout = %d\n", timeout);
        if (guard) {
            if (Wptr[TLink] == TimeNotSet_p) {
                DBGMSG( "enbt: This is the first enbt.\n");
                Wptr[TLink] = TimeSet_p;
                GetTimeField(Wptr) = (Time)timeout;
            } else {
                DBGMSG( "enbt: This is not the first enbt.\n");
                if (Wptr[TLink] == TimeSet_p) {
                    if (Time_AFTER(GetTimeField(Wptr), (Time)timeout)) {
#ifdef DEBUG_KERNEL
                        DBGMSG( "enbt: This enbt has an earlier time.\n");
#endif
                        GetTimeField(Wptr) = (Time)timeout;
                    }
                }
            }
        }
        RESTORE;
        SETRETURN(guard);
        return ;
}

/* }}} */

/* {{{  disc*/
void X_disc()
{
    LOADWPTR;
        LOADRETURN;
        THREEPARAMS(process_address, guard, channel_address);
        DBGMSG(
            "disc: process_offset = 0x%x, guard = %d, channel address = 0x%x\n",
            process_address, guard, channel_address);
#ifdef EXTERNAL_COMMS
        if (IsExternal(channel_address)) {
            /* This is an external channel */
            /* enable the internal channel field */
#ifdef SLOW
            channel_address = GetAddrInternal(channel_address);
#else
            channel_address++;
#endif
            /* because external channel defined as word, this will point at
             * next word, which is the internal channel.  Slow will be needed if
             * field is moved in structure, or structure's layout is unknown.
             */
        }
#endif
        if (!guard) {
            fired = false;
        } else {
            if (*channel_address == NotProcess_p) {
                fired = false;
            } else {
                if (*channel_address == (word *)((word)Wptr | 1)) {
                    *channel_address = NotProcess_p;
                    fired = false;
                } else {
                    if (Wptr[Temp] == NoneSelected_o) {
                        Wptr[Temp] = process_address; /* actually offset */
                        fired = true;
                    } else {
                        fired = false;
                    }
                }
            }
        }
#if 0
        /* old code */
        temp1 = (word)*channel_address;
        if (((word)Wptr | 1) != temp1) {
            if (temp1 != NotProcess_p) {
                if (Wptr[Temp] == NoneSelected_o) {
                    Wptr[Temp] = process_address;
                    fired = true;
                }
            }
        } else {
            *channel_address = NotProcess_p;
        }
    }
#endif
#ifdef DEBUG_KERNEL
    DBGMSG( "disc: fired = %d.\n", fired);
#endif
    RETURNVAL(fired);
    RESTORE;
    return ;
}
/* }}} */

/* {{{  diss*/
void X_diss()
{
    LOADWPTR;
    LOADRETURN;
    TWOPARAMS(process_address, guard);
#ifdef DEBUG_KERNEL
    DBGMSG( "diss: entered.\n");
#endif
    if (!guard) {
        fired = false;
    } else {
        if (Wptr[Temp] != NoneSelected_o) {
            fired = false;
        } else {
            Wptr[Temp] = process_address;
            fired = true;
        }
    }
    RETURNVAL(fired);
    RESTORE;
    return ;
}
/* }}} */

#ifdef CSPLIB
/* {{{  C_claim_semaphore*/
void X_C_claim_semaphore()
{
    LOADRETURN;
    LOADWPTR;
    PARAM(semaphore);
    if (semaphore->value > 0) {
        semaphore->value--;
        RESTORE;
        return ;
    } else {
        Wptr[Iptr] = return_address;
        EnQueue(Wptr,&(semaphore->q));
        KJUMP(X_scheduler);
    }
}
/* }}} */

/* {{{  C_release_semaphore*/
void X_C_release_semaphore()
{
    LOADRETURN;
    LOADWPTR;
    PARAM(semaphore);
    other_workspace = GetNextProcessFromASuspendQ(&(semaphore->q));
    if (other_workspace != NotProcess_p) {
        /* {{{  check if a higher priority process is available*/
        if (GetPriority(other_workspace) > GetPriority(Wptr)) {
            /* {{{  Comment*/
            /* We know because we (Wptr) are running currently, that we
             * _were_ the highest priority runnable process, but that now
             * the other process (*channel_address) is, so we deschedule
             * Wptr and continue as (*channel_address).  There is no need
             * to queue because we know that (*channel_address)'s run queue
             * is empty (because Wptr was running before at a lower priority.
             */
            /* }}} */
            Wptr[Iptr] = return_address; /* save Iptr for current process */
            QUEUE(Wptr); /* put it on its appropriate run queue */
            Wptr = other_workspace; /* make the other process the current */
            return_address = Wptr[Iptr]; /* get its continuation address */
            /* set up ebp = Wptr and esi = occam_table for occam */
            RESTORE;
            return ;
        } else {
            QUEUE(other_workspace);
            Wptr[Iptr] = return_address;
            QUEUE(Wptr);
            /* gets a process from Q at current (GetPriority(Wptr)) priority
             * level.
             */
            Wptr = GetNextProcess_NonEmptyQ();
            return_address = Wptr[Iptr];
            /* set up ebp = Wptr and esi = occam_table for occam */
            RESTORE;
            return ;
        }
        /* }}} */
    } else {
        semaphore->value++;
        RESTORE;
        return ;
    }
}
/* }}} */

/* {{{  C_release_all_semaphore*/
void X_C_release_all_semaphore()
{
    LOADRETURN;
    LOADWPTR;
    PARAM(semaphore);
    current_max_priority = GetPriority(Wptr);
    AppendAllFromQtoQ(&(semaphore->q), &runq);
    if (current_max_priority > GetPriority(Wptr)) {
        Wptr[Iptr] = return_address;
        QUEUE(Wptr);
        Wptr = GetNextProcessFrom(current_max_priority);
        return_address = Wptr[Iptr];
        RESTORE;
        return ;
    } else {
        RESTORE;
        return ;
    }
} 
/* }}} */
#endif
