/*
 * KROC translator - Arithmetic and logical instructions
 *
 * $Source: /extra/play/cvs/kroc/src/target/alpha/f_arith.c,v $
 *
 * $Id: f_arith.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *
 */

static char f_arith_c_rcsid[] = "$Id: f_arith.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $";

#include "octran.h"
#undef CHECK           /* overflows trapped by the instructions */
#define CHECK

/*   DIFF:
 *   ----
 *    Ref: (T9000 ISM  Page 264)
 *      
 *      A' <- B - A
 *      B' <- C
 */
void DIFF ()
{       mnem("subl"); B; A; B; NL;
	POP;
}

/*   ADD:
 *   ---
 *    Ref: (T9000 ISM  Page 230)
 *      
 *      A' <- B +(checked) A
 *      B' <- C
 */
void ADD ()
{       mnem("addlv"); B; A; B; NL;
	CHECK;
	if (strstr(b_comment, " check for replicator overflow") != NULL)
		pop (2);
	else
		POP;
}

/*   PROD:
 *   ----
 *    Ref: (T9000 ISM  Page 399)
 *      
 *      A' <- A * B
 *      B' <- C
 */
void PROD ()
{
	{       mnem("mull"); B; A; B; NL; }
	POP;
}

/*   SUB:
 *   ---
 *    Ref: (T9000 ISM  Page 429)
 *      
 *      A' <- B -(checked) A
 *      B' <- C
 */
void SUB ()
{       mnem("sublv"); B; A; B; NL;
	CHECK;
	POP;
}

/*   REM:
 *   ---
 *    Ref: (T9000 ISM  Page 402)
 *      
 *      if (A = 0) or ( (A=-1) and (B = MININT) )
 *        ...IntegerOverflow
 *      else
 *        A' <- A rem B
 *      
 *      B' <- C
 */
void REM ()
{       if (o_hardware)
	{       
		mnem("divl"); B; A; T0; nl;
		CHECK;
		mnem("mull"); T0; A; T0; nl;
		mnem("subl"); B; T0; B; nl;
	}
	else
	{       mnem("mov"); B; T2; NL;
		mnem("mov"); A; T3; NL;
		CALL_KRN(K_REM);
		/* don't know how to check overflow */
		mnem("mov"); str("$27"); B; nl;
	}
	POP;
}

/*   DIV:
 *   ---
 *    Ref: (T9000 ISM  Page 270)
 *      
 *      if (A = 0) or ( (A=-1) and (B = MININT) )
 *        ...IntegerOverflow
 *      else
 *        A' <- B / A
 *      
 *      B' <- C
 */
void DIV ()
{       if (o_hardware)
	{       
		mnem("divl"); B; A; B; nl;
		CHECK;
	}
	else
	{       mnem("mov"); B; T2; NL;
		mnem("mov"); A; T3; nl;
		CALL_KRN(K_DIV);
		/* don't know how to check overflow */
		mnem("mov"); str("$27"); B; nl;
	}
	POP;
}

/*   NOT:
 *   ---
 *    Ref: (T9000 ISM  Page 393)
 *      
 *      A' <- ~A
 */
void NOT ()
{       mnem("ornot"); Z; A; A; NL;
}

/*   XOR:
 *   ---
 *    Ref: (T9000 ISM  Page 451)
 *      
 *      A' <- B xor C
 *      B' <- C
 */
void XOR ()
{       mnem("xor"); B; A; B; NL;
	POP;
}

/*   SHR:
 *   ---
 *    Ref: (T9000 ISM  Page 414)
 *      
 *      if ( 0 <= A <= 31)
 *         A' <- B >> A
 *      else
 *         A' <- 0
 *       
 *      B' <- C 
 */
void SHR ()
{
	mnem("blt"); B; f(1); nl;
	mnem("srl"); B; A; B; nl;
	mnem("br");  f(2); nl;
	lab(1); nl;
	UNSGNED(B)
	mnem("srl"); B; A; B; nl;
	lab(2); nl;
	POP;
}

/*   SHL:
 *   ---
 *    Ref: (T9000 ISM  Page 413)
 *      
 *      if ( 0 <= A <= 31)
 *         A' <- B << A
 *      else
 *         A' <- 0
 *       
 *      B' <- C 
 */
void SHL ()
{       
	mnem("sll"); B; A; B; nl;
	POP;
}

/*   AND:
 *   ---
 *    Ref: (T9000 ISM  Page 235)
 *      
 *      A' <- B and A
 *      B' <- C 
 */
void AND ()
{       mnem("and"); B; A; B; NL;
	POP;
}

/*   OR:
 *   --
 *    Ref: (T9000 ISM  Page 394)
 *      
 *      A' <- B or A
 *      B' <- C       
 */
void OR ()
{       mnem("or"); B; A; B; NL;
	POP;
}

/*   SUM:
 *   ---
 *    Ref: (T9000 ISM  Page 430)
 *      
 *      A' <- B + A
 *      B' <- C       
 *      
 */
void SUM ()
{       mnem("addl"); B; A; B; NL;
	POP;
}

/*   MUL:
 *   ---
 *    Ref: (T9000 ISM  Page 390)
 *      
 *      A' <- B *(checked) A
 *      B' <- C       
 *      
 */
void MUL ()
{
	{       mnem("mullv"); B; A; B; NL;
		CHECK;
	}
	POP;
}
