/*
 * KROC translator - Miscellaneous "operate" instructions
 *
 * $Source: /extra/play/cvs/kroc/src/target/alpha/f_etc.c,v $
 *
 * $Id: f_etc.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *
 * ALPHA target only.
 */

static char f_etc_c_rcsid[] = "$Id: f_etc.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $";

#include "octran.h"

/*  REV:
 *  ---
 *   Ref: (T9000 ISM  Page 406)  
 *
 *    Areg' <- Breg
 *    Breg' <- Areg
 *    
 */
void REV ()
{       mnem("mov"); A; T0; NL;
	mnem("mov"); B; A; nl;
	mnem("mov"); T0; B; nl;
}

/*  LB:
 *  --
 *   Ref: (T9000 ISM  Page 351)  
 *
 *    Areg'(0..7) <- byte[Areg]
 *    Areg'(8...) <- 0's
 *    
 */
void LB ()
{       if (o_reverse)
	{       mnem("xor"); A; dec(ByteSelectMask); A; NL; }
	LOAD_BYTE_Zero_Extended(Areg, Areg)
}

/*  BSUB:
 *  ----
 *   Ref: (T9000 ISM  Page 240)  
 *
 *    Areg' <- Areg + Breg
 *    Breg' <- Creg
 *    
 */
void BSUB ()
{       mnem("addl"); B; A; B; NL;
	POP;
}

/*  GCALL:
 *  -----
 *   Ref: (T9000 ISM  Page 334)  
 *
 *    Iptr' <- Areg
 *    Areg' <- Iptr
 *    
 */
void GCALL ()
{       mnem("jmp"); RA; indirect(Areg); NL;
	/* (return address never used from Areg, but allow for called proc which
	    expects it to be in RA) */
	istackp = 0;
}

/*  GT:
 *  --
 *   Ref: (T9000 ISM  Page 338)  
 *
 *    Areg' <- (Breg > Areg)
 *    
 *    
 */
void GT ()
{
	mnem("cmplt"); A; B; B; NL
	POP;
}

/*  WSUB:
 *  ----
 *   Ref: (T9000 ISM  Page 447)  
 *
 *    Areg' <- Areg + (Breg * BytePerWord)
 *    Breg' <- Creg
 *    
 */
void WSUB ()
{
	mnem("s4addl"); B; A; B; nl;
	POP;
}

/*  RESETCH:
 *  -------
 *   Ref: (T9000 ISM  Page 403)  
 *
 *    if (process waiting on channel Areg)
 *      Areg' <- process descriptor of waiting process
 *    else
 *      Areg' <- NotProcess.p
 *
 *    word'[Areg] <- NotProcess.p
 */
void RESETCH ()
{       mnem("ldl"); T0; index1(Areg, 0); NL;
	mnem("stl"); MIN; index1(Areg, 0); nl;
	mnem("mov"); T0; A; nl;
}

/*  CSUB0:
 *  -----
 *   Ref: (T9000 ISM  Page 255)  
 *
 *    if (Breg >= Areg)
 *      ...IntegerError
 *    
 *    Areg' <- Breg
 *    Breg' <- Creg
 */
void CSUB0 ()                                   /* what if checks are off? */
{       mnem("cmpult"); B; A; T0; NL;
	mnem("bne"); T0; f(1); NL;
	GEN_TRAP(T_RANGE);
	lab(1); nl;
	if (strcmp(b_comment, " throw the result away") == 0)
		pop (2);
	else
	{       ishuff += 1;
		POP;
	}
}

/*  NORM:
 *  ----
 *   Ref: (T9000 ISM  Page 392)  
 *
 *    if (Breg(unsigned) = 0) and (Areg(unsigned) = 0)
 *       Creg' <- 2*BitsPerWord
 *    else
 *       Creg' <- no of most significant zero bits in LONG.
 *       Areg' <- (LONG << Creg') rem 2^BitsPerWord
 *       Breg' <- ((LONG << Creg') / 2^BitsPerWord) rem 2^BitsPerWord
 *       
 *         where  LONG = (Breg(unsigned) * 2^BitsPerWord) + Areg(unsigned)
 */
void NORM ()
{
	/*
	 *  C' = 0
	 *  while !(LONG & 0x8000000000000000) {C'++; LONG << 1; }
	 *  B',A' = original(LONG) << C'
	 */

	/* UNSGNED(A) */
	/* UNSGNED(B) */
	PACK(B,A,T0)

	PUSH;

	mnem("or");  Z; dec(64); C; NL;
	mnem("beq"); T0; f(2); NL;          /* Special Case: Zero */

	mnem("sll"); C; dec(57); T1; nl;    /* Set msb of T1 for tests */
	mnem("clr"); C; nl;

lab(1);
	mnem("and");  T0; T1; T2; nl;
	mnem("bne");  T2; f(2); NL;         /* Exit while loop */

	mnem("addl"); C; dec(1); C; NL;
	mnem("sll"); T0; dec(1); T0; NL; 
	mnem("br"); b(1); NL;
lab(2);
	UNPACK(B,A,T0)
	/* SGNEXT(A) */
	/* SGNEXT(B) */
}

/*  LDPI:
 *  ----
 *   Ref: (T9000 ISM  Page 367)  
 *
 *    Areg' <-  Areg + next instruction
 *    
 *    
 */
void LDPI ()
{       mnem("lda"); RA; f(1); COMMENT("get next .t PC"); NL;
	mnem("addl"); A; RA; A; nl;
lab(1); nl;
}

/*  XDBLE:
 *  -----
 *   Ref: (T9000 ISM  Page 450)  
 *
 *    if (Areg >= 0)
 *       Breg' <- 0
 *    else
 *       Breg' <- 1
 *       
 *    Creg' <- Breg
 *       
 */
void XDBLE ()
{       PUSH;
	mnem("mov"); B; A; NL;
	mnem("sra"); B; dec(BitsPerWord-1); B; COMMENT("B := sign(B)"); nl;
}

/*  LDPRI:
 *  -----
 *   Ref: (T9000 ISM  Page 368)  
 *
 *    Areg' <- Priority
 *    Breg' <- Areg
 *    Creg' <- Breg
 *
 */
void LDPRI ()                                   /* could use a bit in STATUS */
{       PUSH;
	mnem("or"); Z; dec(1); A; COMMENT("low priority only"); nl;
/*      mnem("srl"); STATUS; dec(2); A; NL;     */
}

/*  RET:
 *  ---
 *   Ref: (T9000 ISM  Page 405)  
 *
 *    Iptr' <- word[Wptr @ 0]
 *    Wptr' <- Wptr @ 4
 *    
 */
void RET ()                                     /* unadjusted PC used internally */
{       int i, temp;
	mnem("ldl"); RA; index1(Wptr, 0); NL;
	mnem("addl"); W; dec(4*BytesPerWord); W; nl;
	mnem("jmp"); Z; indirect(RAreg);  nl;
	temp = 0;
	if (proctab[stcount].S_PTYPE == P_FUNCTION)
	{       proctab[stcount].S_ISTACK = istackp;
		proctab[stcount].S_FSTACK = fstackp;
		proctab[stcount].S_ISHUFF = ishuff;
		temp = 0;
		for (i = fstackp; i > 0; --i)
			temp = (temp << 1) | ftype[i];
		proctab[stcount].S_FTYPES = temp;
	}
	else
	{       proctab[stcount].S_ISTACK = 0;
		proctab[stcount].S_FSTACK = 0;
		proctab[stcount].S_ISHUFF = 0;
		proctab[stcount].S_FTYPES = 0;  /* unnecessary */
	}
	if (o_verbose)
	{       fprintf (stderr, "\n... End of %s %s %s\n",
			proctab[stcount].label,
			ptype[proctab[stcount].S_PTYPE],
			proctab[stcount].proc);
		fprintf (stderr, "    istack = %d, fstack = [", istackp);
		for (i = fstackp; i > 0; --i)
			fputc ((ftype[i] == F_D) ? 'D' : 'S', stderr);
		fputc (']', stderr);
		where ();
	}
	istackp = 0;
	fstackp = 0;
	ishuff  = 0;
	BREAK;
}

/*  LEND:
 *  ----
 *   Ref: (T9000 ISM  Page 374)  
 *
 *    word'[Breg @ le.Count] <- word[Breg @ le.Count] - 1
 *    if (word'[Breg @ le.Count] > 0)
 *    {
 *       word'[Breg @ le.Index] <- word[Breg @ le.Index] + 1
 *       Iptr' <- next instruction - Areg
 *    }
 *    
 */
void LEND ()
{
	mnem("ldl"); T0; index1(Breg, 1); COMMENT("--count"); NL;
	DEC_Reg(T0)
	mnem("ble"); T0; f(2); NL;
	mnem("stl"); T0; index1(Breg, 1); NL;
	mnem("ldl"); T0; index1(Breg, 0); COMMENT("++control var."); nl;
	INC_Reg(T0)
	mnem("stl"); T0; index1(Breg, 0); nl;

	mnem("lda"); RA; f(1); COMMENT("Next .t Instr "); NL;
	mnem("subl"); RA;  A; RA; NL;  
	mnem("jmp");  Z; indirect(RAreg); NL;  
lab(1); nl;

	BREAK;                                          /* 4 */
lab(2); nl;
	istackp = 0;
	fstackp = 0;
	ishuff = 0;
}

/*  LDTIMER:
 *  -------
 *   Ref: (T9000 ISM  Page 373)  
 *
 *    Areg' <- ClockReg[Priority]
 *    Breg' <- Areg
 *    Creg' <- Breg
 */
void LDTIMER () 
{       PUSH;
	if (o_time)
	    { CALL_KRN(K_LDTIMER); }
	else
	    { CALL_RTN("$utime") }
	mnem("mov"); R0; A; nl;
}

/*  STTIMER:
 *  -------
 *   Ref: (T9000 ISM  Page 428)  
 *
 *    ClockReg'[0] <- Areg
 *    ClockReg'[1] <- Areg
 *    ...Start timers
 *    Areg' <- Breg
 *    Breg' <- Creg
 */
void STTIMER ()
{       unimp ();               /* don't understand */
	POP;
}

/*  TESTPRANAL:
 *  ----------
 *   Ref: (T9000 ISM  Page 437)  
 *
 *    if error since last reset
 *       Areg' <- true
 *    else
 *       Areg' <- false
 *
 *    Breg' <- Areg
 *    Creg' <- Breg
 */
void TESTPRANAL ()
{       PUSH;
	mnem("clr"); A; NL;
}

/*  BCNT:
 *  ----
 *   Ref: (T9000 ISM  Page 236)  
 *
 *    Areg' <- Areg * BytesPerWord
 *    
 *    
 */
void BCNT ()
{       mnem("sll"); A; dec(2); A; NL;
}

/*  XWORD:
 *  -----
 *   Ref: (T9000 ISM  Page 453)  
 *
 *        Sign extend N bit signed no in Breg into a full word.
 *        bit N-1 of Areg when set indicates the value of N. 
 *
 *    if (Areg is not a power of 2)
 *       Areg' <- Undefined
 *    else if (Areg = MININT)
 *       Areg' <- Breg
 *    else if (Breg >= 0) and (Breg < Areg)
 *       Areg' <- Breg
 *    else if (Breg >= Areg) and ((Breg >> 1) < Areg)
 *       Areg' <- Breg - (Areg << 1)
 *    else
 *       Areg' <- Undefined
 */
void XWORD ()
{       mnem("and"); A; B; T0; NL;
	mnem("beq"); T0; f(1); nl;
	mnem("sll"); A; dec(1); T0; nl;
	mnem("subl"); B; T0; B; nl;
	lab(1); nl;
	POP;
}

/*  SB:
 *  --
 *   Ref: (T9000 ISM  Page 408)  
 *
 *    byte'[Areg] <- Breg[0..7]
 *    Areg' <- Creg
 *    
 */
void SB ()
{       if (o_reverse)
	{       mnem("xor"); A; dec(ByteSelectMask); A; NL; }
	SAVE_BYTE(Breg, Areg)
	pop (2);
}

/*  GAJW:
 *  ----
 *   Ref: (T9000 ISM  Page 333)  
 *
 *    Wptr' <- Areg
 *    Areg' <- Wptr
 *    
 */
void GAJW ()
{       mnem("mov"); A; T0; NL;
	mnem("mov"); W; A; nl;
	mnem("mov"); T0; W; nl;
}

/*  WCNT:
 *  ----
 *   Ref: (T9000 ISM  Page 445)  
 *
 *    Areg' <- (Areg and WordSelectMask) / BytesPerWord
 *    Breg' <- Areg and ByteSelectMask
 *    
 */
void WCNT ()
{       PUSH;
	mnem("sra"); B; dec(2); A; NL;
	mnem("and"); B; dec(3); B; nl;
}

/*  MINT:
 *  ----
 *   Ref: (T9000 ISM  Page 382)  
 *
 *    Areg' <- MININT 
 *    Breg' <- Areg
 *    Creg' <- Breg
 */
void MINT ()
{       PUSH;
	mnem("mov"); MIN; A; NL;
}

/*  MOVE:
 *  ----
 *   Ref: (T9000 ISM  Page )  
 *
 *    Copy (Areg) bytes to Address Breg from Address Creg.
 *    
 *    
 */
void MOVE ()                                    /* A >= 0 (CWG, p.21) */
{
#if carl
	if (o_reverse)
	{       CALL_KRN(K_MOVE);       }
	else
	{       
		mnem("beq"); A; f(2); nl;
	lab(1); 
		LOAD_BYTE_Zero_Extended(Creg, Temp3)
		SAVE_BYTE(Temp3,Breg)
		INC_Reg(B)
		INC_Reg(C)
		DEC_Reg(A)
		mnem("bgt"); A; b(1); nl;
	lab(2); nl;
	}
#else
	CALL_KRN(K_MOVE);
#endif
	istackp = 0;
	ishuff = 0;
}

/*  CSNGL:
 *  -----
 *   Ref: (T9000 ISM  Page 253)  
 *
 *    if ((Areg >= 0) and (Breg != 0)) or ((Areg <0) and (Breg != -1))
 *       ...IntegerError
 *    
 *    Breg' <- Creg
 *
 */
void CSNGL ()                                   /* what if checks are off? */
{
	mnem("srl");   A; dec(31); T0; NL;
	mnem("blbc");  T0; f(1); NL;
	mnem("not");   B; B; NL;
lab(1);
	mnem("beq");   B; f(2); NL;
	GEN_TRAP(T_RANGE);
lab(2);
	mnem("mov"); A; B; nl;
	POP;
}

/*  CCNT1:
 *  -----
 *   Ref: (T9000 ISM  Page 245)  
 *
 *    if (Breg =0) or (Breg(Unsigned) > Arge(unsigned))
 *       ...IntegerError
 *
 *    Areg' <- Breg
 *    Breg' <- Creg
 *    
 */
void CCNT1 ()                                   /* what if checks are off? */
{
	DEC_Reg2(B, T0);
	mnem("cmpult"); T0; A; T1; nl;
	mnem("bne"); T1; f(1); NL;
	GEN_TRAP(T_RANGE);
	lab(1); nl;
	if (strcmp(b_comment, " throw the result away") == 0)
		pop (2);
	else
	{       ishuff += 1;
		POP;
	}
}

/*  CWORD:
 *  -----
 *   Ref: (T9000 ISM  Page 256)  
 *
 *    if (Breg < -Areg) or (Breg >= Areg)
 *       ...IntegerError
 *    
 *    Areg' <- Breg
 *    Breg' <- Creg
 */
void CWORD ()                                   /* what if checks are off? */
{
	mnem("addl"); A; B; T0; NL;
	mnem("sll"); A; dec(1); A; nl;
	mnem("cmpult"); T0; A; T1; nl;
	mnem("bne"); T1; f(1); NL;
	GEN_TRAP(T_RANGE);
	lab(1); nl;
	POP;
}







