/*
 * KROC ALPHA translator - Floating-point instructions
 *
 * $Source: /extra/play/cvs/kroc/src/target/alpha/f_float.c,v $
 *
 * $Id: f_float.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *
 * ALPHA target only.
 */

static char f_float_c_rcsid[] = "$Id: f_float.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $";

#include "octran.h"

/*   FPDUP
 *   -----
 *     Ref:  (T9000 ISM  Page 290)
 *
 *       reg' <- FPAreg
 *       FPBreg' <- FPAreg
 *       FPCreg' <- FPBreg
 *       
 *       Areg' <- Areg
 *       Breg' <- Breg
 *       Creg' <- Creg
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPDUP ()
{       FPUSH;
        FAtype = FBtype;
        FMOV_Reg(FBreg, FAreg)
}

/*   FPREV
 *   -----
 *     Ref:  (T9000 ISM  Page 321)
 *
 *       FPAreg' <- FPBreg
 *       FPBreg' <- FPAreg
 *       FPCreg' <- FPCreg
 *       
 *       Areg' <- Areg
 *       Breg' <- Breg
 *       Creg' <- Creg
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPREV ()
{
  int  temp;

        FMOV_Reg(FAreg, FTreg)
        FMOV_Reg(FBreg, FAreg)
        FMOV_Reg(FTreg, FBreg)

        temp = FAtype;
        FAtype = FBtype;
        FBtype = temp;
}

/*   FPLDNLSN
 *   --------
 *     Ref:  (T9000 ISM  Page 306)
 *
 *       FPAreg' <- word [Areg]
 *       FPBreg' <- FPAreg
 *       FPCreg' <- FPBreg
 *
 *       Areg' <- Breg
 *       Breg' <- Creg
 *       Creg' <- Undefined
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPLDNLSN ()
{       FPUSH;
        fmnem("lds"); FA; index1(Areg, LO); NL;

        FAtype = F_S;
        POP;
}

/*   FPLDNLDB
 *   --------
 *     Ref:  (T9000 ISM  Page 302)
 *
 *       FPAreg' <- double [Areg]
 *       FPBreg' <- FPAreg
 *       FPCreg' <- FPBreg
 *
 *       Areg' <- Breg
 *       Breg' <- Creg
 *       Creg' <- Undefined
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPLDNLDB ()
{       FPUSH;
        fmnem("ldt"); FA; index1(Areg, 0); NL;
        FAtype = F_D;
        POP;
}

/*   FPLDNLSNI
 *   ---------
 *     Ref:  (T9000 ISM  Page 307)
 *
 *       FPAreg' <- word [Areg @ Breg]
 *       FPBreg' <- FPAreg
 *       FPCreg' <- FPBreg
 *
 *       Areg' <- Creg
 *       Breg' <- Undefined
 *       Creg' <- Undefined
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPLDNLSNI ()
{       FPUSH;
        mnem("s4addq"); B; A; A; nl;
        fmnem("lds"); FA; index1(Areg, 0); nl;
        FAtype = F_S;
        pop (2);
}

/*   FPLDNLDBI
 *   ---------
 *     Ref:  (T9000 ISM  Page 303)
 *
 *       FPAreg' <- double [Areg @ (2 * Breg)]
 *       FPBreg' <- FPAreg
 *       FPCreg' <- FPBreg
 *
 *       Areg' <- Creg
 *       Breg' <- Undefined
 *       Creg' <- Undefined
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPLDNLDBI ()
{       FPUSH;
        mnem("s8addq"); B; A; A; nl;
        fmnem("ldt"); FA; index1(Areg, 0); nl;
        FAtype = F_D;
        pop (2);
}

/*   FPSTNLSN
 *   --------
 *     Ref:  (T9000 ISM  Page 331)
 *
 *       word [Areg]' <- FPAreg
 *       FPAreg' <- FPBreg
 *       FPBreg' <- FPCreg
 *       FPCreg' <- Undefined
 *
 *       Areg' <- Breg
 *       Breg' <- Creg
 *       Creg' <- Undefined
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPSTNLSN ()
{       fscheck (F_S);
        fmnem("sts"); FA; index1(Areg, LO); NL;
        fstackp -= 1;
        POP;
}

/*   FPSTNLDB
 *   --------
 *     Ref:  (T9000 ISM  Page 329)
 *
 *       double [Areg]' <- FPAreg
 *       FPAreg' <- FPBreg
 *       FPBreg' <- FPCreg
 *       FPCreg' <- Undefined
 *
 *       Areg' <- Breg
 *       Breg' <- Creg
 *       Creg' <- Undefined
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPSTNLDB ()
{       fscheck (F_D);
        fmnem("stt"); FA; index1(Areg, 0); NL;
        fstackp -= 1;
        POP;
}

/*   FPURN
 *   -----
 *     Ref:  (T9000 ISM  Page 323)
 *
 *       FPAreg' <- FPAreg
 *       FPBreg' <- FPBreg
 *       FPCreg' <- FPCreg
 *
 *       Areg' <- Areg
 *       Breg' <- Breg
 *       Creg' <- Creg
 *
 *       RoundMode' <- Nearest 
 *
 */
void FPURN ()
{       fround (R_N);
}

/*   FPURZ
 *   -----
 *     Ref:  (T9000 ISM  Page 326)
 *
 *       FPAreg' <- FPAreg
 *       FPBreg' <- FPBreg
 *       FPCreg' <- FPCreg
 *
 *       Areg' <- Areg
 *       Breg' <- Breg
 *       Creg' <- Creg
 *
 *       RoundMode' <- Zero
 *       
 */
void FPURZ ()
{       fround (R_Z);
}

/*   FPURP
 *   -----
 *     Ref:  (T9000 ISM  Page 324)
 *
 *       FPAreg' <- FPAreg
 *       FPBreg' <- FPBreg
 *       FPCreg' <- FPCreg
 *
 *       Areg' <- Areg
 *       Breg' <- Breg
 *       Creg' <- Creg
 *
 *       RoundMode' <- Plus Infinity
 *
 */
void FPURP ()
{       fround (R_P);
}

/*   FPURM
 *   -----
 *     Ref:  (T9000 ISM  Page 322)
 *
 *       FPAreg' <- FPAreg
 *       FPBreg' <- FPBreg
 *       FPCreg' <- FPCreg
 *
 *       Areg' <- Areg
 *       Breg' <- Breg
 *       Creg' <- Creg
 *
 *       RoundMode' <- Minus Infinity
 *       
 */
void FPURM ()
{       fround (R_M);
}

/*   FPADD
 *   -----
 *     Ref:  (T9000 ISM  Page 283)
 *
 *       FPAreg' <- FPAreg + FPBreg
 *       FPBreg' <- FPCreg
 *       FPCreg' <- Undefined
 *
 *       Areg' <- Areg
 *       Breg' <- Breg
 *       Creg' <- Creg
 *
 *       RoundMode' <- Nearest 
 *       
 */
void FPADD ()
{       fscheck (FBtype);

        switch(FAtype)
            {
            case F_S:   fmnem("adds"); FB; FA; FB; NL;  break;
            default:    fmnem("addt"); FB; FA; FB; NL;  break;
            }

        fstackp -= 1;
}

/*   FPSUB
 *   -----
 *     Ref:  (T9000 ISM  Page 332)
 *
 *       FPAreg' <- FPBreg - FPAreg
 *       FPBreg' <- FPCreg
 *       FPCreg' <- Undefined
 *
 *       Areg' <- Areg
 *       Breg' <- Breg
 *       Creg' <- Creg
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPSUB ()
{       fscheck (FBtype);

        switch(FAtype)
            {
            case F_S:   fmnem("subs"); FB; FA; FB; NL;  break;
            default:    fmnem("subt"); FB; FA; FB; NL;  break;
            }

        fstackp -= 1;
}

/*   FPMUL
 *   -----
 *     Ref:  (T9000 ISM  Page 311)
 *
 *       FPAreg' <- FPAreg * FPBreg
 *       FPBreg' <- FPCreg
 *       FPCreg' <- Undefined
 *
 *       Areg' <- Areg
 *       Breg' <- Breg
 *       Creg' <- Creg
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPMUL ()
{       fscheck (FBtype);

        switch(FAtype)
            {
            case F_S:   fmnem("muls"); FB; FA; FB; NL;  break;
            default:    fmnem("mult"); FB; FA; FB; NL;  break;
            }

        fstackp -= 1;
}

/*   FPDIV
 *   -----
 *     Ref:  (T9000 ISM  Page 288)
 *
 *       FPAreg' <- FPBreg / FPAreg
 *       FPBreg' <- FPCreg
 *       FPCreg' <- Undefined
 *
 *       Areg' <- Areg
 *       Breg' <- Breg
 *       Creg' <- Creg
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPDIV ()
{       fscheck (FBtype);

        switch(FAtype)
            {
            case F_S:   fmnem("divs"); FB; FA; FB; NL;  break;
            default:    fmnem("divt"); FB; FA; FB; NL;  break;
            }

        fstackp -= 1;
}

/*   FPUSQRTFIRST
 *   ------------
 *     Ref:  (CWG  Page 153)
 *
 *       FPAreg' <- Undefined
 *       FPBreg' <- Undefined
 *       FPCreg' <- Undefined
 *
 *       Areg' <- Areg
 *       Breg' <- Breg
 *       Creg' <- Creg
 *
 *       RoundMode' <- Nearest 
 *
 */
void FPUSQRTFIRST ()
{       dummy("fpusqrtfirst"); NL;
}

/*   FPUSQRTSTEP
 *   -----------
 *     Ref:  (CWG  Page 154)
 *
 *       FPAreg' <- Undefined
 *       FPBreg' <- Undefined
 *       FPCreg' <- Undefined
 *
 *       Areg' <- Areg
 *       Breg' <- Breg
 *       Creg' <- Creg
 *
 *       RoundMode' <- Nearest 
 *       
 */
void FPUSQRTSTEP ()
{       dummy("fpusqrtstep"); NL;
}

/*   FPUSQRTLAST
 *   ----------
 *     Ref:  (T9000 ISM  Page 327)
 *
 *       FPAreg' <- SQRT(FPAreg)
 *       FPBreg' <- Undefined
 *       FPCreg' <- Undefined
 *
 *       Areg' <- Areg
 *       Breg' <- Breg
 *       Creg' <- Creg
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPUSQRTLAST ()
{       
        CALL_KRN(K_FPSQRT);
}

/*   FPREMFIRST
 *   ----------
 *     Ref:  (CWG  Page 154)
 *
 *       FPAreg' <- Undefined
 *       FPBreg' <- Undefined
 *       FPCreg' <- Undefined
 *
 *       Areg' <- Undefined
 *       Breg' <- Areg
 *       Creg' <- Breg
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPREMFIRST ()                      /* short cut for representable quotients */
{       fscheck (FBtype);       /* would be better in double throughout */
                                        /* unnecessary? */

        fmnem("divt"); FB; FA; FT; nl;   

        /* Convert to: Nearest Integer */
        fmnem("cvttq"); FT; FT; nl;
        fmnem("cvtqt"); FT; FT; nl;

        fmnem("mult");  FA; FT; FT; nl;
        fmnem("subt");  FB; FT; FA; nl;

        fmnem("cpys");  FA; FA; FB; NL;

        if (FAtype == F_S) { fmnem("cvtts"); FA; FA; nl; }
        if (FBtype == F_S) { fmnem("cvtts"); FB; FB; nl; }


        fstackp = 1;                    /* tricky! */
        istackp = 1;                            /* just the flag */
        cond = C_ALWAYS;                                /* how can you tell? */
}

/*   FPREMSTEP
 *   ---------
 *     Ref:  (CWG  Page 154)
 *
 *       FPAreg' <- Undefined
 *       FPBreg' <- Undefined
 *       FPCreg' <- Undefined
 *
 *       Areg' <- Undefined
 *       Breg' <- Areg
 *       Creg' <- Breg
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPREMSTEP ()                               /* should set FB */
{       dummy("fpremstep"); COMMENT("remember ALWAYS"); NL;
        fround (R_N);
        fstackp = 1;                            /* assume FB ignored */
        istackp = 1;                            /* just the flag */
        cond = C_ALWAYS;
}

/*   FPLDZEROSN
 *   ----------
 *     Ref:  (T9000 ISM  Page 309)
 *
 *       FPAreg' <- 0.0 (single)
 *       FPBreg' <- FPAreg
 *       FPCreg' <- FPBreg
 *       
 *       Areg' <- Areg
 *       Breg' <- Breg
 *       Creg' <- Creg
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPLDZEROSN ()
{       FPUSH;
        FCLR_Reg(FAreg)
        FAtype = F_S;
}

/*   FPLDZERODB
 *   ----------
 *     Ref:  (T9000 ISM  Page 308)
 *
 *       FPAreg' <- 0.0 (double)
 *       FPBreg' <- FPAreg
 *       FPCreg' <- FPBreg
 *       
 *       Areg' <- Areg
 *       Breg' <- Breg
 *       Creg' <- Creg
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPLDZERODB ()
{       FPUSH;
        FCLR_Reg(FAreg);
        FAtype = F_D;
}

/*   FPUMULBY2
 *   ---------
 *     Ref:  (T9000 ISM  Page 312)
 *
 *       FPAreg' <- FPAreg * 2.0
 *       FPBreg' <- FPBreg
 *       FPCreg' <- FPCreg
 *       
 *       Areg' <- Areg
 *       Breg' <- Breg
 *       Creg' <- Creg
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPUMULBY2 ()
{

        switch(FAtype)
            {
            case F_S:   fmnem("adds"); FA; FA; FA; NL;  break;
            default:    fmnem("addt"); FA; FA; FA; NL;  break;
            }

}

/*   FPUDIVBY2
 *   ---------
 *     Ref:  (T9000 ISM  Page 289)
 *
 *       FPAreg' <- FPAreg / 2.0
 *       FPBreg' <- FPBreg
 *       FPCreg' <- FPCreg
 *       
 *       Areg' <- Areg
 *       Breg' <- Breg
 *       Creg' <- Creg
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPUDIVBY2 ()
{

        switch(FAtype)
            {
            case F_S:   fmnem("muls"); FA; HALF; FA; NL;  break;
            default:    fmnem("mult"); FA; HALF; FA; NL;  break;
            }


}

/*   FPUEXPINC32
 *   -----------
 *     Ref:  (T9000 ISM  Page 293)
 *
 *       FPAreg' <- FPAreg * 2^32
 *       FPBreg' <- FPBreg
 *       FPCreg' <- FPCreg
 *       
 *       Areg' <- Areg
 *       Breg' <- Breg
 *       Creg' <- Creg
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPUEXPINC32 ()
{

        switch(FAtype)
            {
            case F_S:   fmnem("muls"); FA; TWO32; FA; NL;  break;
            default:    fmnem("mult"); FA; TWO32; FA; NL;  break;
            }

}

/*   FPUEXPDEC32
 *   -----------
 *     Ref:  (T9000 ISM  Page 292)
 *
 *       FPAreg' <- FPAreg * 2^(-32)
 *       FPBreg' <- FPBreg
 *       FPCreg' <- FPCreg
 *       
 *       Areg' <- Areg
 *       Breg' <- Breg
 *       Creg' <- Creg
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPUEXPDEC32 ()
{

        switch(FAtype)
            {
            case F_S:   fmnem("muls"); FA; TWOM32; FA; NL;  break;
            default:    fmnem("mult"); FA; TWOM32; FA; NL;  break;
            }

}

/*   FPUABS
 *   ------
 *     Ref:  (T9000 ISM  Page 282)
 *
 *       FPAreg' <- 0.0 - FPAreg
 *       FPBreg' <- FPBreg
 *       FPCreg' <- FPCreg
 *       
 *       Areg' <- Areg
 *       Breg' <- Breg
 *       Creg' <- Creg
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPUABS ()
{
        FABS_Reg(FAreg,FAreg)
}

/*   FPLDNLADDSN
 *   -----------
 *     Ref:  (T9000 ISM  Page 301)
 *
 *       FPAreg' <- FPAreg + word [Areg]
 *       FPBreg' <- FPBreg
 *       FPCreg' <- undefined
 *       
 *       Areg' <- Breg
 *       Breg' <- Creg
 *       Creg' <- undefined
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPLDNLADDSN ()
{       fscheck (F_S);
        fmnem("lds"); FT; index1(Areg, LO); NL;
        fmnem("adds"); FA; FT; FA; nl;
        POP;
}

/*   FPLDNLADDDB
 *   -----------
 *     Ref:  (T9000 ISM  Page 300)
 *
 *       FPAreg' <- FPAreg + double [Areg]
 *       FPBreg' <- FPBreg
 *       FPCreg' <- undefined
 *       
 *       Areg' <- Breg
 *       Breg' <- Creg
 *       Creg' <- undefined
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPLDNLADDDB ()
{       fscheck (F_D);
        fmnem("ldt"); FT; index1(Areg, 0); NL;
        fmnem("addt"); FA; FT; FA; nl;
        POP;
}

/*   FPLDNLMULSN
 *   -----------
 *     Ref:  (T9000 ISM  Page 305)
 *
 *       FPAreg' <- FPAreg * word [Areg]
 *       FPBreg' <- FPBreg
 *       FPCreg' <- undefined
 *       
 *       Areg' <- Breg
 *       Breg' <- Creg
 *       Creg' <- undefined
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPLDNLMULSN ()
{       fscheck (F_S);
        fmnem("lds"); FT; index1(Areg, LO); NL;
        fmnem("muls"); FA; FT; FA; nl;
        POP;
}

/*   FPLDNLMULDB
 *   -----------
 *     Ref:  (T9000 ISM  Page 304)
 *
 *       FPAreg' <- FPAreg * double [Areg]
 *       FPBreg' <- FPBreg
 *       FPCreg' <- undefined
 *       
 *       Areg' <- Breg
 *       Breg' <- Creg
 *       Creg' <- undefined
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPLDNLMULDB ()
{       fscheck (F_D);
        fmnem("ldt"); FT; index1(Areg, 0); NL;
        fmnem("mult"); FA; FT; FA; nl;
        POP;
}

/*   FPGT
 *   ----
 *     Ref:  (T9000 ISM  Page 295)
 *
 *       if (FPAreg is a NaN) or (FPBreg is a NaN)
 *         Areg' <- false
 *       else
 *         Areg' <- (FPBreg > FPAreg)
 *       
 *       FPAreg' <- FPCreg
 *       FPBreg' <- undefined
 *       FPCreg' <- undefined
 *       
 *       Breg' <- Areg
 *       Creg' <- Breg
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPGT ()
{       fscheck (FBtype);
        PUSH;
        fmnem("cmptlt"); FA; FB; FPUCC; 

        if (o_opt)
            {
               COMMENT ("remember FGT"); NL;
               cond = C_FGT;
            }
        else
            {
               NL;
               FPCC_INTO(A)
            }

        fstackp -= 2;
}

/*   FPEQ
 *   ----
 *     Ref:  (T9000 ISM  Page 291)
 *
 *       if (FPAreg is a NaN) or (FPBreg is a NaN)
 *         Areg' <- false
 *       else
 *         Areg' <- (FPBreg = FPAreg)
 *       
 *       FPAreg' <- FPCreg
 *       FPBreg' <- undefined
 *       FPCreg' <- undefined
 *       
 *       Breg' <- Areg
 *       Creg' <- Breg
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPEQ ()
{       fscheck (FBtype);
        PUSH;
        fmnem("cmpteq"); FA; FB; FPUCC;

        if (o_opt)
            {
               COMMENT ("remember FEQ"); NL;
               cond = C_FEQ;
            }
        else
            {
               NL;
               FPCC_INTO(A)
            }

        fstackp -= 2;
}

/*   FPORDERED
 *   ---------
 *     Ref:  (T9000 ISM  Page 315)
 *
 *       if (FPAreg is a NaN) or (FPBreg is a NaN)
 *         Areg' <- false
 *       else
 *         Areg' <- true
 *       
 *       FPAreg' <- FPAreg
 *       FPBreg' <- FPBreg
 *       FPCreg' <- FPCreg
 *       
 *       Breg' <- Areg
 *       Creg' <- Breg
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPORDERED ()
{       fscheck (FBtype);
        PUSH;
        fmnem("cmptun"); FA; FB; FPUCC;

        if (o_opt)
            {
               COMMENT ("remember FO"); NL;
               cond = C_FO;
            }
        else
            {
               NL;
               FPCC_INTO(A)
               mnem ("xor"); A; dec(1); A; NL
            }
}

/*   FPNAN
 *   -----
 *     Ref:  (T9000 ISM  Page 313)
 *
 *       if (FPAreg is a NaN)
 *         Areg' <- true
 *       else
 *         Areg' <- false
 *       
 *       FPAreg' <- FPAreg
 *       FPBreg' <- FPBreg
 *       FPCreg' <- FPCreg
 *       
 *       Breg' <- Areg
 *       Creg' <- Breg
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPNAN ()
{       PUSH;
        fmnem("cmptun"); FA; ZERO; FPUCC;

        if (o_opt)
            {
               COMMENT ("remember FU"); NL;
               cond = C_FU;
            }
        else
            {
               NL;
               FPCC_INTO(A)
            }
}

/*   FPNOTFINITE
 *   -----------
 *     Ref:  (T9000 ISM  Page 314)
 *
 *       if (FPAreg is a NaN) or (FPAreg = Infinity)
 *         Areg' <- true
 *       else
 *         Areg' <- false
 *       
 *       FPAreg' <- FPAreg
 *       FPBreg' <- FPBreg
 *       FPCreg' <- FPCreg
 *       
 *       Breg' <- Areg
 *       Creg' <- Breg
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPNOTFINITE ()
{       PUSH;
        fmnem("sts"); FA; FPtemp; NL;
        mnem("ldl");  T0; FPtemp; NL;
      #ifdef carl
        mnem("sll");  T0; dec(1);  T0; NL;
        mnem("srl");  T0; dec(24); CC; NL;
        cond = C_NE;
      #else
        mnem("ldah"); T1; hex(0x7f80); NL;
        mnem("and"); T0; T1; A; NL;
        mnem("cmpeq"); A; T1; A; nl;
      #endif
}

/*   FPUR32TOR64
 *   -----------
 *     Ref:  (T9000 ISM  Page 316)
 *
 *       if (FPAreg is a signalling NaN)
 *         FPAreg' <- real64 (Q(FPAreg))
 *       else
 *         FPAreg' <- real64 (FPAreg)
 *       
 *       FPBreg' <- FPBreg
 *       FPCreg' <- FPCreg
 *       
 *       Areg' <- Areg
 *       Breg' <- Breg
 *       Creg' <- Creg
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPUR32TOR64 ()
{       fscheck (F_S);
        dummy("ieee 32->64"); FA; NL;
        fround (R_N);
        FAtype = F_D;
}

/*   FPUR64TOR32
 *   -----------
 *     Ref:  (T9000 ISM  Page 317)
 *
 *       if (FPAreg is a NaN)
 *         FPAreg' <- R64ToR32NaN
 *       else
 *         FPAreg' <- real32 (FPAreg)
 *       
 *       FPBreg' <- FPBreg
 *       FPCreg' <- FPCreg
 *       
 *       Areg' <- Areg
 *       Breg' <- Breg
 *       Creg' <- Creg
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPUR64TOR32 ()
{       fscheck (F_D);
        fmnem("cvtts"); FA; FA; nl;
        /* dummy("ieee 64->32"); FA; NL; */
        fround (R_N);
        FAtype = F_S;
}

/*   FPINT
 *   -----
 *     Ref:  (T9000 ISM  Page 298)
 *
 *       if (FPAreg is a signalling NaN)
 *         FPAreg' <- Q(FPAreg)
 *       else if (FPAreg is a quiet NaN) or (FPAreg = Infinity)
 *         FPAreg' <- FPAreg
 *       else
 *         FPAreg' <- fpint (FPAreg)
 *       
 *       FPBreg' <- FPBreg
 *       FPCreg' <- FPCreg
 *       
 *       Areg' <- Areg
 *       Breg' <- Breg
 *       Creg' <- Creg
 *       
 *       RoundMode' <- Nearest 
 */
void FPINT ()
{
        fmnem("cvttq"); FA; FA; NL;
        switch (FAtype)
            {
            case F_S:  fmnem("cvtqs"); FA; FA; nl;  break;
            default:   fmnem("cvtqt"); FA; FA; nl;  break;
            }

}

/*   FPSTNLI32
 *   ---------
 *     Ref:  (T9000 ISM  Page 330)
 *
 *       word [Areg] <- int64 (FPAreg) && (2^32 -1)
 *       
 *       FPAreg' <- FPBreg
 *       FPBreg' <- FPCreg
 *       FPCreg' <- undefined
 *       
 *       Areg' <- Breg
 *       Breg' <- Creg
 *       Creg' <- undefined
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPSTNLI32 ()
{
        /* fround(R_M); */
        fmnem("cvttq"); FA; FA; NL;
        fmnem("stt"); FA; FPtemp; nl;
        mnem("ldq");  T0; FPtemp; nl;
        mnem("stl"); T0; index1(Areg, 0); nl;
        POP;
        fstackp -= 1;
}

/*   FPUCHKI32
 *   ---------
 *     Ref:  (T9000 ISM  Page 286)
 *
 *       if (-2^31 <= FPAreg) and (FPAreg < 2^31)
 *          ...FPInvalidOp
 *       
 *       FPAreg' <- FPAreg
 *       FPBreg' <- FPBreg
 *       FPCreg' <- FPCreg
 *       
 *       Areg' <- Areg
 *       Breg' <- Breg
 *       Creg' <- Creg
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPUCHKI32 ()       /* the conversion following should be checked anyway */
{       dummy("fpuchki32"); NL;
        fround (R_N);
}

/*   FPUCHKI64
 *   ---------
 *     Ref:  (T9000 ISM  Page 287)
 *
 *       if (-2^63 <= FPAreg) and (FPAreg < 2^63)
 *          ...FPInvalidOp
 *       
 *       FPAreg' <- FPAreg
 *       FPBreg' <- FPBreg
 *       FPCreg' <- FPCreg
 *       
 *       Areg' <- Areg
 *       Breg' <- Breg
 *       Creg' <- Creg
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPUCHKI64 ()
{       dummy("fpuchki64"); NL;
        fround (R_N);
}

/*   FPRTOI32
 *   --------
 *     Ref:  (T9000 ISM  Page 325)
 *
 *       if (-2^31 <= FPAreg) and (FPAreg < 2^31)
 *          ...FPInvalidOp
 *       
 *       FPAreg' <- fpint(FPAreg)
 *       FPBreg' <- FPBreg
 *       FPCreg' <- FPCreg
 *       
 *       Areg' <- Areg
 *       Breg' <- Breg
 *       Creg' <- Creg
 *
 *       RoundMode' <- Nearest
 *       
 */
void FPRTOI32 ()
{
        fmnem("cvttq"); FA; FA; NL;

        switch (FAtype)
            {
            case F_S:  fmnem("cvtqs"); FA; FA; nl;  break;
            default:   fmnem("cvtqt"); FA; FA; nl;  break;
            }

        /* FPINT (); */
        FPUCHKI32 ();
}

/*   FPI32TOR32
 *   ----------
 *     Ref:  (T9000 ISM  Page 296)
 *
 *       FPAreg' <- real32(word[Areg])
 *       
 *       FPBreg' <- FPAreg
 *       FPCreg' <- FPBreg
 *       
 *       Areg' <- Breg
 *       Breg' <- Creg
 *       Creg' <- undefined
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPI32TOR32 ()
{
        fstackp += 1;

        rround = sround;
        mnem("ldl");  T0; index1(Areg, LO); NL;
        mnem("stq");  T0; FPtemp; nl;

        fround(R_N);
        fmnem("ldt"); FA; FPtemp; nl;
        fround(rround);
        fmnem("cvtqs"); FA; FA; NL;
        FAtype = F_S;
        POP;
}

/*   FPI32TOR64
 *   ----------
 *     Ref:  (T9000 ISM  Page 297)
 *
 *       FPAreg' <- real64(word[Areg])
 *       
 *       FPBreg' <- FPAreg
 *       FPCreg' <- FPBreg
 *       
 *       Areg' <- Breg
 *       Breg' <- Creg
 *       Creg' <- undefined
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPI32TOR64 ()
{       fstackp += 1;

        rround = sround;
        mnem("ldl");  T0; index1(Areg, 0); NL;
        mnem("stq");  T0; FPtemp; nl;

        fround(R_N);
        fmnem("ldt"); FA; FPtemp; nl;
        fround(rround);
        fmnem("cvtqt"); FA; FA; NL;
        FAtype = F_D;
        POP;
}

/*   FPB32TOR64
 *   ----------
 *     Ref:  (T9000 ISM  Page 285)
 *
 *       FPAreg' <- real64(unsigned(word[Areg]))
 *       
 *       FPBreg' <- FPAreg
 *       FPCreg' <- FPBreg
 *       
 *       Areg' <- Breg
 *       Breg' <- Creg
 *       Creg' <- undefined
 *       
 *       RoundMode' <- Nearest 
 *       
 */
void FPB32TOR64 ()
{       fstackp += 1;

        rround = sround;
        mnem("ldl");  T0; index1(Areg, 0); NL;
        mnem("stq");  T0; FPtemp; nl;

        fround(R_N);
        fmnem("ldt"); FA; FPtemp; nl;
        fround(rround);
        fmnem("cvtqt"); FA; FA; NL;
        FAtype = F_D;
        POP;
}

/*   FPUNOROUND
 *   ----------
 *     Ref:  (CWG  Page 160)
 *
 *       FPAreg' <- single(FPAreg)   without rounding
 *
 *       FPBreg' <- FPBreg
 *       FPCreg' <- FPCreg
 *       
 *       Areg' <- Areg
 *       Breg' <- Breg
 *       Creg' <- Creg
 *
 *       RoundMode' <- Nearest 
 *       
 */
void FPUNOROUND ()
{       fscheck (F_D);
        fround (R_Z);
        fmnem("cvtts"); FA; FA; NL;
        FAtype = F_S;
}

/*   FPENTRY
 *   -------
 *     Ref:  (CWG  Page 150)
 *
 *       Areg' <- Breg
 *       Breg' <- Creg
 *       Creg' <- Undefined
 *       
 *       FPAreg' <- ???
 *       FPBreg' <- ???
 *       FPCreg' <- ???
 *       
 *       RoundMode' <- ???
 *       
 *       ... execute operation Areg on floating point unit
 *       
 */
void FPENTRY ()
{       unimp ();
        POP;
}
