/*
 * KROC ALPHA translator - Two-Dimensional block move instructions
 *
 * $Source: /extra/play/cvs/kroc/src/target/alpha/f_move2d.c,v $
 *
 * $Id: f_move2d.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *
 * ALPHA target only.
 */

static char f_move2d_c_rcsid[] = "$Id: f_move2d.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $";

#include "octran.h"

/*   MOVE2DINIT
 *   ----------
 *     Ref: (T9000 ISM  Page 387)
 *
 *       BMreg0' <- Areg
 *       BMreg1' <- Breg
 *       BMreg2' <- Creg
 *
 */
void MOVE2DINIT ()                              /* (CWG, p.95) */
{       mnem("mov"); A; G1; COMMENT("length"); NL;
	mnem("mov"); B; G2; COMMENT("destination stride"); nl;
	mnem("mov"); C; G3; COMMENT("source stride"); nl;
	istackp = 0;
	ishuff = 0;
}

typedef enum { MOVE_ALL, MOVE_ZERO, MOVE_NONZERO } move_types;

void MOVE2D (move_types move_case)
{
	mnem ("beq"); G1; f(4); NL;
	lab(1);
		DEC_Reg(G1);

	if (o_reverse)
	    {   
	       switch (move_case)     /* A bytes from [C] to [B] */
		   {
		   case MOVE_ALL:     CALL_KRN(K_MOVE);  break;

		   case MOVE_ZERO:    CALL_KRN(K_MOVE_ZERO);  break;

		   case MOVE_NONZERO: CALL_KRN(K_MOVE_NONZERO);  break;
		   }
	    }
	else
	{
		mnem ("beq"); A; f(5); NL;
		mnem("mov"); A; R0; nl;
	    lab(2); 
		DEC_Reg(R0);
		mnem("addl"); C; R0; T3; nl;
		LOAD_BYTE_Zero_Extended(Temp3, Temp2)  /* uses T0, T1 */

		    switch (move_case)    /* Decide whether to copy byte. */
		   {
		   case MOVE_ALL:
			break;

		   case MOVE_ZERO:    
			mnem("bne");  T2; f(3);  COMMENT("Skip nonzero bytes"); nl;
			break;

		   case MOVE_NONZERO: 
			mnem("beq");  T2; f(3);  COMMENT("Skip zero bytes"); nl;
			break;
		   }

		mnem("addl"); B; R0; T3; nl;
		SAVE_BYTE(Temp2, Temp3)       /* uses T0, T1, T2 */
	    if (move_case != MOVE_ALL)
	      lab(3);
		mnem("bne");  R0; b(2); nl;
	    lab(5);
	}

	mnem("addl"); B; G2; B; COMMENT("destination address"); nl;
	mnem("addl"); C; G3; C; COMMENT("source address"); nl;
	mnem("bne");  G1; b(1); nl;
	lab(4);

	istackp = 0;
	ishuff = 0;
}




/*   MOVE2DALL
 *   ---------
 *     Ref: (T9000 ISM  Page 386)
 *
 *       if (Overlap)
 *          ... Undefined
 *       else
 *          for y = 0 .. (BMreg0(unsigned)-1)
 *            for x = 0 .. (Areg(unsigned) - 1)
 *              byte' [Breg + (y * BMreg1) + x] <- byte [Creg + (y * BMreg2) + x]
 *       
 */
void MOVE2DALL ()                               /* A >= 0 (CWG, p.80) */
{
    MOVE2D (MOVE_ALL);
}


/*   MOVE2DNONZERO
 *   -------------
 *     Ref: (T9000 ISM  Page 388)
 *
 *       
 *       if (Overlap)
 *          ... Undefined
 *       else
 *          for y = 0 .. (BMreg0(unsigned)-1)
 *            for x = 0 .. (Areg(unsigned) - 1)
 *              if (byte [Creg + (y * BMreg2) + x] <> 0)
 *                byte' [Breg + (y * BMreg1) + x] <- byte [Creg + (y * BMreg2) + x]
 *
 */
void MOVE2DNONZERO ()
{       
    MOVE2D (MOVE_NONZERO);
}

/*   MOVE2DZERO
 *   ----------
 *     Ref: (T9000 ISM  Page 389)
 *
 *       
 *       if (Overlap)
 *          ... Undefined
 *       else
 *          for y = 0 .. (BMreg0(unsigned)-1)
 *            for x = 0 .. (Areg(unsigned) - 1)
 *              if (byte [Creg + (y * BMreg2) + x] = 0)
 *                byte' [Breg + (y * BMreg1) + x] <- byte [Creg + (y * BMreg2) + x]
 *
 */
void MOVE2DZERO ()
{       
    MOVE2D (MOVE_ZERO);
}





