/*
 * KROC ALPHA translator - Instructions involved in ALT constructs and timeouts
 *
 * $Source: /extra/play/cvs/kroc/src/target/arm/f_alt.c,v $
 *
 * $Id: f_alt.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *
 * ALPHA target only.
 */

static char f_alt_c_rcsid[] = "$Id: f_alt.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $";

#include "octran.h"


/*  ALT:
 *  ---
 *  Ref: (T9000 ISM  Page 232)
 *
 *    word'[Wptr @ pw.state] <- Enabling.p
 *
 */
void ArmALT ()
{

    /* word'[Wptr @ pw.state] <- Enabling.p */
    mnem("mov"); T0; immdec(Enabling_p); NL;
    mnem("str"); T0; index1(Wptr, pw_State); nl;
}


/*  ALTWT:
 *  -----
 *  Ref: (T9000 ISM  Page 234)
 *
 *    if (word'[Wptr @ pw.state] != Ready.p)
 *    {
 *       word'[Wptr @ pw.state] <- Waiting.p
 *       ... deschedule process, wait for guard.
 *    }
 *
 *    word'[Wptr @ pw.Temp] <- NoneSelected.o
 *
 */
void ArmALTWT ()
{
    CALL_KRN(K_ALTWT);
    istackp = 0;
    ishuff = 0;
}


/*  ALTEND:
 *  ------
 *  Ref: (T9000 ISM  Page 233)
 *    
 *    IptrReg' <- next instruction + word[Wptr @ pw.Temp]
 *    
 */
void ArmALTEND ()
{
    mnem("ldr"); T0; index1(Wptr, pw_Temp); NL;
    mnem("adr"); T1; f(1); COMMENT("get Iptr"); NL;
    
    mnem("add"); PC; T0; T1; NL;
    lab(1); nl;
    BREAK;
    istackp = 0;
    ishuff = 0;
}


/*  ENBC:
 *  ----
 *  Ref: (T9000 ISM  Page 272)
 *    
 *    if (Areg != false)
 *    {
 *       if (word[Breg] = Notprocess.p)    word'[Breg] <- WdescRef
 *       else if (word[Breg] != WdescReg)  word'[Wptr @ pw.State] <- Ready.p
 *    }
 *    
 *    Breg' <- Creg
 *
 */
void ArmENBC ()
{
    mnem("IMPORT"); barstr("$enbc"); nl;
    
    mnem("cmp"); A; immdec(0); nl;
    mnem("blne"); barstr("$enbc"); nl;
    
    POP;
}


/*  ENBS:
 *  ----
 *  Ref: (T9000 ISM  Page 274)
 *    
 *    if (Areg != false)  word'[Wptr @ pw.State] <- Ready.p
 *    
 */
void ArmENBS ()
{
    mnem("cmp"); A; immdec(0); nl;
    mnem("movne"); T0; immdec(Ready_p); NL;
    mnem("strne"); T0; index1(Wptr, pw_State); NL;
    ishuff += 1;
}


/*  DISC:
 *  ----
 *  Ref: (T9000 ISM  Page 265)
 *    
 *    if (Breg = false)  Areg' <- false
 *    else if (word[Creg] = NotProcess.p)  Areg' <- false
 *    else if (word[Creg] = WdescReg)
 *    {
 *       word'[Creg] <- NotProcess.p
 *       Areg' <- false
 *    }
 *    else if (word[Wptr @ pw.Temp] = NoneSelected.o)
 *    {
 *       word'[Wptr @ pw.Temp] <- Areg
 *       Areg' <- true
 *    }
 *    else  Areg' <- false
 *    
 */
void ArmDISC ()
{
    CALL_KRN(K_DISC);
    ishuff = 1;
    pop (2);
}


/*  DISS:
 *  ----
 *  Ref: (T9000 ISM  Page 268)
 *    
 *    if (Breg = false)  Areg' <- false
 *    else if (word[Wptr @ pw.Temp] != NoneSelected.o)  Areg' <- false
 *    else
 *    {
 *       word'[Wptr @ pw.Temp] <- Areg
 *       Areg' <- true
 *    }
 *
 *    Breg' <- Creg
 *    
 */
void ArmDISS ()
{
    CALL_KRN(K_DISS);
    /* ishuff += 1; */
    POP;
}


/*      Timer stuff     */


/*  TIN:
 *  ---
 *  Ref: (T9000 ISM  Page 439)
 *    
 *    if not (ClockReg[Priority] after Areg)
 *    {
 *       word'[Wptr @ pw.State] <- Enabling.p
 *       word'[Wptr @ pw.Time]  <- (Areg + 1)
 *       ...Insert into timer list with time (Areg +1) and start next process
 *    }
 *
 */
void ArmTIN ()
{
    shufflecount += shuffle (istackp, 1);   /* requires ONE item on stack */
    CALL_KRN(K_TIN);
    istackp = 0;
}


/*  TALT:
 *  ----
 *  Ref: (T9000 ISM  Page 435)
 *    
 *    word'[Wptr @ pw.State]  <- Enabling.p
 *    word'[Wptr @ pw.TLink]  <- TimeNotSet.p
 *
 */
void ArmTALT ()
{
    /* word'[Wptr @ pw.State]  <- Enabling.p */
    mnem("mov"); T0; immdec(Enabling_p); NL;
    mnem("str"); T0; index1(Wptr, pw_State); NL;
    
    /* word'[Wptr @ pw.TLink]  <- TimeNotSet.p */
    mnem("mov"); T0; immdec(TimeNotSet_p); NL;
    mnem("str"); T0; index1(Wptr, pw_TLink); NL;
}


/*  TALTWT:
 *  ------
 *  Ref: (T9000 ISM  Page 436)
 *    
 *    if (word[Wptr @ pw.State] = Ready.p)  word'[Wptr @ pw.Time] <- ClockReg[Priority]
 *    else if (word[Wptr @ pw.TLink] = TimeNotSet.p)
 *    {
 *       word'[Wptr @ pw.State] <- Waiting.p
 *       ... deschedule and wait for guard.
 *    }
 *    else if (word[Wptr @ pw.TLink] = TimeNotSet.p)
 *    {
 *       if (ClockReg[Priority] after word[Wptr @ pw.Time])
 *       {
 *          word'[Wptr @ pw.State] <- Ready.p 
 *          word'[Wptr @ pw.Time]  <- ClockReg[Priority]
 *       }
 *       else
 *       {
 *          word'[Wptr @ pw.Time] <- (word[Wptr @ pw.Time] +1)
 *          ...insert into timer list with alarm time (word[Wptr @ pw.Time] +1)
 *          if (no-guards-ready)
 *          {
 *             word'[Wptr @ pw.State] <- Waiting.p
 *             ...deschedule and wait for guard
 *          }
 *       }
 *    }
 *    
 *    word'[Wptr @ pw.Temp]  <- NoneSelected.o
 *
 */
void ArmTALTWT ()
{
    CALL_KRN(K_TALTWT);
    istackp = 0;
    ishuff = 0;
}


/*  ENBT:
 *  ----
 *  Ref: (T9000 ISM  Page 275)
 *    
 *    if (Areg != false)
 *    {
 *       if (word[Wptr @ pw.TLink] = TimeNotSet.p)
 *       {
 *          word'[Wptr @ pw.TLink] <- TimeSet.p
 *          word'[Wptr @ pw.Time]  <- Breg
 *       }
 *       else if (word[Wptr @ pw.TLink] = TimeSet.p)
 *       {
 *          if (word[Wptr @ pw.Time] after Breg)
 *             word'[Wptr @ pw.Time] <- Breg
 *       }
 *    }
 *
 *    Breg <- Creg
 *
 */
void ArmENBT ()
{
    mnem("IMPORT"); barstr("$enbt"); nl;
    
    mnem("cmp"); A; immdec(0); nl;
    mnem("blne"); barstr("$enbt"); nl;
    
    /* ishuff++; */
    POP;
}


/*  DIST:
 *  ----
 *  Ref: (T9000 ISM  Page 269)
 *    
 *    if (Breg = false)  Areg' <- false
 *    else if (word[Wptr @ pw.TLink] = TimeNotSet.p)  Areg' <- false
 *    else if (word[Wptr @ pw.TLink] = TimeSet.p)
 *    { 
 *       if not (word[Wptr @ pw.Time] after Creg)  Areg' <- false
 *       else if (word[Wptr @ pw.Temp] = NoneSelected.o)
 *       {
 *          word'[Wptr @ pw.Temp] <- Areg
 *          Areg' <- true
 *       }
 *       else  Areg' <- false
 *    }
 *    else  Areg' <- false
 * 
 *    ...remove process from timer list
 *
 */
void ArmDIST ()
{
    CALL_KRN(K_DIST);
    ishuff += 1;
    pop (2);
}
