/*
 * KROC translator - Arithmetic and logical instructions
 *
 * $Source: /extra/play/cvs/kroc/src/target/arm/f_arith.c,v $
 *
 * $Id: f_arith.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 *
 * arm/f_arith.c:
 * (C) Copyright 1998 R. Ivimey-Cook <ruthc@sharra.demon.co.uk>
 *
 * Derived from alpha/f_arith.c:
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *   /u3/proj/kroc/develop/src/alpha/RCS/f_arith.c,v
 *   f_arith.c,v 1.14 1996/06/26 15:32:09 mdp2 Exp
 *
 */

static char f_arith_c_rcsid[] = "$Id: f_arith.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $";

#include <string.h>
#include "octran.h"

static int imported_sdiv = 0;


/*   DIFF:
 *   ----
 *    Ref: (T9000 ISM  Page 264)
 *      
 *      A' <- B - A
 *      B' <- C
 */
void ArmDIFF ()
{
    mnem("sub"); B; B; A; NL;
    POP;
}

/*   ADD:
 *   ---
 *    Ref: (T9000 ISM  Page 230)
 *      
 *      A' <- B +(checked) A
 *      B' <- C
 */
void ArmADD ()
{
    mnem("add"); B; A; B; NL;
    if (strstr(b_comment, " check for replicator overflow") != NULL)
        pop (2);
    else
        POP;
}

/*   PROD:
 *   ----
 *    Ref: (T9000 ISM  Page 399)
 *      
 *      A' <- A * B
 *      B' <- C
 */
void ArmPROD ()
{
    mnem("mul"); B; A; B; NL;
    POP;
}

/*   SUB:
 *   ---
 *    Ref: (T9000 ISM  Page 429)
 *      
 *      A' <- B -(checked) A
 *      B' <- C
 */
void ArmSUB ()
{
    mnem("sub"); B; B; A; NL;
    POP;
}

/*   REM:
 *   ---
 *    Ref: (T9000 ISM  Page 402)
 *      
 *      if (A = 0) or ( (A=-1) and (B = MININT) )
 *        ...IntegerOverflow
 *      else
 *        A' <- A rem B
 *      
 *      B' <- C
 */
void ArmREM ()
{
    if (!imported_sdiv)
    {
        mnem("IMPORT"); str("__rt_sdiv"); NL;
        imported_sdiv = 1;
    }
    
    mnem("mov"); T0; A; NL;
    mnem("mov"); T1; B; NL;
    mnem("bl"); str("__rt_sdiv"); NL;
    mnem("mov"); B; T1;  COMMENT("div returns r0 = div, r1 = mod"); nl;
    POP;
}

/*   DIV:
 *   ---
 *    Ref: (T9000 ISM  Page 270)
 *      
 *      if (A = 0) or ( (A=-1) and (B = MININT) )
 *        ...IntegerOverflow
 *      else
 *        A' <- B / A
 *      
 *      B' <- C
 */
void ArmDIV ()
{
    if (!imported_sdiv)
    {
        mnem("IMPORT"); str("__rt_sdiv"); NL;
        imported_sdiv = 1;
    }
    
    mnem("mov"); T0; A; NL;
    mnem("mov"); T1; B; NL;
    mnem("bl"); str("__rt_sdiv"); NL;
    mnem("mov"); B; T0;  COMMENT("div returns r0 = div, r1 = mod"); nl;
    
    POP;
}

/*   NOT:
 *   ---
 *    Ref: (T9000 ISM  Page 393)
 *      
 *      A' <- ~A
 */
void ArmNOT ()
{
    mnem("rsb"); A; A; immdec(0); NL;
}

/*   XOR:
 *   ---
 *    Ref: (T9000 ISM  Page 451)
 *      
 *      A' <- B xor C  * really? *
 *      B' <- C
 */
void ArmXOR ()
{
    mnem("eor"); B; A; B; NL;
    POP;
}

/*   SHR:
 *   ---
 *    Ref: (T9000 ISM  Page 414)
 *      
 *      if ( 0 <= A <= 31)
 *         A' <- B >> A
 *      else
 *         A' <- 0
 *       
 *      B' <- C 
 */
void ArmSHR ()
{
    mnem("mov"); B; B; shftop2("lsr", Areg); nl;
    POP;
}

/*   SHL:
 *   ---
 *    Ref: (T9000 ISM  Page 413)
 *      
 *      if ( 0 <= A <= 31)
 *         A' <- B << A
 *      else
 *         A' <- 0
 *       
 *      B' <- C 
 */
void ArmSHL ()
{       
    mnem("mov"); B; B; shftop2("lsl", Areg); nl;
    POP;
}

/*   AND:
 *   ---
 *    Ref: (T9000 ISM  Page 235)
 *      
 *      A' <- B and A
 *      B' <- C 
 */
void ArmAND ()
{
    mnem("and"); B; B; A; NL;
    POP;
}

/*   OR:
 *   --
 *    Ref: (T9000 ISM  Page 394)
 *      
 *      A' <- B or A
 *      B' <- C       
 */
void ArmOR ()
{
    mnem("orr"); B; B; A; NL;
    POP;
}

/*   SUM:
 *   ---
 *    Ref: (T9000 ISM  Page 430)
 *      
 *      A' <- B + A
 *      B' <- C       
 *      
 */
void ArmSUM ()
{
    mnem("add"); B; B; A; NL;
    POP;
}

/*   MUL:
 *   ---
 *    Ref: (T9000 ISM  Page 390)
 *      
 *      A' <- B *(checked) A
 *      B' <- C       
 *      
 */
void ArmMUL ()
{
    mnem("mul"); T0; B; A; NL;
    mnem("mov"); A; T0; NL;
    POP;
}
