/*
 * KROC ALPHA translator - Instructions concerned with run-time errors
 *                 (excluding floating point)
 *
 * $Source: /extra/play/cvs/kroc/src/target/arm/f_error.c,v $
 *
 * $Id: f_error.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 * arm/f_error.c:
 * (C) Copyright 1998 R. Ivimey-Cook <ruthc@sharra.demon.co.uk>
 *
 * Derived from alpha/f_error.c:
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *   /u3/proj/kroc/develop/src/alpha/RCS/f_error.c,v
 *   f_error.c,v 1.16 1996/06/26 15:42:22 mdp2 Exp
 *
 */

static char f_error_c_rcsid[] = "$Id: f_error.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $";

#include "octran.h"

/*
 *  SETERR:
 *  ------
 *   Ref: (T9000 ISM  Page )
 *
 *     A' <- A
 *     B' <- B
 *     C' <- C
 *     ErrorFlag' <- True
 *     
 */
void ArmSETERR ()
{
    switch (o_checks)
    {
	case CH_NONE:
	case CH_SETFLAG:
            SET_EFLAG;
            break;
            
	case CH_TRAP:
	case CH_HALT:
            GEN_TRAP(T_SETERR);
            break;
    }
}

/*
 *  TESTERR:
 *  -------
 *
 *     A' <- ~ ErrorFlag
 *     B' <- A
 *     C' <- B
 *     ErrorFlag' <- clear
 *     
 *     
 *     
 */
void ArmTESTERR ()
{
    PUSH;
    switch (o_checks)
    {
	case CH_NONE:
	case CH_TRAP:
            dummy("testerr"); COMMENT("remember NEVER"); NL;
            cond = C_NEVER;
            break;
            
	case CH_SETFLAG:
	case CH_HALT:
            TST_EFLAG(A);
            CLR_EFLAG;
            cond = C_NE;
    }
}

/*
 *  STOPERR:
 *  -------
 *
 *     if ErrorFlag
 *      then  
 *       A' <- Undefined
 *       B' <- Undefined
 *       C' <- Undefined
 *       W[ ] <- i
 *      else
 *       A' <- A
 *       B' <- B
 *       C' <- C
 *
 *     ErrorFlag' <- ErrorFlag
 *     
 */
void ArmSTOPERR ()
{
    switch (o_checks)
    {
	case CH_NONE:
	case CH_TRAP:
	case CH_HALT:
            dummy("stoperr"); COMMENT("off"); NL;
            break;
            
	case CH_SETFLAG:
            TST_EFLAG(CC);
            mnem("beq"); CC; f(1); nl;
            CALL_KRN(K_STOPP);
            lab(1); nl;
    }
}

/*
 *  CLRHALTERR:
 *  ----------
 *
 *     A' <- A
 *     B' <- B
 *     C' <- C
 *
 *     ErrorFlag' <- ErrorFlag
 *     HaltFlag'  <- False
 *     
 *     
 */
void ArmCLRHALTERR ()
{
    CLR_HFLAG;
}

/*
 *  SETHALTERR:
 *  ----------
 *
 *     A' <- A
 *     B' <- B
 *     C' <- C
 *
 *     ErrorFlag' <- ErrorFlag
 *     HaltFlag'  <- True
 *     
 *     
 *     
 */
void ArmSETHALTERR ()
{
    SET_HFLAG;
}

/*
 *  TESTHALTERR:
 *  ----------
 *
 *     A' <- HaltFlag
 *     B' <- A
 *     C' <- B
 *
 *     ErrorFlag' <- ErrorFlag
 *     HaltFlag'  <- HaltFlag
 *     
 *     
 *     
 */
void ArmTESTHALTERR ()
{
    PUSH;
    TST_HFLAG(A);
    cond = C_NE;
}


#ifdef FLOAT
/*
 *  FPCHKERR:
 *  --------
 *   Ref: (T9000 ISM  Page )
 *
 *     
 *     
 *     
 */
void ArmFPCHKERR ()
{
    /* assume floating point errors trapped by software interrupt to OS */
    fround (R_N);
}

/*
 *  FPTESTERR:
 *  ---------
 *   Ref: (T9000 ISM  Page )
 *
 *     
 *     
 *     
 */
void ArmFPTESTERR ()
{
    PUSH;
    /* assume floating point errors trapped by software interrupt to OS */
    fround (R_N);
}

/*
 *  FPUSETERR:
 *  ---------
 *     
 */
void ArmFPUSETERR ()
{
    switch (o_checks)
    {
	case CH_NONE:
	case CH_SETFLAG:
	case CH_HALT:
            dummy("fpuseterr"); COMMENT("off"); NL;
            break;
            
	case CH_TRAP:
            GEN_TRAP(T_FLOAT);
            COMMENT("force floating-point error"); NL;
            break;
    }
}

/*
 *  FPUCLRERR:
 *  ---------
 *     
 */
void ArmFPUCLRERR ()
{
    switch (o_checks)
    {
	case CH_NONE:
	case CH_TRAP:
	case CH_SETFLAG:
	case CH_HALT:
            dummy("fpuclrerr"); COMMENT("off"); NL;
            break;
    }
}

#endif /* FLOAT */
