/*
 * KROC translator - Miscellaneous "operate" instructions
 *
 * $Source: /extra/play/cvs/kroc/src/target/arm/f_etc.c,v $
 *
 * $Id: f_etc.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 * f_etc.c:
 * (C) Copyright 1998 R. Ivimey-Cook <ruthc@sharra.demon.co.uk>
 *
 *
 * Derived from alpha/f_etc.c:
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *   /u3/proj/kroc/develop/src/alpha/RCS/f_etc.c,v
 *   f_etc.c,v 1.25 1996/10/28 16:56:50 mdp2 Exp
 *
 *
 */

static char f_etc_c_rcsid[] = "$Id: f_etc.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $";

#include "octran.h"

/*  REV:
 *  ---
 *   Ref: (T9000 ISM  Page 406)  
 *
 *    Areg' <- Breg
 *    Breg' <- Areg
 *    
 */
void ArmREV ()
{
    mnem("mov"); T0; A; NL;
    mnem("mov"); A; B; nl;
    mnem("mov"); B; T0; nl;
}

/*  LB:
 *  --
 *   Ref: (T9000 ISM  Page 351)  
 *
 *    Areg'(0..7) <- byte[Areg]
 *    Areg'(8...) <- 0's
 *    
 */
void ArmLB ()
{
    mnem("ldrb"); A; index1(Areg, 0); NL;
}

/*  BSUB:
 *  ----
 *   Ref: (T9000 ISM  Page 240)  
 *
 *    Areg' <- Areg + Breg
 *    Breg' <- Creg
 *    
 */
void ArmBSUB ()
{
    mnem("add"); B; A; B; NL;
    POP;
}

/*  GCALL:
 *  -----
 *   Ref: (T9000 ISM  Page 334)  
 *
 *    Iptr' <- Areg
 *    Areg' <- Iptr
 *    
 */
void ArmGCALL ()
{
    mnem("mov"); T0; A; nl;    
    mnem("mov"); A; str("pc"); COMMENT("Areg := pc + 8 => ins after branch"); nl;
    mnem("b"); index1(Temp0, 0); NL;
    istackp = 0;
}

/*  GT:
 *  --
 *   Ref: (T9000 ISM  Page 338)  
 *
 *    Areg' <- (Breg > Areg)
 *    
 *    
 */
void ArmGT ()
{
    mnem("cmp"); B; A; NL;
    mnem("movgt"); A; immdec(1); NL;
    mnem("movle"); A; immdec(0); NL;
    POP;
}

/*  WSUB:
 *  ----
 *   Ref: (T9000 ISM  Page 447)  
 *
 *    Areg' <- Areg + (Breg * BytePerWord)
 *    Breg' <- Creg
 *    
 */
void ArmWSUB ()
{
    mnem("mov"); T0; B; str("lsl #2"); COMMENT("#2 is log2 BytesPerWord"); nl;
    mnem("add"); A; A; T0; NL;
    POP;
}

/*  RESETCH:
 *  -------
 *   Ref: (T9000 ISM  Page 403)  
 *
 *    if (process waiting on channel Areg)
 *      Areg' <- process descriptor of waiting process
 *    else
 *      Areg' <- NotProcess.p
 *
 *    word'[Areg] <- NotProcess.p
 */
void ArmRESETCH ()
{
    mnem("ldr"); T0; index1(Areg, 0); NL;
    mnem("mov"); T1; immdec(NotProcess_p); NL;
    mnem("str"); T1; index1(Areg, 0); nl;
    mnem("mov"); A; T0; nl;
}

/*  CSUB0:
 *  -----
 *   Ref: (T9000 ISM  Page 255)  
 *
 *    if (Breg >= Areg)
 *      ...IntegerError
 *    
 *    Areg' <- Breg
 *    Breg' <- Creg
 */
void ArmCSUB0 ()                                   /* what if checks are off? */
{
    mnem("cmp"); B; A; NL;
    mnem("blt"); f(1); NL;
    
    GEN_TRAP(T_RANGE);

    lab(1); nl;
    if (strcmp(b_comment, " throw the result away") == 0)
        pop (2);
    else
    {
        ishuff += 1;
        POP;
    }
}

/*  NORM:
 *  ----
 *   Ref: (T9000 ISM  Page 392)  
 *
 *    if (Breg(unsigned) = 0) and (Areg(unsigned) = 0)
 *       Creg' <- 2*BitsPerWord
 *    else
 *       Creg' <- no of most significant zero bits in LONG.
 *       Areg' <- (LONG << Creg') rem 2^BitsPerWord
 *       Breg' <- ((LONG << Creg') / 2^BitsPerWord) rem 2^BitsPerWord
 *       
 *         where  LONG = (Breg(unsigned) * 2^BitsPerWord) + Areg(unsigned)
 */
void ArmNORM ()
{
    mnem("cmp"); B; immdec(0); NL;
    mnem("cmpne"); A; immdec(0); NL;
    mnem("bne"); f(1); NL;
    mnem(""); NL;
    mnem(""); NL;
    mnem("b"); f(2); NL;
    
  lab(1); nl;
    mnem("mov"); C; immdec(BitsPerWord * 2); NL;
    
  lab(2); nl;
}

/*  LDPI:
 *  ----
 *   Ref: (T9000 ISM  Page 367)  
 *
 *    Areg' <-  Areg + next instruction
 *    
 *    
 */
void ArmLDPI ()
{
    mnem("add"); A; PC; A; nl;
    
/*   mnem("sub"); A; A; immdec(4); COMMENT("ARM pc 8 ahead but TP only 4 ahead"); nl; */
}

/*  XDBLE:
 *  -----
 *   Ref: (T9000 ISM  Page 450)  
 *
 *    if (Areg >= 0)
 *       Breg' <- 0
 *    else
 *       Breg' <- 1
 *       
 *    Creg' <- Breg
 *       
 */
void ArmXDBLE ()
{
    PUSH;
    mnem("mov"); A; B; NL;
    mnem("mov"); B; B; str("asl"); immdec(BitsPerWord-1); COMMENT("B := sign(B)"); nl;
}

/*  LDPRI:
 *  -----
 *   Ref: (T9000 ISM  Page 368)  
 *
 *    Areg' <- Priority
 *    Breg' <- Areg
 *    Creg' <- Breg
 *
 */
void ArmLDPRI ()                                   /* could use a bit in STATUS */
{
    PUSH;
    mnem("mov"); A; immdec(1); COMMENT("low priority only"); nl;
}

/*  RET:
 *  ---
 *   Ref: (T9000 ISM  Page 405)  
 *
 *    Iptr' <- word[Wptr @ 0]
 *    Wptr' <- Wptr @ 4
 *    
 */
void ArmRET ()
{
#ifdef FLOAT
    int i, temp;
#endif
    
    /* this could be optimised! */
    mnem("add"); W; W; immdec(4*BytesPerWord); nl;
    mnem("ldr"); PC; index1(Wptr, -4*BytesPerWord); nl;
    
    if (proctab[stcount].S_PTYPE == P_FUNCTION)
    {
        proctab[stcount].S_ISTACK = istackp;
        proctab[stcount].S_ISHUFF = ishuff;
#ifdef FLOAT
        proctab[stcount].S_FSTACK = fstackp;
        temp = 0;
        for (i = fstackp; i > 0; --i)
            temp = (temp << 1) | ftype[i];
        proctab[stcount].S_FTYPES = temp;
#endif
    }
    else
    {
        proctab[stcount].S_ISTACK = 0;
        proctab[stcount].S_ISHUFF = 0;
#ifdef FLOAT
        proctab[stcount].S_FSTACK = 0;
        proctab[stcount].S_FTYPES = 0;  /* unnecessary */
#endif
    }
    if (o_verbose)
    {
        fprintf (stderr, "\n... End of %s %s %s\n",
                 proctab[stcount].label,
                 ptype[proctab[stcount].S_PTYPE],
                 proctab[stcount].proc);
        fprintf (stderr, "    istack = %d", istackp);
#ifdef FLOAT
        fprintf (stderr, ", fstack = [", istackp);
        for (i = fstackp; i > 0; --i)
            fputc ((ftype[i] == F_D) ? 'D' : 'S', stderr);
        fputc (']', stderr);
#endif
        where ();
    }
    BREAK;
    mnem("LTORG"); NL; /* tell asm it can put data tables here */
    istackp = 0;
    fstackp = 0;
    ishuff  = 0;
    BREAK;
}

/*  LEND:
 *  ----
 *   Ref: (T9000 ISM  Page 374)  
 *
 *    word'[Breg @ le.Count] <- word[Breg @ le.Count] - 1
 *    if (word'[Breg @ le.Count] > 0)
 *    {
 *       word'[Breg @ le.Index] <- word[Breg @ le.Index] + 1
 *       Iptr' <- next instruction - Areg
 *    }
 *    
 */
void ArmLEND ()
{
    mnem("ldr"); T0; index1(Breg, 1); COMMENT("--count"); NL;
    mnem("subs"); T0; T0; immdec(1); NL;
    mnem("str"); T0; index1(Breg, 1); NL;
    mnem("beq"); f(1); COMMENT("reach zero -> end"); NL; 
    
    mnem("ldr"); T0; index1(Breg, 0); COMMENT("++control var."); nl;
    mnem("add"); T0; T0; immdec(1); NL;
    mnem("str"); T0; index1(Breg, 0); nl;
    
    mnem("add"); A; A; immdec(4); COMMENT("Adjust b/c PC too far ahead"); NL;
    mnem("sub"); PC; PC; A; NL;  
    
    BREAK;
    mnem("LTORG"); COMMENT("asm data area"); NL; /* tell asm it can put data tables here */
    BREAK;
    
    lab(1); nl;
    
    istackp = 0;
    fstackp = 0;
    ishuff = 0;
}

/*  LDTIMER:
 *  -------
 *   Ref: (T9000 ISM  Page 373)  
 *
 *    Areg' <- ClockReg[Priority]
 *    Breg' <- Areg
 *    Creg' <- Breg
 */
void ArmLDTIMER () 
{
    PUSH;
    CALL_KRN(K_LDTIMER);
    mnem("mov"); A; T0; COMMENT("Temp0 <-> APCS return reg"); nl;
}

/*  STTIMER:
 *  -------
 *   Ref: (T9000 ISM  Page 428)  
 *
 *    ClockReg'[0] <- Areg
 *    ClockReg'[1] <- Areg
 *    ...Start timers
 *    Areg' <- Breg
 *    Breg' <- Creg
 */
void ArmSTTIMER ()
{
    unimp ();               /* don't understand */
    POP;
}

/*  TESTPRANAL:
 *  ----------
 *   Ref: (T9000 ISM  Page 437)  
 *
 *    if error since last reset
 *       Areg' <- true
 *    else
 *       Areg' <- false
 *
 *    Breg' <- Areg
 *    Creg' <- Breg
 */
void ArmTESTPRANAL ()
{
    PUSH;
    mnem("mov"); A; immdec(0); NL;
}

/*  BCNT:
 *  ----
 *   Ref: (T9000 ISM  Page 236)  
 *
 *    Areg' <- Areg * BytesPerWord
 *    
 *    
 */
void ArmBCNT ()
{
    mnem("mov"); A; A; shftop1("asr", 2); NL;
}

/*  XWORD:
 *  -----
 *   Ref: (T9000 ISM  Page 453)  
 *
 *        Sign extend N bit signed no in Breg into a full word.
 *        bit N-1 of Areg when set indicates the value of N. 
 *
 *    if (Areg is not a power of 2)
 *       Areg' <- Undefined
 *    else if (Areg = MININT)
 *       Areg' <- Breg
 *    else if (Breg >= 0) and (Breg < Areg)
 *       Areg' <- Breg
 *    else if (Breg >= Areg) and ((Breg >> 1) < Areg)
 *       Areg' <- Breg - (Areg << 1)
 *    else
 *       Areg' <- Undefined
 */
void ArmXWORD ()
{
    mnem("tst"); A; B; NL;
    mnem("beq"); f(1); nl;
    mnem("mov"); T0; A; shftop1("lsl", 1); nl;
    mnem("sub"); B; B; T0; nl;
    lab(1); nl;
    POP;
}

/*  SB:
 *  --
 *   Ref: (T9000 ISM  Page 408)  
 *
 *    byte'[Areg] <- Breg[0..7]
 *    Areg' <- Creg
 *    
 */
void ArmSB ()
{
    mnem("strb"); B; index1(Areg, 0); nl;
    pop (2);
}

/*  GAJW:
 *  ----
 *   Ref: (T9000 ISM  Page 333)  
 *
 *    Wptr' <- Areg
 *    Areg' <- Wptr
 *    
 */
void ArmGAJW ()
{
    mnem("mov"); T0; A; NL;
    mnem("mov"); A; W; nl;
    mnem("mov"); W; T0; nl;
}

/*  WCNT:
 *  ----
 *   Ref: (T9000 ISM  Page 445)  
 *
 *    Areg' <- (Areg and WordSelectMask) / BytesPerWord
 *    Breg' <- Areg and ByteSelectMask
 *    
 */
void ArmWCNT ()
{
    PUSH;
    mnem("mov"); A; B; str("asr"); immdec(2); NL;
    mnem("and"); B; B; immdec(3); nl;
}

/*  MINT:
 *  ----
 *   Ref: (T9000 ISM  Page 382)  
 *
 *    Areg' <- MININT 
 *    Breg' <- Areg
 *    Creg' <- Breg
 */
void ArmMINT ()
{
    PUSH;
    mnem("mov"); A; immdec(MostNeg); NL;
}

/*  MOVE:
 *  ----
 *   Ref: (T9000 ISM  Page )  
 *
 *    Copy (Areg) bytes to Address Breg from Address Creg.
 *    
 *    
 */
void ArmMOVE ()                                    /* A >= 0 (CWG, p.21) */
{
    CALL_KRN(K_MOVE);
    istackp = 0;
    ishuff = 0;
}

/*  CSNGL:
 *  -----
 *   Ref: (T9000 ISM  Page 253)  
 *
 *    if ((Areg >= 0) and (Breg != 0)) or ((Areg <0) and (Breg != -1))
 *       ...IntegerError
 *    
 *    Breg' <- Creg
 *
 */
void ArmCSNGL ()                                   /* what if checks are off? */
{
    mnem("cmp"); A; immdec(0); NL;     
    mnem("blt"); f(1); NL;             /* a < 0, check b. */
    
    mnem("cmp"); B; immdec(0); NL;     /* a >= 0, check b.  */
    mnem("beq"); f(3); NL;             /* a >= 0 but b == 0, so all ok. */
    mnem("b"); f(2);                   /* else in trouble... */
    
    lab(1); nl;
    mnem("cmp"); B; immdec(-1); NL;    /* a < 0; check b != -1 */
    mnem("beq"); f(2); NL;             /* a < 0 but b == -1, so all ok. */
    
    lab(2); nl;
    GEN_TRAP(T_RANGE);
    
    lab(3); nl;
    
    POP;
}

/*  CCNT1:
 *  -----
 *   Ref: (T9000 ISM  Page 245)  
 *
 *    if (Breg = 0) or (Breg(Unsigned) > Areg(unsigned))
 *       ...IntegerError
 *
 *    Areg' <- Breg
 *    Breg' <- Creg
 *    
 */
void ArmCCNT1 ()                                   /* what if checks are off? */
{
    mnem("cmp"); B; immdec(0); nl;
    mnem("cmpne"); B; A; nl;
    mnem("bls"); f(1); NL;
    GEN_TRAP(T_RANGE);
    
    lab(1); nl;
    if (strcmp(b_comment, " throw the result away") == 0)
        pop (2);
    else
    {
        ishuff += 1;
        POP;
    }
}

/*  CWORD:
 *  -----
 *   Ref: (T9000 ISM  Page 256)  
 *
 *    if (Breg < -Areg) or (Breg >= Areg)
 *       ...IntegerError
 *    
 *    Areg' <- Breg
 *    Breg' <- Creg
 */
void ArmCWORD ()                                   /* what if checks are off? */
{
    mnem("add"); T0; A; B; NL;
    mnem("mov"); A; A; shftop1("lsl", 1); nl;
    mnem("cmp"); T0; A; nl;
    mnem("blo"); f(1); NL;
    GEN_TRAP(T_RANGE);
    lab(1); nl;
    POP;
}







