/*
 * KROC translator - Long arithmetic instructions
 *
 * $Source: /extra/play/cvs/kroc/src/target/arm/f_long.c,v $
 *
 * $Id: f_long.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 * arm/f_long.c:
 * (C) Copyright 1998 R. Ivimey-Cook <ruthc@sharra.demon.co.uk>
 *
 * Derived from alpha/f_arith.c:
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *   /u3/proj/kroc/develop/src/alpha/RCS/f_long.c,v
 *   f_long.c,v 1.15 1996/08/09 13:39:30 mdp2 Exp
 *
 */

static char f_long_c_rcsid[] = "$Id: f_long.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $";

#include "octran.h"


/*
 *  LADD:
 *  ----
 *   Ref (T9000 ISM  Page 350)  [CSL - Areg after Overflow is not fixed-up]
 *
 *     if (sum > MAXINT)
 *       Areg' <- sum - 2^BitsPerWord
 *       ... IntegerOverflow
 *     else if (sum < MININT) 
 *       Areg' <- sum + 2^BitsPerWord
 *       ... IntegerOverflow
 *     else
 *       Areg' <- sum
 *     
 *        where sum = Areg + Breg + Creg[bit 0]
 */
void ArmLADD ()
{
    /* sum = Areg + Breg + Creg[bit 0] */
    mnem("and"); C; C; immdec(1); NL;
    mnem("add"); A; B; A; NL;
    mnem("add"); A; C; C; NL;

    pop (2);
}

/*
 *  LDIV:
 *  ----
 *   Ref (T9000 ISM  Page 361)
 *
 *     if (Creg(unsigned) >= Areg(unsigned))
 *       ... IntegerOverflow
 *     else
 *       Areg'(unsigned) <- LONG(unsigned) / Areg(unsigned)
 *       Breg'(unsigned) <- LONG(unsigned) rem Areg(unsigned)
 *
 *         where LONG = (Creg(unsigned) * 2^BitsPerWord) + Breg(unsigned)
 */
void ArmLDIV ()                                    /* muldiv == _SW? */
{
    unimp();
    POP;
}

/*
 *  LMUL:
 *  ----
 *   Ref (T9000 ISM  Page 375)
 *
 *       Areg'(unsigned) <- PROD(unsigned) rem 2^BitsPerWord
 *       Breg'(unsigned) <- PROD(unsigned) / 2^BitsPerWord
 *
 *         where PROD = (Breg(unsigned) * Areg(unsigned)) + Creg(unsigned)
 *     
 */
void ArmLMUL ()
{
    /* PROD = (Breg(unsigned) * Areg(unsigned)) + Creg(unsigned) */
    mnem("smull"); T0; T1; B; A; NL;
    mnem("adds");  T0; T0; C; nl;
    mnem("addcs"); T1; immdec(1); nl;
    
    /* Breg'(unsigned) <- PROD(unsigned) / 2^BitsPerWord */
    mnem("mov"); B; T0; NL;    
    mnem("mov"); C; T1; NL;
    
    ishuff = 1;
    POP;
}

/*
 *  LSHR:
 *  ----
 *   Ref (T9000 ISM  Page 378)
 *
 *     Areg' <- (LONG >> Areg(unsigned)) rem 2^BitsPerWord 
 *     Breg' <- (LONG >> Areg(unsigned) / 2^BitsPerWord) rem 2^BitsPerWord
 *     
 *       where LONG = (Creg(unsigned) * 2^BitsPerWord)) + Breg(unsigned)
 *     
 */
void ArmLSHR ()
{
    mnem("movs"); C; C; str("lsr"); A; nl;
    mnem("mov");  B; B; str("rrx"); nl;
    /* rrx is only 1 bit... if more, use lsr again */
    mnem("subs"); A; A; immdec(1); nl;
    mnem("beq"); f(1); nl;    
    mnem("mov");  B; str("lsr"); A; nl;
    
    lab(1);
    
    ishuff = 1;
    POP;
}

/*
 *  LSHL:
 *  ----
 *   Ref (T9000 ISM  Page 377)
 *
 *     Areg' <- (LONG << Areg(unsigned)) rem 2^BitsPerWord 
 *     Breg' <- (LONG << Areg(unsigned) / 2^BitsPerWord) rem 2^BitsPerWord
 *     
 *       where LONG = (Creg(unsigned) * 2^BitsPerWord)) + Breg(unsigned)
 *     
 */
void ArmLSHL ()
{
    
    ishuff = 1;
    POP;
}

/*
 *  LSUM:
 *  ----
 *   Ref (T9000 ISM  Page 380)
 *
 *     if (SUM(unsigned) > MostPosUnsigned)
 *        Areg' <- SUM(unsigned) - 2^BitsPerWord
 *        Breg' <- 1
 *     else
 *        Areg' <- SUM(unsigned)
 *        Breg' <- 0
 *     
 *         where  SUM = Areg(unsigned) + Breg(unsigned) + Creg[0]
 */
void ArmLSUM ()
{
    mnem("and");   C; C; dec(1); NL;
    mnem("adds");  A; C; A; NL;
    mnem("adds");  B; A; B; NL;

    mnem("movcc"); C; immdec(0); nl;
    mnem("movcs"); C; immdec(1); nl;
    
    ishuff = 1;
    POP;
}

/*
 *  LSUB:
 *  ----
 *   Ref (T9000 ISM  Page 379)
 *
 *     if (DIFF > MAX)
 *       ...IntegerOverflow
 *     else if (DIFF < MIN)
 *       ...IntegerOverflow
 *     else
 *       Areg' <- DIFF
 *
 *         where  DIFF = (Breg - Areg) - Creg(0)
 */
void ArmLSUB ()
{
    /* DIFF = (Breg - Areg) - Creg(0) */
    mnem("sub"); T0; B; A; NL;
    mnem("and"); C; C; immdec(1); nl;
    mnem("sub"); C; T0; C; NL;

    pop (2);
}

/*
 *  LDIFF:
 *  ----
 *   Ref (T9000 ISM  Page 360)
 *
 *     if (DIFF(unsigned) <= MAXUINT)
 *       Areg'(unsigned) <- DIFF(unsigned)
 *       Breg' <- 0
 *     else
 *       Areg'(unsigned) <- DIFF(unsigned) + 2^BitsPerWord
 *       Breg' <- 1
 *
 *         where  DIFF = (Breg(unsigned) - Areg(unsigned)) - Creg[0]
 *     
 */
void ArmLDIFF ()
{
    mnem("and"); C; C; immdec(1); nl;
    mnem("sub"); T0; B; A; NL;
    mnem("sub"); T0; T0; C; NL;
    
    /* Breg' <- 0 */
    mnem("mov"); C; immdec(0); NL;
    
    ishuff = 1;
    POP;
}
