/*
 * KROC ARM translator - Instructions concerned with processes
 *
 * $Source: /extra/play/cvs/kroc/src/target/arm/f_proc.c,v $
 *
 * $Id: f_proc.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 *
 * arm/f_move2d.c:
 * (C) Copyright 1998 R. Ivimey-Cook <ruthc@sharra.demon.co.uk>
 *
 * Derived from alpha/f_move2d.c:
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *   /u3/proj/kroc/develop/src/alpha/RCS/f_proc.c,v
 *   f_proc.c,v 1.14 1996/11/06 17:15:06 mdp2 Exp
 * 
 */

static char f_proc_c_rcsid[] = "$Id: f_proc.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $";

#include "octran.h"

/*
 *  ENDP:
 *  ----
 *   Ref: (T9000 ISM  Page 276)
 *
 *     word'[Areg @ pp.Count] <- word[Areg @ pp.Count] - 1
 *     if (word'[Areg @ pp.Count] = 0)
 *     {
 *       Iptr' <- word[Areg @ pp.Iptrsucc]
 *       Wptr' <- Areg
 *     }
 *     else
 *       ... start next process
 */
void ArmENDP ()
{
    /* word'[Areg @ pp.Count] <- word[Areg @ pp.Count] - 1 */
    mnem("ldr"); T0; index1(Areg, 1); NL;
    mnem("subs"); T0; T0; immdec(1); nl;
    mnem("str"); T0; index1(Areg, 1); NL;
    
    /* if (word'[Areg @ pp.Count] = 0) */
    mnem("beq"); f(1); nl;
    CALL_RTN("scheduler");  /* CALL_RTN("$scheduler") */
    lab(1); nl;
    
    /*  Iptr' <- word[Areg @ pp.Iptrsucc] */
    /*  Wptr' <- Areg */
    mnem("ldr"); T0; index1(Areg, 0); nl;
    mnem("mov"); W; A; NL;
    mnem("mov"); PC; T0; COMMENT("should combine with ldr above..."); NL;
    
    istackp = 0;
    fstackp = 0;
    ishuff = 0;
    BREAK;
}

/*
 *  STARTP:
 *  ------
 *   Ref: (T9000 ISM  Page 418)
 *
 *     word'[Areg @ pw.Iptr] <- next .t instruction + Breg
 *     ... put process Areg onto the back of the queue.
 *     
 *     
 */
void ArmSTARTP ()
{
    /* word'[Areg @ pw.Iptr] <- next instruction + Breg */
    mnem("adr"); T0; f(4); COMMENT("get PC of next .t instruction"); NL;
    lab(1); mnem("add"); T0; T0; B; nl;
    mnem("str"); T0; index1(Areg, -1); nl;

	/* ... put process Areg onto the back of the queue. */
    mnem("ldr"); T0; ldrcon(MOSTNEG_INT32); nl;
    mnem("cmp"); FP; T0; nl;
    mnem("bne"); f(2); nl;
    mnem("str"); A; index1(Bptr, -2); nl;
    mnem("b");   f(3); NL;
    lab(2); mnem("mov"); FP; A; nl;
    lab(3); mnem("mov"); BP; A; nl;
    lab(4); nl;
    istackp = 0;
    fstackp = 0;
    ishuff = 0;
}

/*
 *  STOPP:
 *  -----
 *   Ref: (T9000 ISM  Page 425)
 *
 *      word'[Wptr @ pw.Iptr] <- next instruction
 *     
 *     
 *     
 */
void ArmSTOPP ()
{
    CALL_KRN(K_STOPP);
    istackp = 0;
    fstackp = 0;
    ishuff = 0;
}

/*
 *  STHF:
 *  ----
 *   Ref: (T800 ITT  Page 198)
 *
 *     HPBP' <- A
 *     Areg' <- B
 *     Breg' <- C
 *     
 */
void ArmSTHF ()
{
    unimp ();
/*  mnem("mov"); A; FP; NL; */
    POP;
}

/*
 *  STLB:
 *  ----
 *   Ref: (T800 ITT  Page 201)
 *
 *     BP'   <- A
 *     Areg' <- B
 *     Breg' <- C
 *     
 */
void ArmSTLB ()
{
    mnem("mov"); BP; A;; NL;
    POP;
}

/*
 *  STLF:
 *  ----
 *   Ref: (T800 ITT  Page 202)
 *
 *     FP'   <- A
 *     Areg' <- B
 *     Breg' <- C
 *
 */
void ArmSTLF ()
{
    mnem("mov"); FP; A; NL;
    POP;
}

/*
 *  RUNP:
 *  ----
 *   Ref: (T9000 ISM  Page 407)
 *
 *     ... put process Areg onto back of process list
 *     
 *     
 *     
 */
void ArmRUNP ()
{
    /* ... put process Areg onto the back of the queue. */
    mnem("bic"); A; immdec(3); A; NL;
    mnem("mov"); T0; immdec(MOSTNEG_INT32);
    mnem("cmp"); FP; T0; nl;
    mnem("bne"); f(1); nl;
    mnem("str"); A; index1(Bptr, -2); nl;
    mnem("b");   f(2); NL;
    lab(1); mnem("mov"); FP; A; nl;
    lab(2); mnem("mov"); BP; A; nl;

    istackp = 0;
    fstackp = 0;
    ishuff = 0;
}

/*
 *  SAVEH:
 *  -----
 *   Ref: (T800 ITT  Page 189)
 *
 *     
 *     
 *     
 *     
 */
void ArmSAVEH ()
{
    unimp ();
/*      mnem("stl"); BP; index1(Areg, 1); NL;   */
/*      mnem("stl"); FP; index1(Areg, 0); nl;   */
}

/*
 *  SAVEL:
 *  -----
 *   Ref: (T800 ITT  Page 190)
 *
 *     word'[Areg @ 1]  <- lpBP
 *     word'[Areg @ 0]  <- lpFP
 *     A' <- B
 *     B' <- C
 *     
 */
void ArmSAVEL ()
{
    mnem("str"); BP; index1(Areg, 1); NL;
    mnem("str"); FP; index1(Areg, 0); nl;
}

/*
 *  STHB:
 *  ----
 *   Ref: (T800 ITT  Page 198)
 *
 *     
 *     
 *     
 *     
 */
void ArmSTHB ()
{
    unimp ();
/*  mnem("mov"); A; BP; NL; */
    POP;
}
