/*
 * KROC ARM translator - Special instructions for the T800
 *
 * $Source: /extra/play/cvs/kroc/src/target/arm/f_t800.c,v $
 *
 * $Id: f_t800.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 *
 * arm/f_move2d.c:
 * (C) Copyright 1998 R. Ivimey-Cook <ruthc@sharra.demon.co.uk>
 *
 * Derived from alpha/f_t800.c:
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *   /u3/proj/kroc/develop/src/alpha/RCS/f_t800.c,v
 *   f_t800.c,v 1.10 1996/06/26 15:35:50 mdp2 Exp
 *
 */

static char f_t800_c_rcsid[] = "$Id: f_t800.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $";

#include "octran.h"


/*
 *  DUP:
 *  ---
 *   Ref: (T9000 ISM  Page 271)
 *
 *    A' <- A
 *    B' <- A
 *    C' <- B
 *
 */
void ArmDUP ()
{
    PUSH;
    mnem("mov"); A; B; NL;
}

/*
 *  FMUL:
 *  ----
 *   Ref: (T9000 ISM  Page 281)
 *
 *    if (Areg = MININT) and (Breg = MININT)
 *      ... Integer Overflow
 *    else
 *      A' <- (B * A) / 2^(Bitsperword-1)
 *    
 *    B' <- C
 *
 */
void ArmFMUL ()
{
    mnem("smull"); T0; B; B; A;; NL;
    POP;
}

/*
 *  CRCWORD:
 *  -------
 *   Ref: (T9000 ISM  Page 251)
 *
 *    Areg' <- temp(BitsPerWord)
 *    Breg' <- Creg
 *    where
 *      temp(0) = Breg
 *      for i = 1 .. BitsPerWord
 *        temp(i) = (temp(i-1) << 1) + Areg[BPW-i] xor
 *                  (Creg * temp(i-1)[BPW-1])
 *
 */
void ArmCRCWORD ()
{       
    unimp();
    ishuff = 1;
    POP;
}

/*
 *  CRCBYTE:
 *  -------
 *   Ref: (T9000 ISM  Page 250)
 *
 *    Areg' <- temp(8)
 *    Breg' <- Creg
 *    where
 *      temp(0) = Breg
 *      for i = 1 .. 8
 *        temp(i) = (temp(i-1) << 1) + Areg[BPW-i] xor
 *                  (Creg * temp(i-1)[BPW-1])
 *    
 *
 */
void ArmCRCBYTE ()
{
    unimp();
    
    ishuff = 1;
    POP;
}

/*
 *  BITCNT:
 *  ------
 *   Ref: (T9000 ISM  Page 237)
 *
 *    Areg' <- Breg + ...bits set in Areg
 *    Breg' <- Creg
 *
 * Pseudo-code 1:
 *   while (A > 0) begin
 *     A := A >> 1
 *     if (carry set) B := B + 1
 *   end
 *
 * Pseudo-code 2:
 *   while (A > 0) begin
 *     A := (A - 1) and A
 *     B := B + 1
 *   end
 */
void ArmBITCNT ()
{
    mnem("cmp"); A; immdec(0); NL;
    mnem("beq"); f(2); NL;
lab(1);
    mnem("movs"); A; A; str("lsr #1"); NL;
    mnem("addcs"); B; immdec(1);     COMMENT("add 1 to b if bit shifted set"); NL;
    mnem("cmp"); A; immdec(0); NL;
    mnem("bne"); b(1); NL;
lab(2);
    
    POP;
}

void ArmBITCNT ()
{
    mnem("cmp"); A; immdec(0); NL;
    mnem("beq"); f(2); nl;
lab(1);
    mnem("sub"); T0; A; immdec(1);; nl;
    mnem("ands"); A; T0; A; nl;
    mnem("add"); B; B; immdec(1); nl;
    mnem("bne"); b(1); nl;
lab(2);
    nl;
    POP;
}

/*
 *  BITREVWORD:
 *  ----------
 *   Ref: (T9000 ISM  Page 239)
 *
 *    Areg' <- reversed Areg
 *
 *   Inlined operation: first swap halfwords with a rotate, then
 * swap in turn the bytes, nibbles, bit-pairs & bits. The smaller
 * chunks are performed using:
 *
 *   res =  ( ( A >> shift ) & const ) | (( A & const ) << shift)
 *
 *  for shift = 8, 4, 2, 1, while const is a mask value selecting bits.
 *
 *  Pseudo-code:
 *  
 *  Areg := (Areg << 16) | (Areg >> 16)             -- rotate 16
 *  shift := 8
 *  const := 0xff00ff
 *  do begin
 *    Areg :=  ( ( Areg >> shift ) & const ) | (( Areg & const ) << shift)
 *    shift := shift >> 1
 *    const := const eor (const << shift)
 *  while (shift > 0)
 * 
 * ARM Cycles (approx):   17
 */
void ArmBITREVWORD ()
{    
    mnem("mov"); T0; #ff;     COMMENT("T1 = temp reg, A = input value, T0 = constant"); nl;
    mnem("orr"); T0; T0; 0xff00; nl;
    
    COMMENT("swap halfwords & bytes");
    mnem("and"); T1; T0; A; ror #24; nl;
    mnem("and"); A; T0; A; lsl #8; nl;
    mnem("orr"); A; T1; A; nl;
    mnem("eor"); T0; T0; T0; lsl #4; nl;

    COMMENT("swap nibbles");
    mnem("and"); T1; T0; A; lsr #4; nl;
    mnem("and"); A; T0; A; nl;
    mnem("orr"); A; T1; A; lsl #4; nl;
    mnem("eor"); T0; T0; T0; lsl #2; nl;

    COMMENT("swap bit-pairs");
    mnem("and"); T1; T0; A; lsr #2; nl;
    mnem("and"); A; T0; A; nl;
    mnem("orr"); A; T1; A; lsl #2; nl;
    mnem("eor"); T0; T0; T0; lsl #1; nl;

    COMMENT("swap bits");
    mnem("and"); T1; T0; A; lsr #1; nl;
    mnem("and"); A; T0; A; nl;
    mnem("orr"); A; T1; A; lsl #1; nl;
}

/*
 *  BITREVNBITS:
 *  -----------
 *   Ref: (T9000 ISM  Page 238)
 *
 *    Areg'[0..Areg-1] <- reversed Breg[0..Areg-1]
 *    Areg'[Areg..BitsPerWord-1] <- 0
 *    
 *    Breg' <- Creg
 *
 *   Shift Breg down, extracting the bottom bit on each
 * shift and putting it into a temporary, which is shifted up
 * each time.
 *
 * Pseudo-code:
 *   
 *   temp := 0
 *   while ( Areg > 0 ) begin
 *     temp := (Breg & 1) | (temp << 1)
 *     Breg := Breg >> 1
 *     Areg := Areg - 1
 *   end
 *   Breg = temp
 *
 * ARM Cycles (approx):
 *   Areg <> 0 : 4 + (Areg) * 6
 *   Areg == 0: 5
 *
 */
void ArmBITREVNBITS ()
{
    mnem("mov"); T2; dec(0); nl;
    mnem("cmp"); A; dec(0); NL;
    mnem("beq"); f(2); nl;
lab(1);
    mnem("and"); T0; B; dec(1); nl;
    mnem("orr"); T2; T0; T2; lsl #1; nl;
    mnem("mov"); B; B; lsr #1; nl;
    mnem("subs"); A; A; dec(1); nl;
    mnem("bne"); b(1); nl;
lab(2);
    mnem("mov"); B; T2; nl;
    POP;
}

/*
 *  WSUBDB:
 *  ------
 *   Ref: (T9000 ISM  Page 448)
 *
 *    Areg' <- Areg + (2 * Bytesperword * Breg)
 *    Breg' <- Creg
 */
void ArmWSUBDB ()
{
    mnem("mov"); B; B; str("lsl #3"); NL;
    mnem("add"); B; A; B; nl;
    POP;
}
