/*
 * KROC translator: Function definitions and prototypes
 *
 * $Source: /extra/play/cvs/kroc/src/target/asm.h,v $
 *
 * $Id: asm.h,v 1.1 1999/09/27 21:24:00 rivimey Exp $
 *
 * asm.h
 * (C) Copyright 1999 R. Ivimey-Cook <ruthc@sharra.demon.co.uk>
 *
 */

#ifndef OCTRAN_ASM_H
#define OCTRAN_ASM_H

typedef void MnemonicProc (const char *mnemonic);
typedef void CommentProc (const char *str);
typedef void GlobalFunctionProc (const char *name);
typedef void LocalFunctionProc (const char *name);
typedef void StartOfFileProc (void);
typedef void EndOfFileProc (void);
typedef void ImportSymbolProc (const char *name);
typedef void ExportSymbolProc (const char *name);
typedef void ConstantDataProc (const int size, long value);
typedef void InitialiseProc (void);
typedef void CloseProc (void);


typedef struct 
{
    InitialiseProc       *asmInit;
    CloseProc            *asmClose;
    MnemonicProc         *asmMnemonic;
    CommentProc          *asmComment;
    GlobalFunctionProc   *asmGlobalFn;
    LocalFunctionProc    *asmLocalFn;
    StartOfFileProc      *asmStartFile;
    EndOfFileProc        *asmEndFile;
    ImportSymbolProc     *asmImport;
    ExportSymbolProc     *asmExport;
    ConstantDataProc     *asmConstantData;
    AlignProc            *asmAlign;
    
} AssemblerProcVec;

/* routines to hide some of the nitty gritty details when writing code... */
extern AssemblerProcVec *currentAsm;

#define mnem(str)       currentAsm->asmMnemonic(str)
#define align(sz)       currentAsm->asmAlign(sz)
#define comment(str)    currentAsm->asmComment(str)
#define importsym(sym)  currentAsm->asmImport(sym)
#define exportsym(sym)  currentAsm->asmExport(sym)


#endif /* OCTRAN_ASM_H */
