/*
 * KROC SPARC translator: Deal with conditionals.
 *
 * $Source: /extra/play/cvs/kroc/src/target/sparc/cond.c,v $
 *
 * $Id: cond.c,v 1.1 1999/09/27 21:10:13 rivimey Exp $
 *
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *
 * Mostly depends on target, but something similar will be needed.
 *
 * (E.g.: 68000 has conditional set instructions, but only 68020 etc.
 * has conditional traps.)
 */

static char cond_c_rcsid[] = "$Id: cond.c,v 1.1 1999/09/27 21:10:13 rivimey Exp $";

#include "octran.h"

cond_type condtab[C_SIZE] =

/*	Must be in true/false even/odd pairs	*/

/*	{"name",    "branch",  "trap" } */      /* code */

{	{"...",     "...",     "..."  },	/*  0 */
	{"...",     "...",     "..."  },	/*  1 */

	{"GT",      "bg",      "tg"   },	/*  2 */
	{"LE",      "ble",     "tle"  },	/*  3 */

	{"EQ",      "be",      "te"   },	/*  4 */
	{"NE",      "bne",     "tne"  },	/*  5 */

	{"GE",      "bge",     "tge"  },	/*  6 */
	{"LT",      "bl",      "tl"   },	/*  7 */

	{"VS",      "bvs",     "tvs"  },	/*  8 */
	{"VC",      "bvc",     "tvc"  },	/*  9 */

	{"GU",      "bgu",     "tgu"  },	/* 10 */
	{"LEU",     "bleu",    "tleu" },	/* 11 */

	{"GEU",     "bgeu",    "tgeu" },	/* 12 */
	{"LU",      "blu",     "tlu"  },	/* 13 */

	{"NEG",     "bneg",    "tneg" },	/* 14 */
	{"POS",     "bpos",    "tpos" },	/* 15 */

	{"FGT",     "fbg",     "..."  },	/* 16 */
	{"FLE",     "fble",    "..."  },	/* 17 */

	{"FGE",     "fbge",    "..."  },	/* 18 */
	{"FLT",     "fbl",     "..."  },	/* 19 */

	{"FEQ",     "fbe",     "..."  },	/* 20 */
	{"FNE",     "fbne",    "..."  },	/* 21 */

	{"FLG",     "fblg",    "..."  },	/* 22 */
	{"FUE",     "fbue",    "..."  },	/* 23 */

	{"FO",      "fbo",     "..."  },	/* 24 */
	{"FU",      "fbu",     "..."  },	/* 25 */

	{"ALWAYS",  "ba",      "ta"   },	/* 26 */
	{"NEVER",   "bn",      "tn"   }		/* 27 */
};

void defcond (c)		/* do deferred condition */
int c;
{	if (istackp == 0)	/* stackdepth = 0 earlier */
	{	dummy ("Boolean lost"); nl; }
	if (c == C_ALWAYS)
	{	mnem("mov"); TRUE; A; COMMENT("deferred ALWAYS"); nl; }
	else if (c == C_NEVER)
	{	mnem("mov"); FALSE; A; COMMENT("deferred NEVER"); nl; }
	else
	{	mnema(condtab[c].branch); f(1); lineup();
			fprintf(out, "%c+ deferred %s", CCHAR, condtab[c].name); nl;
		mnem("mov"); TRUE; A; annul("true"); nl;
		mnem("mov"); FALSE; A; COMMENT("false"); nl;
	lab(1);	nl;
	}
	cond = C_NONE;
}

void check (c, chk)		/* generate error-checking code */
int c, chk;
{	static const char *chkname[6] =
	{	"SETERR",	/* 0 */
		"OFLOW",	/* 1 */
		"RANGE",	/* 2 */
		"FLOAT",	/* 3 */
		"UNREC",	/* 4 */
		"UNIMP"		/* 5 */
	};
	comma = 0;
	switch (o_checks)
	{
	case CH_NONE:
		dummy("checks off"); nl;
		break;
	case CH_SETFLAG:
		mnema(condtab[c].branch); f(1); nl;
		mnem("bset"); EF; STATUS; ANNUL; nl;
	lab(1);	nl;
		break;
	case CH_TRAP:
		mnem(condtab[c].trap); dec(T_ZERO + chk); lineup();
			fprintf(out, "%c+ %s", CCHAR, chkname[chk]); nl;
	/*	mnem(condtab[c].branch); str(errlab[chk]); nl;	*/
	/*	mnem("nop"); nl;				*/
		break;
	case CH_HALT:		/* should HALT on error, not TRAP */
		mnem(condtab[INV(c)].branch); f(1); nl;
		mnem("btst"); HOEF; STATUS; DELAY; nl;
		mnem("tne"); dec(T_ZERO + chk); nl;
	lab(1);	nl;
	/*	mnem("bne"); str(errlab[chk]); nl;		*/
	/*	mnem("nop"); nl;				*/
		break;
	}
}

/*	See also CJ in "f_direct.c".	*/
