/*
 * KROC SPARC translator: Instructions concerned with run-time errors (including floating point).
 *
 * $Source: /extra/play/cvs/kroc/src/target/sparc/f_error.c,v $
 *
 * $Id: f_error.c,v 1.1 1999/09/27 21:10:13 rivimey Exp $
 *
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *
 * SPARC target only.
 */

static char f_error_c_rcsid[] = "$Id: f_error.c,v 1.1 1999/09/27 21:10:13 rivimey Exp $";

#include "octran.h"

void SETERR ()
{	switch (o_checks)
	{
	case CH_NONE:
		dummy("seterr"); COMMENT("off"); NL;
		break;
	case CH_SETFLAG:
		mnem("bset"); EF; STATUS; NL;
		break;
	case CH_TRAP:
		mnem("ta"); dec(T_ZERO + T_SETERR); COMMENT("SETERR"); NL;
		break;
	case CH_HALT:
		mnem("bset"); EF; STATUS; NL;
		mnem("btst"); HOEF; STATUS; nl;
		mnem("tne"); dec(T_ZERO + T_SETERR); COMMENT("SETERR"); nl;
	}
}

void TESTERR ()
{	PUSH;
	switch (o_checks)
	{
	case CH_NONE:
	case CH_TRAP:
		dummy("testerr"); COMMENT("remember NEVER"); NL;
		cond = C_NEVER;
		break;
	case CH_SETFLAG:
	case CH_HALT:
		mnem("btst"); EF; STATUS; NL;
		mnem("bclr"); EF; STATUS; nl;
		cond = C_NE;
	}
	ishuff += 1;		/* may be used just to clear error flag */
}

void STOPERR ()
{	switch (o_checks)
	{
	case CH_NONE:
	case CH_TRAP:
		dummy("stoperr"); COMMENT("off"); NL;
		break;
	case CH_SETFLAG:
	case CH_HALT:
		mnem("btst"); EF; STATUS; NL;
		mnem("be"); f(1); nl;
		mnem("nop"); nl;
		extcall(X_STOPP); nl;
		ADJUST; nl;
	lab(1);	nl;
	}
}

void CLRHALTERR ()
{	mnem("bclr"); HOEF; STATUS; NL;
}

void SETHALTERR ()
{	mnem("bset"); HOEF; STATUS; NL;
}

void TESTHALTERR ()
{	PUSH;
	mnem("btst"); HOEF; STATUS; NL;
	cond = C_NE;
}

#ifdef FLOAT

void FPCHKERR ()
{	switch (o_checks)
	{
	case CH_NONE:				/* ignore it */
	case CH_TRAP:				/* can't happen */
		dummy("fpchkerr"); COMMENT("off"); NL;
		break;
	case CH_SETFLAG:
	case CH_HALT:
		mnem("st"); FSR; FPtemp; NL;
		mnem("ld"); FPtemp; T0; nl;
		mnem("btst"); hex(AEXC); T0; nl;
		check (C_NE, T_FLOAT);		/* Halt-On-Error stuff */
	}
	fround (R_N);
}

void FPTESTERR ()
{	PUSH;
	switch (o_checks)
	{
	case CH_NONE:
	case CH_TRAP:
		dummy("fptesterr"); COMMENT("remember NEVER"); NL;
		cond = C_NEVER;
		break;
	case CH_SETFLAG:
	case CH_HALT:
		mnem("st"); FSR; FPtemp; NL;	/* have to preserve RN */
		mnem("ld"); FPtemp; T0; nl;
		mnem("btst"); hex(AEXC); T0; nl;
		mnem("bclr"); hex(AEXC); T0; nl;
		mnem("st"); T0; FPtemp; nl;
		mnem("ld"); FPtemp; FSR; nl;
	/*	Could have all eight combinations of RN and AEXC fields	*/
	/*	(Would save only two instuctions)			*/
		cond = C_NE;
	}
	fround (R_N);
}

void FPUSETERR ()
{	switch (o_checks)
	{
	case CH_NONE:
		dummy("fpuseterr"); COMMENT("off"); NL;
		break;
	case CH_TRAP:
		mnem("ta"); dec(T_ZERO + T_FLOAT);
			COMMENT("force floating-point error"); NL;
		break;
	case CH_SETFLAG:
	case CH_HALT:
		mnem("ld"); Aexc; FSR; NL;
	}
}

void FPUCLRERR ()
{	switch (o_checks)
	{
	case CH_NONE:
	case CH_TRAP:
		dummy("fpuclrerr"); COMMENT("off"); NL;
		break;
	case CH_SETFLAG:
	case CH_HALT:
		mnem("ld"); Zero; FSR; NL;
	}
}

#endif /* FLOAT */
