/*
 * KROC SPARC translator: Floating-point instructions.
 *
 * $Source: /extra/play/cvs/kroc/src/target/sparc/f_float.c,v $
 *
 * $Id: f_float.c,v 1.1 1999/09/27 21:10:13 rivimey Exp $
 *
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *
 * SPARC target only.
 */

static char f_float_c_rcsid[] = "$Id: f_float.c,v 1.1 1999/09/27 21:10:13 rivimey Exp $";

#include "octran.h"

int ftype[4];			/* types of f.p. registers */

void FPDUP ()
{	fpush (FAtype);
	if (DBLE)		/* faster than two "fmovs"es */
	{	mnem("faddd"); FBH; ZERO; FAH; NL; }
	else
	{	mnem("fmovs"); FBH; FAH; NL; }
	fround (R_N);
}

void FPREV ()
{	int temp;
	if (DBLE)
	{	mnem("faddd"); FAH; ZERO; FTH; NL;
		mnem("faddd"); FBH; ZERO; FAH; nl;
		mnem("faddd"); FTH; ZERO; FBH; nl;
	}
	else
	{	mnem("fmovs"); FAH; FTH; NL;
		mnem("fmovs"); FBH; FAH; nl;
		mnem("fmovs"); FTH; FBH; nl;
	}
	fround (R_N);
	temp = FAtype;
	FAtype = FBtype;
	FBtype = temp;
}

void FPLDNLSN ()
{	fpush (F_S);
	mnem("ld"); sindex(Areg, 0); FAH; NL;
	fround (R_N);
	POP;
}

void FPLDNLDB ()
{	fpush (F_D);
	if (o_real64)
	{	mnem("ldd"); sindex(Areg, 0); FAH; NL; }
	else
	{	mnem("ld"); sindex(Areg, HI); FAH; NL;
		mnem("ld"); sindex(Areg, LO); FAL; nl;
	}
	fround (R_N);
	POP;
}

void FPLDNLSNI ()
{	fpush (F_S);
	mnem("sll"); B; dec(2); B; NL;
	mnem("ld"); index2(Areg, Breg); FAH; nl;
	fround (R_N);
	pop (2);
}

void FPLDNLDBI ()
{	fpush (F_D);
	mnem("sll"); B; dec(3); B; NL;	/* CWG p.54, NOT p.144! */
	if (o_real64)
	{	mnem("ldd"); index2(Areg, Breg); FAH; nl; }
	else
	{	mnem("add"); A; B; A; nl;
		mnem("ld"); sindex(Areg, HI); FAH; nl;
		mnem("ld"); sindex(Areg, LO); FAL; nl;
	}
	fround (R_N);
	pop (2);
}

void FPSTNLSN ()
{	fscheck (F_S);
	mnem("st"); FAH; sindex(Areg, 0); NL;
	fround (R_N);
	fstackp -= 1;
	POP;
}

void FPSTNLDB ()
{	fscheck (F_D);
	if (o_real64)
	{	mnem("std"); FAH; sindex(Areg, 0); NL; }
	else
	{	mnem("st"); FAH; sindex(Areg, HI); NL;
		mnem("st"); FAL; sindex(Areg, LO); nl;
	}
	fround (R_N);
	fstackp -= 1;
	POP;
}

void FPURN ()
{	fround (R_N);
}

void FPURZ ()
{	fround (R_Z);
}

void FPURP ()
{	fround (R_P);
}

void FPURM ()
{	fround (R_M);
}

void FPADD ()
{	fscheck (FBtype);
	fmnem("fadd"); FBH; FAH; FBH; NL;
	fround (R_N);
	fstackp -= 1;
}

void FPSUB ()
{	fscheck (FBtype);
	fmnem("fsub"); FBH; FAH; FBH; NL;
	fround (R_N);
	fstackp -= 1;
}

void FPMUL ()
{	fscheck (FBtype);
	fmnem("fmul"); FBH; FAH; FBH; NL;
	fround (R_N);
	fstackp -= 1;
}

void FPDIV ()
{	fscheck (FBtype);
	fmnem("fdiv"); FBH; FAH; FBH; NL;
	fround (R_N);
	fstackp -= 1;
}

void FPUSQRTFIRST ()
{	dummy("fpusqrtfirst"); NL;
}

void FPUSQRTSTEP ()
{	dummy("fpusqrtstep"); NL;
}

void FPUSQRTLAST ()
{	fmnem("fsqrt"); FAH; FAH; NL;
	fround (R_N);
}

void FPREMFIRST ()
{	fscheck (FBtype);

/*	The following ought to work in all cases	*/
/*	Independent of rounding mode			*/
/*	(IEEE 754-1895 5.1 and CWG p.58)		*/

	if (SNGL)
	{	mnem("fstod"); FAH; FAH; NL;
		mnem("fstod"); FBH; FBH; nl;
	}
	mnem("std"); FBH; FPtemp; NL;		/* B\A */
	mnem("ldd"); FPtemp; P0; nl;
	mnem("std"); FAH; FPtemp; nl;
	extcall(X_REMAINDER); nl;	/* Unix math library */
	mnem("ldd"); FPtemp; P2; DELAY; nl;
	/* result in %f0 = FAH */
	extcall(X_FPCONST); COMMENT("restore f.p. registers"); nl;
	fstackp = 1;
	if (SNGL)
	{	mnem("fdtos"); FAH; FAH; delay("remember ALWAYS"); nl; }
	else
	{	mnem("nop"); COMMENT("remember ALWAYS"); nl; }

#ifdef UNDEFINED

/*	Short cut for representable quotients	*/
/*	Would be better in double throughout	*/

	fround (R_N);
	fmnem("fdiv"); FBH; FAH; FTH; NL;	/* t = b/a */
	fmnem("fcmp"); FTH; ZERO; nl;		/* t = ROUND(t) */
	mnem("fabss"); FTH; FTH; nl;
	fmnem("fadd"); FTH; MAGIC; FTH; nl;
	fmnem("fsub"); FTH; MAGIC; FTH; nl;
	mnem("nop"); COMMENT("shouldn't be necessary"); nl;
	mnem("fbl,a"); f(1); COMMENT("restore sign"); nl;
	mnem("fnegs"); FTH; FTH; ANNUL; nl;
lab(1);	fmnem("fmul"); FTH; FAH; FTH; nl;	/* t = t*a */
	fmnem("fsub"); FBH; FTH;		/* a' = b - t */
		fstackp = 1;			/* tricky! */
			FAH; COMMENT("remember ALWAYS"); nl;

#endif /* UNDEFINED */

	istackp = 1;				/* just the flag */
	cond = C_ALWAYS;
}

void FPREMSTEP ()				/* should set FB */
{	dummy("fpremstep"); COMMENT("remember ALWAYS"); NL;
	fround (R_N);
	fstackp = 1;				/* assume FB ignored */
	istackp = 1;				/* just the flag */
	cond = C_ALWAYS;
}

void FPLDZEROSN ()
{	fpush (F_S);
	mnem("fmovs"); ZERO; FAH; NL;
	fround (R_N);
}

void FPLDZERODB ()
{	fpush (F_D);
/*	mnem("faddd"); ZERO; ZERO; FAH; NL;	*/
	mnem("fstod"); ZERO; FAH; NL;
	fround (R_N);
}

void FPUMULBY2 ()
{	fmnem("fadd"); FAH; FAH; FAH; NL;
/*	As fast or faster on all our SPARCs	*/
/*	Saves two registers			*/
/*	fmnem("fmul"); FAH; TWO; FAH; NL;	*/
	fround (R_N);
}

void FPUDIVBY2 ()
{	fmnem("fmul"); FAH; HALF; FAH; NL;
	fround (R_N);
}

void FPUEXPINC32 ()
{	fmnem("fmul"); FAH; TWO32; FAH; NL;
	fround (R_N);
}

void FPUEXPDEC32 ()
{	fmnem("fmul"); FAH; TWOM32; FAH; NL;
	fround (R_N);
}

void FPUABS ()
{	mnem("fabss"); FAH; FAH; NL;
	fround (R_N);
}

void FPLDNLADDSN ()
{	fscheck (F_S);
	mnem("ld"); sindex(Areg, 0); FTH; NL;
	mnem("fadds"); FAH; FTH; FAH; nl;
	fround (R_N);
	POP;
}

void FPLDNLADDDB ()
{	fscheck (F_D);
	if (o_real64)
	{	mnem("ldd"); sindex(Areg, 0); FTH; NL; }
	else
	{	mnem("ld"); sindex(Areg, HI); FTH; NL;
		mnem("ld"); sindex(Areg, LO); FTL; nl;
	}
	mnem("faddd"); FAH; FTH; FAH; nl;
	fround (R_N);
	POP;
}

void FPLDNLMULSN ()
{	fscheck (F_S);
	mnem("ld"); sindex(Areg, 0); FTH; NL;
	mnem("fmuls"); FAH; FTH; FAH; nl;
	fround (R_N);
	POP;
}

void FPLDNLMULDB ()
{	fscheck (F_D);
	if (o_real64)
	{	mnem("ldd"); sindex(Areg, 0); FTH; NL; }
	else
	{	mnem("ld"); sindex(Areg, HI); FTH; NL;
		mnem("ld"); sindex(Areg, LO); FTL; nl;
	}
	mnem("fmuld"); FAH; FTH; FAH; nl;
	fround (R_N);
	POP;
}

void FPGT ()
{	fscheck (FBtype);
	PUSH;
	if (o_checks == CH_TRAP)
		fmnem("fcmpe");
	else
		fmnem("fcmp");
	FBH; FAH; COMMENT("remember FGT"); NL;
	mnem("nop"); nl;
	fround (R_N);
	cond = C_FGT;
	fstackp -= 2;
}

void FPEQ ()
{	fscheck (FBtype);
	PUSH;
	if (o_checks == CH_TRAP)
		fmnem("fcmpe");
	else
		fmnem("fcmp");
	FBH; FAH; COMMENT("remember FEQ"); NL;
	mnem("nop"); nl;
	fround (R_N);
	cond = C_FEQ;
	fstackp -= 2;
}

void FPORDERED ()
{	fscheck (FBtype);
	PUSH;
	fmnem("fcmp"); FBH; FAH; COMMENT("remember FO"); NL;	/* fcmpe? */
	mnem("nop"); nl;
	fround (R_N);
	cond = C_FO;
/*	The instruction leaves both operands on the stack,	*/
/*	but occam uses it only in (D)ORDERED, which loses them.	*/
	fstackp -= 2;
}

void FPNAN ()
{	PUSH;
	fmnem("fcmp"); FAH; ZERO; COMMENT("remember FU"); NL;	/* fcmpe? */
	mnem("nop"); nl;
	fround (R_N);
	cond = C_FU;
}

void FPNOTFINITE ()
{	PUSH;
	if (SNGL)
	{	mnem("fabss"); FAH; FTH; NL; }
	else
	{	mnem("fdtos"); FAH; FTH; NL;	/* preserves NaN/Inf */
		mnem("fabss"); FTH; FTH; nl;
	}
	mnem("st"); FTH; FPtemp; nl;
	mnem("ld"); FPtemp; T0; nl;
	mnem("set"); hex(0x7f800000); T1; nl;
	mnem("cmp"); T0; T1; COMMENT("remember GE"); nl;
	fround (R_N);
	cond = C_GE;
}

void FPUR32TOR64 ()
{	fscheck (F_S);
	mnem("fstod"); FAH; FAH; NL;
	fround (R_N);
	FAtype = F_D;
}

void FPUR64TOR32 ()
{	fscheck (F_D);
	mnem("fdtos"); FAH; FAH; NL;
	fround (R_N);
	FAtype = F_S;
}

void FPINT ()
{	fmnem("fcmp"); FAH; ZERO; NL;		/* fcmpe? */
	mnem("fabss"); FAH; FAH; nl;
	fmnem("fadd"); FAH; MAGIC; FAH; nl;
	fmnem("fsub"); FAH; MAGIC; FAH; nl;
	mnem("nop"); COMMENT("shouldn't be necessary"); nl;
	mnem("fbl,a"); f(1); COMMENT("restore sign"); nl;
	mnem("fnegs"); FAH; FAH; ANNUL; nl;
lab(1);	nl;
	fround (R_N);
}

void FPSTNLI32 ()
{	if (SNGL)
		mnem("fstoi");
	else
		mnem("fdtoi");
	FAH; FAH; NL;
	mnem("st"); FAH; sindex(Areg, 0); nl;
	fround (R_N);
	POP;
	fstackp -= 1;
}

void FPUCHKI32 ()		/* the conversion following should be checked anyway */
{	dummy("fpuchki32"); NL;
	fround (R_N);
}

void FPUCHKI64 ()
{	dummy("fpuchki64"); NL;
	fround (R_N);
}

void FPRTOI32 ()
{	FPINT ();
	FPUCHKI32 ();
}

void FPI32TOR32 ()
{	fpush (F_S);
	mnem("ld"); sindex(Areg, 0); FAH; NL;
	mnem("fitos"); FAH; FAH; nl;
	fround (R_N);
	POP;
}

void FPI32TOR64 ()
{	fpush (F_D);
	mnem("ld"); sindex(Areg, 0); FAH; NL;
	mnem("fitod"); FAH; FAH; nl;
	fround (R_N);
	POP;
}

void FPB32TOR64 ()
{	fpush (F_D);
	mnem("ld"); sindex(Areg, 0); FAH; NL;
	mnem("fcmpd"); FAH, ZERO; nl;
	mnem("fitod"); FAH; FAH; nl;
	mnem("nop"); COMMENT("shouldn't be necessary"); nl;
	mnem("fbl,a"); f(1); nl;
	mnem("faddd"); FAH; TWO32D; FAH; ANNUL; nl;
lab(1);	nl;
	fround (R_N);
	POP;
}

void FPUNOROUND ()
{	fscheck (F_D);
	fround (R_Z);
	mnem("fdtos"); FAH; FAH; NL;
	fround (R_N);
	FAtype = F_S;
}

void FPENTRY ()
{	unimp ();
	POP;
}
