/*
 * KROC Scheduler startup code for occam
 *
 * $Source: /extra/play/cvs/kroc/src/target/unix/rtsmain.c,v $
 *
 * $Id: rtsmain.c,v 1.1 1999/09/27 21:10:13 rivimey Exp $
 *
 * (C) Copyright 1995,1996,1997 D.C. Wood <D.C.Wood@ukc.ac.uk> and 
 *                              P.H. Welch <P.H.Welch@ukc.ac.uk>
 * University of Kent at Canterbury
 *
 * Based on work by: Peter Shephard @ QMC Sept. 1984 for
 * UNIX OPS 2.0 (Release 2.0.1 : 24th. July, 1986)
 */

static char rtsmain_c_rcsid[] = "$Id: rtsmain.c,v 1.1 1999/09/27 21:10:13 rivimey Exp $";

#define RTS_MAIN
#include "rts.h"

void keyboard_process (int user_pid, int to_user_fd, int from_user_fd, bool is_a_tty);
void user_process (int kbd_pid, int to_kbd_fd, int from_kbd_fd, bool is_a_tty);

/* RTS common variables */

/* Two pipes (socketpairs) */
int kbd_chars[2];
int kbd_ack[2];

#ifdef DEBUG
bool trace;
#endif

char * short_cmdline;
char * long_cmdline;

int kroc_argc;
char ** kroc_argv;


static int my_fork (void);
static void savecmdline (int argc, char **argv);


static int
my_fork (void) {
  int n = 5;
  int cpid;
  do {
    if ((cpid = fork ()) != ERR)
      return (cpid);
    else
      sleep (2);
  } while (--n >= 0);

  fputs ("KROC: fork() failed\n", stderr);
  exit (-1);
}  /* my_fork */


static void
savecmdline (int argc, char **argv)
{
  static char cmdline[512];
  int i = 0;

  cmdline[0] = '\0';
  long_cmdline = &cmdline[0];
  short_cmdline = &cmdline[0];
  while (i < argc) {
    strcat (cmdline, argv[i]);
    i++;
    if (i < argc)
      strcat (cmdline, " ");  /* separate arguments except the last one */
    if (i == 1)
      short_cmdline = &cmdline[strlen(cmdline)];
  }
}


void
main (int argc, char **argv)
{
  int cpid;
  int is_a_tty;
  
  kroc_argc = argc;
  kroc_argv = argv;
  
  
  savecmdline (argc, argv);
  /*
   * create pipes down which :-
   *       (a) parent will send keyboard chars to occam;
   *       (b) child will acknowledge with same char
   *               (only when an Occam process inputs it);
   */
  if (pipe (kbd_chars) == ERR || pipe (kbd_ack) == ERR) {
    fputs ("failed to create pipe (s)\n", stderr);
    exit (-3);
  }

#ifdef DEBUG
  trace =  TRUE;
#endif /* DEBUG */
  
  is_a_tty = isatty (0);

  if (is_a_tty)
    save_tty_state ();

  /* fork off process to run occam kernel */
  if ((cpid=my_fork ())) {
    /* parent */
    close (kbd_chars[1]);
    close (kbd_ack[0]);
    
    /* Close stdin = fd 0 = keyboard */
    /* fclose(stdin); */
    
    user_process (cpid, kbd_ack[1], kbd_chars[0], is_a_tty);
  } else {
    /* child */
    close (kbd_chars[0]);
    close (kbd_ack[1]);
    
    fclose(stdout);
    
    keyboard_process (getppid(), kbd_chars[1], kbd_ack[0], is_a_tty);
  }

}  /* main */
